//
// Cyrano Chess engine
//
// Copyright (C) 2007  Harald JOHNSEN
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
//
//



#include "engine.hpp"
#include "gproto.hpp"
#include "eval.hpp"
#include "moves.hpp"
#include "genmagic.hpp"

#include <windows.h>

#ifdef	DEBUG
static char const s_aszModule[] = __FILE__;
#endif


enum {_WHITE,_BLACK};
enum {_EMPTY,_WKING,_WQUEEN,_WROOK,_WBISHOP,_WKNIGHT,_WPAWN,
_BKING,_BQUEEN,_BROOK,_BBISHOP,_BKNIGHT,_BPAWN};
enum {LOAD_NONE,LOAD_4MEN,SMART_LOAD,LOAD_5MEN};

#define _NOTFOUND 99999

typedef int (*PPROBE_EGBB) (int player, int w_king, int b_king,
							int piece1, int square1,
							int piece2, int square2,
							int piece3, int square3);

typedef void (*PLOAD_EGBB) (char* path,int cache_size,int load_options);
static PPROBE_EGBB probe_egbb;

int egbb_is_loaded = 0;
int egbb_load_type = LOAD_4MEN;

/*
Load the dll and get the address of the load and probe functions.
*/

#ifdef _MSC_VER
#    define EGBB_NAME "egbbdll.dll"
#else
#    define EGBB_NAME "egbbso.so"
#    define HMODULE void*
#    define LoadLibrary(x) dlopen(x,RTLD_LAZY)
#    define GetProcAddress dlsym
#endif

int LoadEgbbLibrary(char* main_path,int egbb_cache_size) {
	HMODULE hmod;
	PLOAD_EGBB load_egbb;
	char path[256];

	strcpy(path,main_path);
	strcat(path,EGBB_NAME);
	if(hmod = LoadLibrary(path)) {
		load_egbb = (PLOAD_EGBB) GetProcAddress(hmod,"load_egbb_5men");
     	probe_egbb = (PPROBE_EGBB) GetProcAddress(hmod,"probe_egbb_5men");
        load_egbb(main_path,egbb_cache_size,egbb_load_type);
		return true;
	} else {
		printf("EgbbProbe not Loaded!\n");
		return false;
	}
}
/*
Probe:
Change interanal scorpio board representaion to [A1 = 0 ... H8 = 63]
board representation and then probe bitbase.
*/

bool probe_bitbases(PCON pcon, PSTE pste, int player, int& score) {

    int piece[32],square[32];
    int count = 0;
    piece[0] = 0;
    piece[1] = 0;
    piece[2] = 0;
    square[0] = 0;
    square[1] = 0;
    square[2] = 0;

#define ADD_PIECE(list,type) \
    for (bitboard pc = list; pc; ) {\
        int sqr = PopLSB(&pc);\
        piece[count] = type;\
        square[count] = sqr;\
        count++;\
    }

    ADD_PIECE(pcon->pieces[PAWN|W],_WPAWN);
    ADD_PIECE(pcon->pieces[KNIGHT|W],_WKNIGHT);
    ADD_PIECE(pcon->pieces[BISHOP|W],_WBISHOP);
    ADD_PIECE(pcon->pieces[ROOK|W],_WROOK);
    ADD_PIECE(pcon->pieces[QUEEN|W],_WQUEEN);
    ADD_PIECE(pcon->pieces[PAWN|B],_BPAWN);
    ADD_PIECE(pcon->pieces[KNIGHT|B],_BKNIGHT);
    ADD_PIECE(pcon->pieces[BISHOP|B],_BBISHOP);
    ADD_PIECE(pcon->pieces[ROOK|B],_BROOK);
    ADD_PIECE(pcon->pieces[QUEEN|B],_BQUEEN);
    if( count > 3 )
        return false;
    score = probe_egbb(player ^ 1, pcon->king[W], pcon->king[B],
        piece[0],square[0],piece[1],square[1],piece[2],square[2]);
	
    if(score != _NOTFOUND) {
        return true;
    }

    return false;
}

typedef enum WDL {loss = -1, draw = 0 , win = 1} EGPROBE_result ;

typedef unsigned char byte;

class WDtable {
private:
    byte *table;
    int size;
public:
    WDtable(int _size) {
        size = _size;
        table = new byte[size/8];
    }
    ~WDtable() {
        delete table;
    }
    void set(int idx, enum WDL result) {
        if( result == loss )
            result = win;
        Assert(idx < size);
        Assert(result == draw || result == win);
        byte mask = 1 << (idx & 7);
        table[idx / 8] = table[idx / 8] & ~ mask | ((result & 1) << (idx & 7));
    }
    enum WDL get(int idx) {
        Assert(idx < size);
        return enum WDL((table[idx / 8] >> (idx & 7)) & 1);
    }
    void save(char *filename) {
        FILE *f = fopen(filename, "wb");
        fwrite(table, 1, size/8, f);
        fclose(f);
    }
    void load(char *filename) {
        FILE *f = fopen(filename, "rb");
        if( f ) {
            fread(table, 1, size/8, f);
            fclose(f);
        } else {
            VPrSendComment("***** could not load file %s", filename);
        }
    }
};

void WDL_set(byte *table, int idx, enum WDL result) {
    Assert(result == draw || result == win || result == loss);
    byte mask = 3 << (idx & 3);
    table[idx / 4] = table[idx / 4] & ~ mask | ((result & 3 ) << (idx & 3));
}
enum WDL WDL_get(byte *table, int idx) {
    return enum WDL((table[idx / 4] >> (idx & 3)) & 3);
}

WDtable KPK_bitbase(64*64*64*2);


static inline int sqDistance( int isqFrom, int isqTo ) {

    int dr = Rank( isqFrom ) - Rank( isqTo );
    int df = File( isqFrom ) - File( isqTo );
    if( dr < 0 ) dr = -dr;
    if( df < 0 ) df = -df;
    if( dr > df )
        return dr;
    return df;
}

static inline int idxFromSq(int sqWK, int sqBK, int sqWP, int side) {
    Assert(sqWP >= A2 && sqWP <= H7);
    return ((sqWK * 64 + sqBK) * 64 + sqWP) * 2 + side;
}


static bool testKPKpos(PCON pcon, PSTE pste, int sqWK, int sqBK, int sqWP, int side) {
    int thisIdx = idxFromSq( sqWK, sqBK, sqWP, side);

    if( sqWK == sqBK || sqWK == sqWP || sqBK == sqWP || sqDistance(sqWK, sqBK) < 2 )
        return false;

    pste->checkf    = 0;
	pste->castle	= 0;
	pste->ep_t		= 0;
	memset(pcon->pos, 0, sizeof(pcon->pos));
	memset(pcon->pieces, 0, sizeof(pcon->pieces));
	pste->hashkPc = 0;		// Going to XOR the all pieces/pawns into this.
	pste->hashkPn = 0;		// Going to XOR the pawns only into this.
    pste->side = side;
    pste->xside = side ^ 1;
    pste->mlist_start = pcon->move_list;

    AddPiece(pcon, sqWK, KING|W);
    AddPiece(pcon, sqBK, KING|B);
    AddPiece(pcon, sqWP, PAWN|W);

    int xside = side ^ 1;
	// Initialize pin/check data
	pste->pinner = 0;
	pste->pinned = 0;

    int score;
    if( probe_bitbases(pcon, pste, side, score) ) {
        if( score == 0 )
            KPK_bitbase.set( thisIdx,  draw );
        else if( score < 0 )
            KPK_bitbase.set( thisIdx, loss );
        else
            KPK_bitbase.set( thisIdx, win );
        return true;
    } else {
        score = score;
    }

    return false;
}


void initEG(PCON pcon, PSTE pste) {
    VPrSendComment("# start of initEG");
    int t = TmNow();

#ifdef _DEBUG
    LoadEgbbLibrary("F:/Chess/TB/egbbs_3_4/", 4*1024*1024);

    for(int sqWP = A2; sqWP <= H7; sqWP++)
        for(int sqWK = A1; sqWK <= H8; sqWK++)
            for(int sqBK = A1; sqBK <= H8; sqBK++)
                for( int side = B; side <= W ; side ++) {
                    testKPKpos(pcon, pste, sqWK, sqBK, sqWP, side);
                }

    KPK_bitbase.save("kpk.bit");
#else
    KPK_bitbase.load("kpk.bit");
#endif
    t = TmNow() - t;
//    testKPKpos(pcon, pste, A8, C8, A7, B);
//    t = probeKPKpos(pcon, pste, A8, C8, A7, B);
//    VPrSendComment("# val %d", KPK_bitbase.get( idxFromSq( B6, B8, A6, B) ));
//    VPrSendComment("# val %d", KPK_bitbase.get( idxFromSq( B6, B8, A6, W) ));
//    VPrSendComment("# val %d", KPK_bitbase.get( idxFromSq( A6, C8, A5, W) ));
    VPrSendComment("# end of initEG t=%d", t);
}

enum WDL probeKPKpos(PCON pcon, PSTE pste, int sqWK, int sqBK, int sqWP, int side) {
    int thisIdx = idxFromSq( sqWK, sqBK, sqWP, side);
    return KPK_bitbase.get( thisIdx );
}


//
// King & Pawn vs King
// http://en.wikipedia.org/wiki/King_and_pawn_versus_king_endgame
//
// co is the winning side
//
bool KPK(PCON pcon, PSTE pste, int co, int & val) {
    bitboard pawns = pcon->pieces[PAWN|co];
    int isqPawn = PopLSB(&pawns) ;
    int nc = co ^ 1 ;
#if 1
    isqPawn = remapSq[nc][ isqPawn ];

    int isqBKing = remapSq[nc][pcon->king[co ^ 1]];
    int isqWKing = remapSq[nc][pcon->king[co]];
    int rankPawn = Rank( isqPawn );
    int thisIdx = idxFromSq( isqWKing, isqBKing, isqPawn, co == W ? pste->side : pste->side ^ 1);
    int res = KPK_bitbase.get( thisIdx );
//    if( co == pste->side )
//        res = -res;
    int delta = - sqDistance( isqBKing, isqPawn ) * 2 + sqDistance( isqWKing, isqPawn ) * 1 + rankPawn * 8;
    switch( res ) {
        case 0:
            val = (25 + delta); // 0 gives less nodes but why not try to draw a lost game :p
            break;
        case 1:
            val = 700 + delta;
            break;
        case -1:
            val = -( 700  + delta);
            break;
    }
    return true;
#else
    // R1: square rule
    // if dist to promote < dist from king then win
    if( SquareRuleForPawn(pcon, pste, co, isqPawn, val) )
        return true;
    isqPawn = remapSq[nc][ isqPawn ];

    int isqBKing = remapSq[nc][pcon->king[co ^ 1]];
    int isqWKing = remapSq[nc][pcon->king[co]];
    int pawnFile = File(isqPawn);
    int rankPawn = Rank( isqPawn );
    int filePawn = File( isqPawn );
    if( pawnFile == filH || pawnFile == filA ) {
        if( pawnFile == File(isqBKing) ) {
            val = 50 + Rank( isqWKing );
            return true;
        }
        int dot1, dot2;
        if( pawnFile == filH ) {
            dot1 = G8; dot2 = G7;
            if( isqBKing == F8 || isqBKing == F7 || isqBKing == G8 || isqBKing == G7 ) {
                val = 50 - sqDistance( isqWKing, dot2 );
                return true;
            }
        } else {
            dot1 = B8; dot2 = B7;
            if( isqBKing == C8 || isqBKing == C7 || isqBKing == B8 || isqBKing == B7 ) {
                val = 50 - sqDistance( isqWKing, dot2 );
                return true;
            }
        }
        val = 700 + sqDistance( isqBKing, dot2 ) * 32 - sqDistance( isqWKing, dot2 ) * 4 + rankPawn;
        return true;
    } else {
        bitboard keys = 0;
        keys |= BitIdx( rankPawn + 2, filePawn - 1);
        keys |= BitIdx( rankPawn + 2, filePawn);
        keys |= BitIdx( rankPawn + 2, filePawn + 1);
        if( rankPawn >= 4 ) {
            keys |= BitIdx( rankPawn + 1, filePawn - 1);
            keys |= BitIdx( rankPawn + 1, filePawn);
            keys |= BitIdx( rankPawn + 1, filePawn + 1);
        }
        if( rankPawn == 6 ) {
            keys |= BitIdx( rankPawn, filePawn - 1);
            keys |= BitIdx( rankPawn, filePawn);
            keys |= BitIdx( rankPawn, filePawn + 1);
        }
        bool white_on_square = (IdxToU64( isqWKing ) & keys) != 0;
        bool black_on_square = (IdxToU64( isqBKing ) & keys) != 0;
        if( pawnFile != filB && pawnFile != filG ) {
            if( white_on_square ) {
                val = 700 + rankPawn*4;
                return true;
            }
            if(black_on_square ) {
                val = 50;
                return true;
            }
        }
    }
#endif
    return false;
}

// contain the distance of a square to the corner
// used to eval simple mat positions to speedup the finding of the mat

const static int c_argdistcorner[64] = {
//  a   b   c   d   e   f   g   h
	0,	5,	10,	20,	30,	40,	50,	60,
	5,	5,	10,	20,	30,	40,	50,	50,
	10,	10,	10,	20,	30,	40,	40,	40,
	20,	20,	20,	20,	30,	30,	30,	30,
	30,	30,	30,	30,	20,	20,	20,	20,
	40,	40,	40,	30,	20,	10,	10,	10,
	50,	50,	50,	30,	20,	10,	5,	5,
	60,	50,	40,	30,	20,	10,	5,	0,
};

static unsigned int colorOfSquare (unsigned int sq) {
   return  (0xAA55AA55 >> sq) & 1;
}

//
// King & Bishop & Knight vs King
//
// co is the winning side
//
bool KBNK(PCON pcon, PSTE pste, int co, int & val) {
    bitboard bishops = pcon->pieces[BISHOP|co];
    int isqBishop = LSB(bishops);
    bitboard knights = pcon->pieces[KNIGHT|co];
    int isqKnight = LSB(knights);
    int idxKnight = isqKnight;
    int nc = co ^ 1 ;
    int isqBKing = pcon->king[co ^ 1];
    int idxBKing = isqBKing;
    int isqWKing = pcon->king[co];
    int idxWKing = isqWKing;

    if( colorOfSquare( isqBishop) == 0 ) {
        idxBKing = Idx(Rank(isqBKing), 7-File(isqBKing));
        idxWKing = Idx(Rank(isqWKing), 7-File(isqWKing));
        idxKnight= Idx(Rank(isqKnight), 7-File(isqKnight));
    }
    //sqDistance
    // must lead the black king to the corner of the bishop
//    val = 800 - c_argdistcorner[idxBKing] - c_argdistcorner[idxWKing] / 2 - c_argdistcorner[idxKnight] / 4;
    val = 800 - c_argdistcorner[idxBKing] - sqDistance(isqWKing, isqBKing) * 2 - sqDistance(isqKnight, isqBKing);
    return true;
}
