//
// Cyrano Chess engine
//
// Copyright (C) 2007  Harald JOHNSEN
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
//
//


#include "engine.hpp"
#include "eval.hpp"
#include "moves.hpp"
#include "magicmoves.h"
#include "genmagic.hpp"
#include <stdlib.h>     // for dummy rand()

#ifdef _DEBUG
static char const s_aszModule[] = __FILE__;
#endif


const U64 fileAMask	= 0x0101010101010101ULL;
const U64 fileHMask = 0x8080808080808080ULL;
//#define fileAMask	0x0101010101010101ULL
//#define fileHMask   0x8080808080808080ULL

const int promotionSqrFileA[2] = {A1,A8};
const int promotionSqrFileH[2] = {H1,H8};

static bitboard attackMap[coMAX];
static int posval[coMAX];       // position value
static int tacval[coMAX];       // tactical value
static int otherval[coMAX];     // other value ?
static int endgval[coMAX];      // end game values
static int notpawn[coMAX];
static int tempoval[coMAX];
static int attackval[coMAX];
static int c_mobility[coMAX];
static int mobility[coMAX];
static int defSpace[coMAX];
static int pawnDist[coMAX];
// count of pieces by type and color
static int c_pieces[PIECES];
static bitboard knight_moves[coMAX];
static bitboard tmp_attack[coMAX];
static int bishopSquare[coMAX];
static PHASHPawn phashp;
static int mat[coMAX];
static int underdev[coMAX];

static struct mobDebug {
    int piece;
    int mobility;
} mobDebugTable[32];
static int mobDebugIdx;

//==== ATTACK TABLE ============================================================

// I only look at the attacker type
//U8 attackMap[2][64];
typedef struct _posdir {
    U8 pos;
    U8 dir;
} posdir;
static posdir PosDirTab[8][64][64];

// attacker type sorted by value
#define ATK_PAWN    0x80
#define ATK_MINOR   0x20    // *4 => 0x80
#define ATK_ROOK    0x08    // *4 => 0x20
#define ATK_QUEEN   0x02    // *4 => 0x08
#define ATK_KING    0x01

/* piece code defines */
#define APC_no_piece 0
#define APC_pawn 1
#define APC_knight 2
#define APC_bishop 3
#define APC_rook 4
#define APC_queen 5
#define APC_king 6
#define APC_bpawn 7
#define APC_BLACKKING   12

//const int KING		= 0;
//const int QUEEN		= 2;
//const int ROOK		= 4;
//const int BISHOP	= 6;
//const int KNIGHT	= 8;
//const int PAWN		= 10;

// mobility counter increment depending on square content
// t[12] = black king
//                      -,WK,BQ,WQ,BR,WR,BB,WB,BN,WN,BP,WP,BK
// 0 is empty, BK replaced with 12 in pos[]
const U8 mobCount[2][16] = {{1,1,0,1,0,1,0,1,0,1,0,1,0},
                            {1,0,1,0,1,0,1,0,1,0,1,0,1}};

static char stringPc[] = "KkQqRrBbNnPp";

#define mobK(n)     (n)
#define queenK(n)   ((3*mobK(n))/4)
#if 1

// mobility adjusted with unsafe mobility
static int KnightMobility[1+8] = {mobK(-30),mobK(-20),mobK(-10),mobK(0),mobK(10),
    mobK(15),mobK(20),mobK(22),mobK(25)};

// 14 * ((64-32) / 64) => 7,        14 * ((64-20)/64) => 9.6
static int BishopMobility[1+14] = {mobK(-35),mobK(-25),mobK(-20),mobK(-10),mobK(-5),mobK(0),mobK(5),
    mobK(10),mobK(12),mobK(15),mobK(20),mobK(25),mobK(30),mobK(32),mobK(35)};

// first and last ranks don't really count
// 14 * ((48-32)/48)   => 4.6,      14 * ((48-20)/48) => 8.1
static int RookMobility[1+14] = {mobK(-30),mobK(-20),mobK(-10),mobK(-5),mobK(0),mobK(5),mobK(10),
    mobK(15),mobK(20),mobK(22),mobK(25),mobK(27),mobK(30),mobK(30),mobK(30)};

// bishop+rook 32 => 11.6       20 => 17.7
static int QueenMobility[1+28] = {queenK(-50),queenK(-40),queenK(-25),queenK(-15),queenK(-7),queenK(-5),queenK(-3),queenK(-1),
/*8*/    queenK(0),queenK(2),queenK(5),queenK(7),queenK(10),queenK(12),queenK(15),queenK(17),queenK(20),queenK(21),queenK(22),
/*19*/    queenK(23),queenK(24),queenK(25),queenK(26),queenK(27),queenK(28),queenK(30),queenK(30),queenK(30),queenK(30)};

#else
// base value
static int KnightMobility[1+8] = {mobK(-30),mobK(-20),mobK(-10),mobK(0),mobK(10),
    mobK(15),mobK(20),mobK(22),mobK(25)};

// 14 * ((64-32) / 64) => 7,        14 * ((64-20)/64) => 9.6
static int BishopMobility[1+14] = {mobK(-30),mobK(-25),mobK(-20),mobK(-15),mobK(-10),mobK(-5),mobK(0),
    mobK(5),mobK(10),mobK(15),mobK(20),mobK(25),mobK(30),mobK(32),mobK(35)};

// first and last ranks don't really count
// 14 * ((48-32)/48)   => 4.6,      14 * ((48-20)/48) => 8.1
static int RookMobility[1+14] = {mobK(-30),mobK(-20),mobK(-15),mobK(-10),mobK(-5),mobK(0),mobK(10),
    mobK(15),mobK(20),mobK(22),mobK(25),mobK(27),mobK(30),mobK(30),mobK(30)};

// bishop+rook 32 => 11.6       20 => 17.7
static int QueenMobility[1+28] = {queenK(-50),queenK(-40),queenK(-30),queenK(-20),queenK(-10),queenK(-7),queenK(-6),queenK(-5),
    queenK(-4),queenK(-3),queenK(-2),queenK(0),queenK(2),queenK(5),queenK(7),queenK(10),queenK(12),queenK(15),queenK(17),
    queenK(20),queenK(22),queenK(25),queenK(27),queenK(30),queenK(30),queenK(30),queenK(30),queenK(30),queenK(30)};
#endif

//==============================================================================


const unsigned char castle_mask[2] = { 12, 3 };

// I don't want to use one array per color for positional values
// this array precomputes rank swap
int const remapSq[2][64] = {
    {
//  a   b   c   d   e   f   g   h
	0,	1,	2,	3,	4,	5,	6,	7,
	8,	9,	10,	11,	12,	13,	14,	15,
	16,	17,	18,	19,	20,	21,	22,	23,
	24,	25,	26,	27,	28,	29,	30,	31,
	32,	33,	34,	35,	36,	37,	38,	39,
	40,	41,	42,	43,	44,	45,	46,	47,
	48,	49,	50,	51,	52,	53,	54,	55,
	56,	57,	58,	59,	60,	61,	62,	63
    },{
	56,	57,	58,	59,	60,	61,	62,	63,
	48,	49,	50,	51,	52,	53,	54,	55,
	40,	41,	42,	43,	44,	45,	46,	47,
	32,	33,	34,	35,	36,	37,	38,	39,
	24,	25,	26,	27,	28,	29,	30,	31,
	16,	17,	18,	19,	20,	21,	22,	23,
	8,	9,	10,	11,	12,	13,	14,	15,
	0,	1,	2,	3,	4,	5,	6,	7
    }
};

//
//	These are piece-square tables.  The award bonuses (or penalties) in
//	centipawns if a piece is sitting on a particular square
//

/*
const int KING		= 0;    => 0
const int QUEEN		= 2;    => 1
const int ROOK		= 4;    => 2
const int BISHOP	= 6;    => 3
const int KNIGHT	= 8;    => 4
const int PAWN		= 10;   => 5
*/
static int const staticPosVal[6+2][64] = {
	//
	//	White king.
	//
//  a   b   c   d   e   f   g   h
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-20,-20,-20,-20,-20,-20,-20,-20,
	-7,	-15,-15,-15,-15,-15,-15,-7,
	-5,	-5,	-12,-12,-12,-12,-5,	-5,
	 3,	 5,	  6,-12, -9,-12, 8,	 6,
	//
	//	White queen.
	//
//  a   b   c   d   e   f   g   h
	-2,	-2,	0,	0,	0,	0,	0,	0,
	-2,	-2,	0,	0,	0,	0,	0,	0,
	-2,	-2,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	2,	2,	0,	0,	0,
	0,	0,	0,	2,	2,	0,	0,	0,
	0,	1,	1,	1,	1,	0,	0,	0,
	0,	0,	1,	1,	1,	0,	0,	0,
	-2,	-2,	-2,	0,	0,	-2,	-2,	-2,
	//
	//	White rook.
	//
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,
	8,	8,	8,	8,	8,	8,	8,	8,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-2,	0,	0,	0,	0,	0,	0,	-2,
	-4,	0,	0,	0,	0,	0,	0,	-4,
	-4,	0,	3,	6,	6,	3,	0,	-4,
	//
	//	White bishop.
	//
//  a   b   c   d   e   f   g   h
	5,	5,	5,	5,	5,	5,	5,	5,
	8,	8,	8,  8,  8,	8,	8,	8,
	8,	8,	8,  8,  8,	8,	8,	8,
	8,	9,	8,  8,  8,	8,  9,	8,
	8,	8,  9,  8,  8,  9,	8,	8,
	6,	8,	8,  8,  8,	8,	8,	6,
	4,  7,	5,  7,  7,	5,  7,	4,
	0, -4,	0,	0,	0,	0, -4,	0,
	//
	//	White knight.
	//
//  a   b   c   d   e   f   g   h
	-9,-5,-5, -5, -5, -5, -5, -9,
	-7,	0,  8,  8,  8,  8,	0, -7,
	-7,	5,  9,  9,  9,  9,	5, -7,
	-7,	7,	7,  9,  9,	7,	7, -7,
	-7,	8,	5,	6,	6,	5,	8, -7,
	-7,	5,  8,	5,	5,  8,	5, -7,
	-7,	0,  5,  7,  7,  5,	0, -7,
	-9,-2, -2, -2, -2, -2, -2, -9,
	//
	//	White pawn.
	//
// TODO:not up to date =============================================================
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,  // a8..h8
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0, 15, 15,	0,	0,	0,
	0,	0,	4,	7,	7,	0,	0,	0,
	0,	0,	5,  7,  7,	4,	0,	0,
	2,	2,	1,	5,	5,	0,	2,	2,
	5,	5,	2,	0,	0,	5,	5,	5,
	0,	0,	0,	0,	0,	0,	0,	0,  // a1..h1
};

//
//

static int	const tempoVal[6+2][64] = {
	//
	//	White king.
	//
//  a   b   c   d   e   f   g   h
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  4,  0,  0,  0,  4,  0,
	//
	//	White queen.
	//
//  a   b   c   d   e   f   g   h
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	//
	//	White rook.
	//
//  a   b   c   d   e   f   g   h
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	0,  4,  4,  4,  4,  4,  4,  0,
	0,  0,  4,  4,  4,  4,  0,  0,
	//
	//	White bishop.
	//
//  a   b   c   d   e   f   g   h
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  0,  4,  4,  0,  4,  4,
	//
	//	White knight.
	//
//  a   b   c   d   e   f   g   h
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	2,  4,  4,  4,  4,  4,  4,  2,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  0,  4,  4,  4,  4,  0,  4,
	//
	//	White pawn.
	//
//  a   b   c   d   e   f   g   h
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  1,  2,  2,  1,  0,  0,
	0,  0,  1,  3,  3,  1,  0,  0,
	0,  0,  3,  3,  3,  1,  0,  0,
	0,  0,  3,  3,  3,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
};

static const int knightOutpostVal[64] = {
	//
	//	White knight.
	//
//  a   b   c   d   e   f   g   h
	 0, 0, 0,  0,  0,  0,  0,  0,
	 0,	0,  7,  7,  7,  7,	0, 0,
	 0,	0,  9,  9,  9,  9,	0, 0,
	 0,	0,	7,  9,  9,	7,	0, 0,
	 0,	0,	3,	5,	5,	3,	0, 0,
	 0, 0, 0,  0,  0,  0,  0,  0,
	 0, 0, 0,  0,  0,  0,  0,  0,
	 0, 0, 0,  0,  0,  0,  0,  0,
};

//	The king is a weird piece because in the engame its eval is approximately
//	opposite of its eval in the middlegame or opening.  At the start of the
//	game, you want to hide your king in the corner, and at the end, you want
//	to run to the middle.

//	This problem is taken care of by having two tables.  If you try to make
//	things work with the middlegame table, the program will play like crap in
//	endings.

static int	const KingEGPosVal[64] = {
	//
	//	White king.
	//
//  a   b   c   d   e   f   g   h
	-5,	0,	0,	0,	0,	0,	0, -5,
	-0,	4,	3,	3,	3,	3,	4,  0,
	-1,	3,	7,	7,	7,	7,	3, -1,
	-2,	3,	7,	8,	8,	7,	3, -2,
	-2,	3,	6,	8,	8,	6,	3, -2,
	-2,	3,	5,	7,	7,	5,	3, -2,
	-3,	0,	4,	3,	3,	4,	0, -3,
	-5, -2,-1,	-1,	-1,	-1, -2, -5,
};

#define	sideK	0
#define	sideC	1
#define	sideQ	2

#define	sideMAX	3

//	I didn't use the #defines because that is hard to format.  This is an
//	array of char instead of int, which is a decision based upon space.
//
//	Note that c1 is considered a q-side square but f1 is not considered a
//	k-side square.  I want to do everything possible to:
//
//	1.	Convince the program to castle, and O-O-O, which puts the king on
//		c1, is a fine thing to do.
//
//	2.	Prevent the program from playing Kf1.  f1 is not a good square for
//		the king, and I'd rather it play O-O if it's going to move its king
//		into the k-side corner.

static char const c_argside[64] = {
//  a   b   c   d   e   f   g   h
	2,	2,	2,	1,	1,	1,	0,	0,
	2,	2,	1,	1,	1,	1,	0,	0,
	1,	1,	1,	1,	1,	1,	1,	1,
	1,	1,	1,	1,	1,	1,	1,	1,
	1,	1,	1,	1,	1,	1,	1,	1,
	1,	1,	1,	1,	1,	1,	1,	1,
	2,	2,	1,	1,	1,	1,	0,	0,
	2,	2,	2,	1,	1,	1,	0,	0,
};

//	These are edits to the pawn positional values based upon king location.
//	This one assumes that we're dealing with a white king on the q-side.
//	Non-zero values will replace the values in the regular pawn table.

static char const XXX_argvalShelter[64] = {
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	10,	11,	18,	0,	0,	0,	0,	0,
	15,	15,	18,	0,	0,	0,	0,	0,
	20,	20,	12,	0,	0,	0,	0,	0,
	35,	35,	12,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
};

static int	XXX_argvalPawn[sideMAX][64];

// contain the distance of a square to the border
// used to eval simple mat positions to speedup the finding of the mat
// higher value when opposite king is near the border

const static int c_argdistborder[64] = {
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	10,	10,	10,	10,	10,	10,	0,
	0,	10,	20,	20,	20,	20,	10,	0,
	0,	10,	20,	30,	30,	20,	10,	0,
	0,	10,	20,	30,	30,	20,	10,	0,
	0,	10,	20,	20,	20,	20,	10,	0,
	0,	10,	10,	10,	10,	10,	10,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
};

// minor value less when there is fewer pawns
static int minor_eg_malus[9] = {100,75,50,25,15,0,0,0,0};

static int pawn_bonus_to_7[coMAX][8] = {{0,90,65,35,17,5,0,0}};

// f(mat) = mat / 50 if mat > Q+B+K else zero
// all mat => 100; mat==Q+R => 0; mat < Q+R => 0;
static int potentialPressure[32] = {0};
// piece:Q+R+R+B+B+N+N + K
// full score when 8 pieces (start of game)
// 0 or near 0 when only the king left
// mid game position tables are worthless in end game
//#define MidGameValue(n,pop)     (((n)*(pop))/8)
//#define EndGameValue(n,pop)     (((n)*(8-pop))/8)
static int MidGameValueCoef[9] = {0,0,0,1,2,5,7,8,8};
static int EndGameValueCoef[9] = {8,8,8,7,6,2,0,0,0};
#define MidGameValue(n,pop)     (((n)*MidGameValueCoef[pop])/8)
#define EndGameValue(n,pop)     (((n)*EndGameValueCoef[pop])/8)

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-


static void Initialize_moves (void);

void VInitMiddlegamePawnTable(void)
{
#if 0
	for (int side = sideK; side <= sideQ; side++) {
		for (int co = coWHITE; co <= coBLACK; co++) {
			memcpy(c_argvalPawn[side][co], c_argvalPos[pcPAWN][co],
				csqMAX * sizeof(int));
			if (side != sideC) {
				bool fFlipFil = (side == sideK);
				bool fFlipRnk = (co == coBLACK);

				for (int rnk = rnk2; rnk <= rnk7; rnk++)
					for (int fil = filA; fil <= filH; fil++) {
						int	isq = IsqFromRnkFil(rnk, fil);

						if (c_argvalShelter[isq] != 0) {
							int	rnkEdit = (fFlipRnk) ? 7 - rnk : rnk;
							int	filEdit = (fFlipFil) ? 7 - fil : fil;
							int	isqEdit = IsqFromRnkFil(rnkEdit, filEdit);

							c_argvalPawn[side][co][isqEdit] =
								c_argvalShelter[isq];
						}
					}
			}
		}
	}
#endif
    Initialize_moves();
    // TODO:check that ==========================================================
    for(int i = 0; i < 8 ; i++)
        pawn_bonus_to_7[coWHITE][7-i] = pawn_bonus_to_7[coBLACK][i];
    for(int i = 0; i < 32; i++) {
        int valMat = i * 128;
        if( valMat < (valQUEEN))
            potentialPressure[i] = 0;
        else {
            potentialPressure[i] = (128*(valMat-valQUEEN)) / (valROOK+valKNIGHT);
            if(potentialPressure[i] > 128)
                potentialPressure[i] = 128;
        }

    }

}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

static int const c_argdsqKing[64] = {
	//
	//	White king.
	//
//  a   b   c   d   e   f   g   h
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	-1,	-1,	0,	0,	1,	1,	0,
};


#define SAFEMOB     4
#define UNSAFEMOB   (2-SAFEMOB)
#if 0
static INLINE unsigned int mob_slider_attacks(const CON *pcon, int piece, int side, int sqr, U8 ctl) {
    posdir *nextPosDir = PosDirTab[piece][sqr];
    int u = nextPosDir[sqr].pos;
    int mob = 0;
    do {
        attackMap[side][u] += ctl;
//        mob += mobCount[side][pcon->pos[u]];
        if( mobCount[side][pcon->pos[u]] ) {
            // empty or ennemy piece/pawn
            mob += SAFEMOB;
            // if allready attacked, it's a lower value pawn/piece that attack this square,
            // not safe to go there
            if( attackMap[side^1][u] >= (ctl*4) )
                mob += UNSAFEMOB;
        }
        if( pcon->pos[u] )
            u = nextPosDir[u].dir;
        else
            u = nextPosDir[u].pos;
    } while (u != sqr);
    return mob>>2;
}

static INLINE unsigned int mob_non_slider_attacks(const CON *pcon, int piece, int side, int sqr, U8 ctl) {
    posdir *nextPosDir = PosDirTab[piece][sqr];
    int u = nextPosDir[sqr].dir;
    int mob = 0;
    do {
        attackMap[side][u] += ctl;
//        mob += mobCount[side][pcon->pos[u]];
        if( mobCount[side][pcon->pos[u]] ) {
            // empty or ennemy piece/pawn
            mob += SAFEMOB;
            // if allready attacked, it's a lower value pawn/piece that attack this square,
            // not safe to go there
            if( attackMap[side^1][u] >= (ctl*4) )
                mob += UNSAFEMOB;
        }
        u = nextPosDir[u].dir;
    } while (u != sqr);
    return mob>>2;
}
static INLINE void mob_pawn_attacks(const CON *pcon, int piece, int side, int sqr, U8 ctl) {
    posdir *nextPosDir = PosDirTab[piece][sqr];
    int u = nextPosDir[sqr].dir;
    attackMap[side][u] |= ctl;
    u = nextPosDir[u].dir;
    if (u != sqr)
        attackMap[side][u] |= ctl;
}

#endif

// from Gerd
static unsigned int colorOfSquare (unsigned int sq) {
   return  (0xAA55AA55 >> sq) & 1;
}
// from h.g.m
static bool sameSquareColor(unsigned int sq1, unsigned int sq2) {
    sq1 ^= sq2;
    return  ( (0xAA55AA55 >> sq1) & 1);
}

static int sqDistance( int isqFrom, int isqTo ) {

    int dr = Rank( isqFrom ) - Rank( isqTo );
    int df = File( isqFrom ) - File( isqTo );
    if( dr < 0 ) dr = -dr;
    if( df < 0 ) df = -df;
    if( dr > df )
        return dr;
    return df;
}
// does not check if there is two bishops
static bool bishopcanreach(const CON *pcon, int side, int sqr ) {
    bitboard bishops = pcon->pieces[BISHOP|side];
    if( bishops == 0 )
        return false;
    int bishop_sqr = LSB( bishops );
    return sameSquareColor(bishop_sqr, sqr);
}

//
// Square rule
// used to see a winning position with KP vs K
// and to give a bonus for passed pawn based on distance to king
//
static bool SquareRuleForPawn(const CON *pcon, const STE *pste, int co, int isqPawn, int & val) {
    // compute ranks from pawn to promotion
    // compute file dist from opposite king to pawn
    // if dist to promote < dist from king then win
    int myturn = pste->side != co;
    int isqKing = pcon->king[co ^ 1];
    int isqWKing = pcon->king[co];
    int distPawn;
    int rankKing = 7 - Rank( isqKing );
    int rankPawn = 7 - Rank( isqPawn );
    val = 0;
    // can not count if own king on pawn file
    if( File(isqWKing) == File(isqPawn) )
        return false;
    if( co != coWHITE ) {
        distPawn = 7 - rankPawn;
        rankKing = 7 - rankKing;
    } else
        distPawn = rankPawn;
    if( rankKing > (distPawn + myturn) ) {
        val = 750 + (7-distPawn)*4;
        return true;
    }
    int distFil = File( isqKing ) - File( isqPawn );
    if( distFil < 0 )
        distFil = - distFil;
    if( distFil >  (distPawn + myturn) ) {
        val = 750 + (7-distPawn)*4;
        return true;
    }
    return false;
}

/*
static bool SquareRuleKPK(PCON pcon, PSTE pste, int co, int & val) {
    // if dist to promote < dist from king then win
    bitboard pawns = pcon->pieces[PAWN|co];
    if(pawns) {
        int isqPawn = PopLSB(&pawns);
        return SquareRuleForPawn(pcon, pste, co, isqPawn, val);
    }
    return false;
}
*/

static bitboard BitIdx(int rank, int file) {
    if( rank >= 0 && rank <= 7 && file >=0 && file <= 7 )
        return IdxToU64( Idx(rank, file) );
    else
        return 0;
}

static void inline mobDebugADD(int pc, int m) {
#ifdef _DEBUG
    mobDebugTable[mobDebugIdx].piece = pc;
    mobDebugTable[mobDebugIdx].mobility = m;
    mobDebugIdx++;
#endif
}


// friendly pawns in front of the king
// ennemy pawns in front of the king
// open files on the king
static int pawnOnKing(const CON *pcon, const STE *pste, int co, int sqr) {
    const U8 *pos = pcon->pos;
    int val = 0;
    int dir = co == W ? 1 : -1;
    int r = Rank(sqr);
    int f = File(sqr);
    bitboard myPawns = pcon->pieces[PAWN|co];
    bitboard theirPawns = pcon->pieces[PAWN|co^1];
    for(int ff = -1; ff <= 1; ff++) {
        int fp = f + ff;
        if( BitIdx(r + dir, fp) & myPawns )
            val += 8;
        else if( BitIdx(r + dir*2, fp) & myPawns )
            val += 5;
        else if( BitIdx(r + dir*3, fp) & myPawns )
            val += 2;
        if( BitIdx(r + dir, fp) & theirPawns )
            val -= 50;
        else if( BitIdx(r + dir*2, fp) & theirPawns )
            val -= 35;
        else if( BitIdx(r + dir*3, fp) & theirPawns )
            val -= 15;
        else {
            if( fp >=0 && fp <= 7 ) {
                // no ennemy pawns on this file ?
                if( phashp->side[co^1].c_pawn[fp] == 0 ) {
                    val -= 20;
                    // no pawn to protect
                    if( phashp->side[co].c_pawn[fp] == 0 )
                        val -= 20;
                }
            }
        }
    }
    return val;
}

// this has mostly to do with king safety
static int evalKingCastle(const CON *pcon, const STE *pste, int co) {
    int sqr = pcon->king[co];
    const U8 *pos = pcon->pos;
    int dir = co == W ? 1 : -1;
    int r = Rank(sqr);
    int f = File(sqr);
    // friendly pawns in front of the king
    // ennemy pawns in front of the king
    // compute safety value of the king present position
    int val = pawnOnKing(pcon, pste, co, sqr);
    // compute safety value of the king future position if the castle can still be done
    // if the king has not castled yet (perhaps because of some tactical moves)
    // and can still castle then blend the 2 scores

    int CastleSqrVal = -999;
    if( (co == W && (pste->castle & cfE1G1)) || (co == B && (pste->castle & cfE8G8))) {
        // king side castle
        CastleSqrVal = pawnOnKing(pcon, pste, co, co == W ? G1 : G8);
        // pieces between king and rook are delaying the castle
        if( pos[sqr+1] != 0 )   CastleSqrVal -= 7;
        if( pos[sqr+2] != 0 )   CastleSqrVal -= 7;
    }
    if( (co == W && (pste->castle & cfE1C1)) || (co == B && (pste->castle & cfE8C8))) {
        // queen side castle
        int QueenCastleSqrVal = pawnOnKing(pcon, pste, co, co == W ? C1 : C8);
        // pieces between king and rook are delaying the castle
        if( pos[sqr-1] != 0 )   QueenCastleSqrVal -= 5;
        if( pos[sqr-2] != 0 )   QueenCastleSqrVal -= 5;
        if( pos[sqr-3] != 0 )   QueenCastleSqrVal -= 5;
        if( QueenCastleSqrVal > CastleSqrVal )
            CastleSqrVal = QueenCastleSqrVal;
    }
    if( CastleSqrVal > val ) {
        // We can still castle and it's better for us, blend the scores.
        // Scores are averaged because the current score is not permanent.
        // how much of each ???
        val = (CastleSqrVal + val) / 2;
    }
    if( pcon->pieces[QUEEN|co^1] == 0 )
        val /= 4;
    return val;
}

    // for each piece of the board...
#define pieceheader(kind) \
		int sqr = PopLSB(&pieces);  \
		int pc = kind|co;          \
        int neutralPc = kind;       \
        /* neutral color square for positional values */ \
        int nSqr = remapSq[co][sqr]; \
        /* count kind of pieces */ \
        c_pieces[kind|co]++; \
        /* basic positional value */ \
        posval[co] += staticPosVal[neutralPc >> 1][nSqr]; \
        tempoval[co] += tempoVal[neutralPc >> 1][nSqr]; \
        if( neutralPc != PAWN ) \
            notpawn[co]++;


static INLINE void evalPAWN(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
	for (bitboard pieces = pcon->pieces[PAWN|co]; pieces; ) {
        pieceheader(PAWN);

        bitboard thispawn = IdxToU64(sqr);
        int rankBonus = pawn_bonus_to_7[co][Rank(sqr)];
        // give a little bonus for passed pawns
        if( phashp->side[co].passed & thispawn ) {
            // protected passed pawn
            if( thispawn & attackMap[co] )
                otherval[co] += 4;
            // this will incite the king to move to candidate/passed pawn of other side
            endgval[co] += 4*sqDistance(sqr, pcon->king[co^1]);
            // this will incite the king to move to own candidate/passed pawn
            endgval[co] -= 2*sqDistance(sqr, pcon->king[co]);
            if( mat[co ^ 1] == 0 ) {
                int specialVal;
                // compute square rule
                SquareRuleForPawn(pcon, pste, co, sqr, specialVal);
                if( specialVal > 0 ) {
                    endgval[co] += (rankBonus + specialVal);
                } else {
                    endgval[co] += (50+rankBonus);
                }
                posval[co] += 50 + rankBonus;
            } else {
//                  posval[co] += 10 + rankBonus / 2;
                posval[co] += rankBonus;
                endgval[co] += (50+rankBonus);
            }
        } else if( phashp->side[co].candidate & thispawn ) {
            // candidate pawn
            // protected candidate
            if( thispawn & attackMap[co] )
                otherval[co] += 3;
//              posval[co] += 5 + rankBonus / 4;
            posval[co] += 5 + rankBonus / 2;
            endgval[co] += (20+rankBonus);
            endgval[co] += 2*sqDistance(sqr, pcon->king[co^1]);
            // this will incite the king to move to own candidate/passed pawn
            endgval[co] -= 1*sqDistance(sqr, pcon->king[co]);
        }
    }
}

static INLINE void evalKNIGHT(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
	for (bitboard pieces = pcon->pieces[KNIGHT|co]; pieces; ) {
        pieceheader(KNIGHT);

        //	Computers are by default very stupid about putting
        //	bishops and knights on d3 and e3 while there is still
        //	a pawn on d2 or e2.  I will detect this case and
        //	penalize heavily.
        //
        if(co == W && (sqr == E3 || sqr == D3) && (pos[sqr - 8] == WP) )
            posval[co] -= 30;
        if(co == B && (sqr == E6 || sqr == D6) && (pos[sqr + 8] == BP) )
            posval[co] -= 30;

        // knight outpost
        if( knightOutpostVal[nSqr] ) {
            bitboard thisSqr = IdxToU64(sqr);
            // not attacked by an ennemy pawn
            if( (thisSqr & attackMap[co^1]) == 0 )
                // defended by my pawns
                if( thisSqr & attackMap[co] ) {
                    posval[co] += knightOutpostVal[nSqr];
                }
        }
        bitboard moves = knight_attacks[sqr];
        tmp_attack[co] |= moves;
        moves &= ~pcon->pieces[B_OCCU|co];
//        bitboard futurOutpost = moves & outpostMap & attackMap[co] & ~attackMap[co^1];
//        int futurOutpostVal = 0;
//        if( futurOutpost ) {
//            futurOutpostVal = 10/2; // outpostValue[nSqr]/2;
//        }
//        int mob = mob_non_slider_attacks(pcon, APC_knight, co, sqr, ATK_MINOR);
        int mob = (popCount(moves) * 2 - popCount(moves & attackMap[co^1]))>>1;
        mobDebugADD(pc,mob);
        c_mobility[co] += mob;
        Assert(mob>=0 && mob <= 8);
        mobility[co] += KnightMobility[mob];
//        int ov = evalKnightOutpost(sqr,co); 
        if( nSqr == B8 || nSqr == G8)
            underdev[co]++;
    }
}

static INLINE void evalBISHOP(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
	for (bitboard pieces = pcon->pieces[BISHOP|co]; pieces; ) {
        pieceheader(BISHOP);
        bishopSquare[co] = sqr;
        //	Computers are by default very stupid about putting
        //	bishops and knights on d3 and e3 while there is still
        //	a pawn on d2 or e2.  I will detect this case and
        //	penalize heavily.
        //
        if(co == W && (sqr == E3 || sqr == D3) && (pos[sqr - 8] == WP) )
            posval[co] -= 30;
        if(co == B && (sqr == E6 || sqr == D6) && (pos[sqr + 8] == BP) )
            posval[co] -= 30;

//        int mob = mob_slider_attacks(pcon, APC_bishop, co, sqr, ATK_MINOR);
        bitboard moves = Bmagic(sqr, pcon->pieces[OCCU]);
        tmp_attack[co] |= moves;
        moves &= ~pcon->pieces[B_OCCU|co];
        int mob = (popCount(moves) * 2 - popCount(moves & attackMap[co^1]))>>1;
        mobDebugADD(pc,mob);
        c_mobility[co] += mob;
        Assert(mob>=0 && mob <= 14);
        mobility[co] += BishopMobility[mob];
        // the bishop is a little better than the knight in eg
        endgval[co] += 25;
        if( nSqr == C8 || nSqr == F8)
            underdev[co]++;
    }
}

static INLINE void evalROOK(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
	for (bitboard pieces = pcon->pieces[ROOK|co]; pieces; ) {
        pieceheader(ROOK);
//        int mob = mob_slider_attacks(pcon, APC_rook, co, sqr, ATK_ROOK);
        bitboard moves = Rmagic(sqr, pcon->pieces[OCCU]);
        tmp_attack[co] |= moves;
        moves &= ~pcon->pieces[B_OCCU|co];
        int mob = (popCount(moves) * 2 - popCount(moves & attackMap[co^1]))>>1;
        mobDebugADD(pc,mob);
        c_mobility[co] += mob;
        Assert(mob>=0 && mob <= 14);
        mobility[co] += RookMobility[mob];

        int rookF = File(sqr);
        // rook on semi-open file
        if(phashp->side[co].c_pawn[rookF] == 0) {
            posval[co] += 7;
            // rook on open file
            if(phashp->side[co^1].c_pawn[rookF] == 0)
                posval[co] += 7;
        }
        int knSqr = remapSq[co][pcon->king[co]];
        // trapped rook
        if( phashp->side[co].c_pawn[rookF] ) {
            if( nSqr == A8 && (knSqr == B8 || knSqr == C8) )
                posval[co] -= 50;
            if( nSqr == B8 && (knSqr == C8) )
                posval[co] -= 50;
            if( nSqr == H8 && (knSqr == F8 || knSqr == G8) )
                posval[co] -= 50;
            if( nSqr == G8 && (knSqr == F8) )
                posval[co] -= 50;
        }
    }
}

static INLINE void evalQUEEN(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
	for (bitboard pieces = pcon->pieces[QUEEN|co]; pieces; ) {
        pieceheader(QUEEN);
        if(co == W && (sqr == E3 || sqr == D3) && (pos[sqr - 8] == WP) )
            posval[co] -= 20;
        if(co == B && (sqr == E6 || sqr == D6) && (pos[sqr + 8] == BP) )
            posval[co] -= 20;
//        int mob = mob_slider_attacks(pcon, APC_queen, co, sqr, ATK_QUEEN);
        bitboard moves = Qmagic(sqr, pcon->pieces[OCCU]);
        tmp_attack[co] |= moves;
        moves &= ~pcon->pieces[B_OCCU|co];
        int mob = (popCount(moves) * 2 - popCount(moves & attackMap[co^1]))>>1;

        mobDebugADD(pc,mob);
        c_mobility[co] += mob;
        Assert(mob>=0 && mob <= 28);
        mobility[co] += QueenMobility[mob];
        // don't move the queen before the minors
        if( (nSqr != D8) && (notpawn[co] >= 5) ) {
            posval[co] -= underdev[co] * 8;
        }
    }
}

static INLINE void evalKING(const CON *pcon, const STE *pste, int co) {
    const U8 *pos = pcon->pos;
    int sqr = pcon->king[co];
    int neutralPc = KING;       
    int pc = KING|co;
    /* neutral color square for positional values */
    int nSqr = remapSq[co][sqr];
    /* count kind of pieces */
    c_pieces[KING|co]++;
    /* basic positional value */
    posval[co] += staticPosVal[neutralPc >> 1][nSqr];
    tempoval[co] += tempoVal[neutralPc >> 1][nSqr];
    notpawn[co]++;

    if( mat[co ^ 1] > 1100 )
        posval[co] += evalKingCastle(pcon, pste, co);

    endgval[co] += KingEGPosVal[nSqr];
    // add king mobility for endgame
    bitboard moves = king_attacks[sqr] & ~attackMap[co^1];
    tmp_attack[co] |= moves;
    moves &= ~pcon->pieces[B_OCCU|co];
    int mob = popCount(moves);
    mobDebugADD(pc,mob);
    endgval[co] += mob*2;
    if( true && notpawn[co] <= 3 ) {
        // accessible squares
        bitboard kingSqr = IdxToU64(sqr);
        bitboard myStuff = pcon->pieces[B_OCCU|co];
        // acessible squares : not occupied my my pawns/pieces, not attacked by theit pawns/pieces
        bitboard path = 0xffffffffffffffffULL & ((~myStuff)|kingSqr) & ~attackMap[co ^ 1];
        // look for pawns not defended by pawns
        bitboard pieces = pcon->pieces[PAWN|co^1] & ~phashp->side[co^1].pawnatk;
	    if ( pieces ) {
//		        int pSqr = PopLSB(&pieces);
//                bitboard thisPawn = IdxToU64(pSqr);
//                int dist = pathLength(kingSqr, thisPawn, path);
            int dist = pathLength(kingSqr, pieces, path);
            if( dist < pawnDist[co] )
                pawnDist[co] = dist;
        }
        // blocked position, for at least one side
        if( pawnDist[co] > 50 ) {
            endgval[co] /= 2;
            otherval[co] /= 2;
        } else
            endgval[co] -= (pawnDist[co]-1)*10;
    }

}

//	This eval function evaluates:
//
//	1.	Pawn structure, which is hashed and handled in another module.
//	2.	Piece location, which is usually related to central occupation or
//		in the case of pawns, simple advancement.
//	3.	Minors blocking d2/e2 pawns.
//	4.	King on f1/g1 with rook on g1/h1.
//	5.	Pawn shelter.
//	6.	Endgame king position values versus middlegame values.
//
//	Cases 3 and 4 are surprisingly important.
//
//	The piece-square tables are designed to induce the program to castle
//	behind a safe pawn structure, to move its central pawns, and to develop
//	its pieces.
//
//	There is also a chance that the program would feel an interest in breaking
//	its opponent's king position, but I don't have high hopes for this.

int ValEval(PCON pcon, PSTE pste, int Alpha, int Beta)
{
    int specialVal;
    bool isPawnEndGame = false;
    int side = pste->side;
    int xside = side ^ 1;
    int isqKingUs = pcon->king[side];
    int isqKingThem = pcon->king[xside];

    mobDebugIdx = 0;

	//	Get value for pawn structure.
	//
    EvalPawns(pcon, pste, phashp);
	//
	//	Get material value.
	//
    int val[coMAX];
    // pawn sctructure score not included in eg, structure is added in posval only
//    val[side] = pste->valPcUs + pste->valPnUs + phashp->side[side].val ;
//    val[xside] = pste->valPcThem + pste->valPnThem + phashp->side[xside].val ;
    val[side] = pste->valPcUs + pste->valPnUs;
    val[xside] = pste->valPcThem + pste->valPnThem;
    mat[side] = pste->valPcUs;
    mat[xside] = pste->valPcThem;

    /////
    /////return val[side] - val[xside];
    /////

    //
    //  Check for insuffisant material
    //
    // No pawn
	if ( (pste->valPnUs + pste->valPnThem) == 0 ) {
        // K vs K
		if ( (pste->valPcUs + pste->valPcThem)== 0 )
            return 0;
        // K + Minor vs K
		if ((pste->valPcUs == valBISHOP || pste->valPcUs == valKNIGHT) && (pste->valPcThem == 0))
			return 0;
		if ((pste->valPcUs == 0) && (pste->valPcThem == valBISHOP || pste->valPcThem == valKNIGHT))
			return 0;
        // Two knights can not force checkmate
        if ((pste->valPcUs == 2*valKNIGHT)) {
            if (pste->valPcThem == 0 || pste->valPcThem == valKNIGHT || pste->valPcThem == valBISHOP)
			    return 50;
        }
        if ((pste->valPcThem == 2*valKNIGHT)) {
            if (pste->valPcUs == 0 || pste->valPcUs == valKNIGHT || pste->valPcUs == valBISHOP)
    			return -50;
        }
        // K + Q vs K
		if ((pste->valPcUs == valQUEEN||pste->valPcUs == valROOK) && (pste->valPcThem == 0))
			return pste->valPcUs + 150 - c_argdistborder[ isqKingThem ] -
                sqDistance( isqKingUs, isqKingThem );
        // K vs K + Q | R
		if ((pste->valPcUs == 0) && (pste->valPcThem == valQUEEN||pste->valPcThem == valROOK))
			return -(pste->valPcThem + 150 - c_argdistborder[ isqKingUs ] -
                sqDistance( isqKingUs, isqKingThem ));
        if (pste->valPcUs == (valBISHOP + valKNIGHT)) {
            if( (pste->valPcThem == valBISHOP||pste->valPcThem == valKNIGHT))
                return 50;
            // KBNK
            else if( (pste->valPcThem == 0))
                if( KBNK( pcon, pste, side, specialVal) )
                    return specialVal;
        }
        if (pste->valPcThem == (valBISHOP + valKNIGHT)) {
            if( (pste->valPcUs == valBISHOP||pste->valPcUs == valKNIGHT))
                return -50;
            else if( KBNK( pcon, pste, xside, specialVal) )
                return -specialVal;
        }
    }
    // No piece
    if ((pste->valPcUs + pste->valPcThem)== 0) {
        isPawnEndGame = true;
        // only a pawn on one side
    	if ( (pste->valPnUs == valPAWN) && (pste->valPnThem == 0) ) {
            // compute ranks from pawn to promotion
            // compute file dist from opposite king to pawn
            // if dist to promote < dist from king then win
            if( KPK( pcon, pste, side, specialVal) )
                return specialVal;
        }
    	if ( (pste->valPnUs == 0) && (pste->valPnThem == valPAWN) ) {
            // compute ranks from pawn to promotion
            // compute file dist from opposite king to pawn
            // if dist to promote < dist from king then win
            if( KPK( pcon, pste, xside, specialVal) )
                return -specialVal;
        }
    }
    // bishop + pawns vs nothing
    if ((pste->valPcUs == valBISHOP || pste->valPcUs == 0) && (pste->valPcThem == 0) &&
        (pste->valPnUs > 0) && (pste->valPnThem == 0)) {
        // pawns on file A
        if( (pcon->pieces[PAWN|side] & ~fileAMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, side, promotionSqrFileA[side]) && sqDistance( pcon->king[xside], promotionSqrFileA[side] ) <= 1 )
                return 50;
        // pawns on file H
        if( (pcon->pieces[PAWN|side] & ~fileHMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, side, promotionSqrFileH[side]) && sqDistance( pcon->king[xside], promotionSqrFileH[side] ) <= 1 )
                return 50;
    }
    // bishop + pawns vs nothing
    if ((pste->valPcThem == valBISHOP || pste->valPcThem == 0) && (pste->valPcUs == 0) &&
        (pste->valPnThem > 0) && (pste->valPnUs == 0)) {
        // pawns on file A
        if( (pcon->pieces[PAWN|xside] & ~fileAMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, xside, promotionSqrFileA[xside]) && sqDistance( pcon->king[side], promotionSqrFileA[xside] ) <= 1 )
                return -50;
        // pawns on file H
        if( (pcon->pieces[PAWN|xside] & ~fileHMask) == 0 )
            // if opposite king is in the corner then we can not win
            if( !bishopcanreach(pcon, xside, promotionSqrFileH[xside]) && sqDistance( pcon->king[side], promotionSqrFileH[xside] ) <= 1 )
                return -50;
    }

    //
    //	Cut off if it appears that this is futile.
    //  Don't do this cut off in endgames !!
    //
#if 1
    // caution with score and pruning
    int tmpval = val[side] - val[xside];
    int margin = 900;
    if( !isPawnEndGame )
        margin = 300;
	if (tmpval - margin >= Beta)		// We're already great?
        return (tmpval - 50) ? (tmpval - 50) : 1;
	if (tmpval + margin <= Alpha) 	// We're already toast?
        return (tmpval + 50) ? (tmpval + 50) : -1;
#endif

    attackMap[B] = attackMap[W] = 0;
    posval[B] = posval[W] = 0;      // position value
    tacval[B] = tacval[W] = 0;      // tactical value
    otherval[B] = otherval[W] = 0;  // other value ?
    endgval[B] = endgval[W] = 0;    // end game values
    notpawn[B] = notpawn[W] = 0;
    tempoval[B] = tempoval[W] = 0;
    attackval[B] = attackval[W] = 0;
    c_mobility[B] = c_mobility[W] = 0;
    mobility[B] = mobility[W] = 0;
    defSpace[B] = defSpace[W] = 0;
    pawnDist[B] = pawnDist[W] = 99;

    const bitboard offenseMapSide[coMAX] = {0x00000000ffffffffULL, 0xffffffff00000000ULL};
    posval[W] = phashp->side[W].val;
    posval[B] = phashp->side[B].val;

    // count of pieces by type and color
    for(int i = 0; i < PIECES ; i++)
        c_pieces[i] = 0;
    //
    //
    _dynamicVal dynamicVal[coMAX];
    computeMobility(pcon, pste, B, ~phashp->side[W].pawnatk, dynamicVal[B]);
    computeMobility(pcon, pste, W, ~phashp->side[B].pawnatk, dynamicVal[W]);

    // pawns attacked and not defended by other pawns
    // gives a penality because we don't want to defend with a piece
    // with 10 points penality, the engine wants to attack with his queen
    // in the begining
    attackval[W] += 10*popCount( pcon->pieces[PAWN|B] & dynamicVal[W].attacks & ~phashp->side[B].pawnatk );
    attackval[B] += 10*popCount( pcon->pieces[PAWN|W] & dynamicVal[B].attacks & ~phashp->side[W].pawnatk );

    // hack alert : change the content of the king square
    pcon->pos[pcon->king[B]] = APC_BLACKKING;

    const U8 *pos = pcon->pos;

    {
	    // - Pawns seven-shift -
	    attackMap[W] |= ((pcon->pieces[PAWN|W] & CLEAR_LEFT) << 7);
	    // - Pawns nine-shift -
	    attackMap[W] |= ((pcon->pieces[PAWN|W] & CLEAR_RIGHT) << 9);
	    // - Pawns seven-shift -
	    attackMap[B] |= ((pcon->pieces[PAWN|B] & CLEAR_RIGHT) >> 7);
	    // - Pawns nine-shift -
	    attackMap[B] |= ((pcon->pieces[PAWN|B] & CLEAR_LEFT) >> 9);
    }

#define foreachside2(c,code)    {int co = c; code;}
#define foreachside(code)       foreachside2(B,code) foreachside2(W,code)

    evalPAWN(pcon, pste, B);
    evalPAWN(pcon, pste, W);

    defSpace[B] = popCount(offenseMapSide[W] & ~attackMap[W] &~pcon->pieces[PAWN|B]);
    defSpace[W] = popCount(offenseMapSide[B] & ~attackMap[B] &~pcon->pieces[PAWN|W]);
    posval[B] += defSpace[B];
    posval[W] += defSpace[W];

    knight_moves[B] = knight_moves[W] = 0;
    tmp_attack[B] = tmp_attack[W] = 0;
    underdev[B] = underdev[W] = 0;

    evalKNIGHT(pcon, pste, B);
    evalKNIGHT(pcon, pste, W);

    bishopSquare[B] = bishopSquare[W] = 0;

    evalBISHOP(pcon, pste, B);
    evalBISHOP(pcon, pste, W);

    bool drawish = false;
    // EG with pawns and bishops of opposite color
    if( (pste->valPcThem == valBISHOP) && (pste->valPcUs == valBISHOP) ) {
        if( abs(pste->valPnThem - pste->valPnUs) <= valPAWN * 2 )
            if( !sameSquareColor( bishopSquare[W], bishopSquare[B] ) )
                drawish = true;
    }
    attackMap[W] |= tmp_attack[W];
    attackMap[B] |= tmp_attack[B];
    tmp_attack[W] = tmp_attack[B] = 0;

    evalROOK(pcon, pste, B);
    evalROOK(pcon, pste, W);

    attackMap[W] |= tmp_attack[W];
    attackMap[B] |= tmp_attack[B];
    tmp_attack[W] = tmp_attack[B] = 0;

    evalQUEEN(pcon, pste, B);
    evalQUEEN(pcon, pste, W);

    attackMap[W] |= tmp_attack[W];
    attackMap[B] |= tmp_attack[B];
    tmp_attack[W] = tmp_attack[B] = 0;

    evalKING(pcon, pste, B);
    evalKING(pcon, pste, W);

    attackMap[W] |= tmp_attack[W];
    attackMap[B] |= tmp_attack[B];

    int valAttacking[coMAX] = {0,0};
    int valHanging[coMAX] = {0,0};

#define updateThreat(bm,val)   \
            if( bm & attackMap[co^1] &~phashp->side[co].pawnatk) { \
                valAttacking[co ^ 1] += val; \
                if( (bm & attackMap[co]) == 0 ) \
                    valHanging[co] += val;\
            }

#if 0
    if( true )  { // if not at leaf
        foreachside({
            updateThreat(pcon->pieces[PAWN|co],100);
            updateThreat((pcon->pieces[KNIGHT|co]|pcon->pieces[BISHOP|co]),325);
            updateThreat(pcon->pieces[ROOK|co],500);
            updateThreat(pcon->pieces[QUEEN|co],975);
            valAttacking[co] = 0;
        })
    } 
#endif

    // hack alert : restore the content of the king square
    pcon->pos[pcon->king[B]] = KING|B;

#ifdef KEYSQUARE
    if( isPawnEndGame ) {
        for(int i = 0; i < 3 ; i ++ )
            if( phashp->keySquare[i].square ) {
                int winner = phashp->keySquare[i].winningSide;
                int dist = phashp->keySquare[i].distance[winner^1][ pcon->king[winner^1] ] - 
                    phashp->keySquare[i].distance[winner][ pcon->king[winner] ];
                // side is side to move next
                if( dist >= 2 || (dist >= 1) && (winner == side) ) {
//                if( dist > 3 || (dist > 2) && (winner == side) ) {
                    endgval[winner] += dist * 10;
                }
            }
    }
#endif

    int c_pawn = c_pieces[PAWN|B] + c_pieces[PAWN|W];

	for (int co = coBLACK; co <= coWHITE; co++) {
        // penalize for not developping minors (this is also a tempo counter in the begining)
        posval[co] -= underdev[co] * 20;

        // lower the value of minor pieces in endgame depending on the pawn count
        if( c_pieces[BISHOP|co] + c_pieces[KNIGHT|co] )
            val[co] -= minor_eg_malus[c_pieces[PAWN|co]];

        // compute the mobility and position control
        // but only in midgame
        if( MidGameValueCoef[notpawn[ co ^ 1]] > 0 ) {
            // old coefs move * 5/4, offense * 5/4, king * 5
            // mobility has increased a lot now
            // new coefs move * 3/4, offense * 3/4, king * 5
            // 3 is too big
#if 0
            tacval[co] += (dynamicVal[co].c_move * 2)/4;
#else
            // mob / 2, adjust mobK macro
            tacval[co] += mobility[co];
#endif
            // we must attack the other side so lets give us a bonus for that
            //
            tacval[co] += dynamicVal[co].c_offense * 5/8;
            tacval[co] += dynamicVal[co].c_king * 2;
        }
        if( c_pieces[BISHOP|co] >= 2 ) {
            // bishop bonus if open position
            posval[co] += (6 - phashp->c_center_pawns);
            // bishop pair
            posval[co] += 45;
        }

        // knight bonus for closed positions
        if( c_pieces[KNIGHT|co] >= 1 && c_pieces[BISHOP|(co^1)] == 1 )
            posval[co] += (phashp->c_center_pawns - 6);
    }

    // tempoval[co] += tempoVal[neutralPc >> 1][nSqr];

//    int dvlp_ratio = notpawn[W] > 4 ? 128 * tempoval[W] / (notpawn[W]*4) : 128;
//    if( dvlp_ratio < 85 )
//        attackval[W] = attackval[W] * dvlp_ratio / 128;
//    dvlp_ratio = notpawn[B] > 4 ? 128 * tempoval[B] / (notpawn[B]*4) : 128;
//    if( dvlp_ratio < 85 )
//        attackval[B] = attackval[B] * dvlp_ratio / 128;

    int finalVal[coMAX];
	finalVal[side] = val[side] + MidGameValue( 2*tempoval[side] + posval[side] + tacval[side] + attackval[side],notpawn[xside]) + 
           EndGameValue( endgval[side], notpawn[xside] ) + otherval[side] + valAttacking[side];
	finalVal[xside] = val[xside] + MidGameValue( 2*tempoval[xside] + posval[xside] + tacval[xside] + attackval[xside],notpawn[side]) + 
           EndGameValue( endgval[xside], notpawn[side] ) + otherval[xside] + valAttacking[xside];

    int delta = finalVal[side] - finalVal[xside];
#if 0
    int sum = 400 + (((finalVal[side] + finalVal[xside]) > 0) ? (finalVal[side] + finalVal[xside]) : - (finalVal[side] + finalVal[xside]));
    int ratioBonus = delta * 128 / sum;
#else
    // pieces start:14=>0, mid:10=>4, eg:4=>10, eg:2=>12, eg:0=>14
    // pawns  start:16=>0, mid:12=>4, eg:8=>8, eq:4=>12
//    int exchange = 4*(14-(notpawn[W]+notpawn[B]))-8*(16-(c_pieces[PAWN|B] + c_pieces[PAWN|W]));
    int exchange = 4*(10-(notpawn[W]+notpawn[B]))-8*(10-(c_pieces[PAWN|B] + c_pieces[PAWN|W]));
//    int ratioBonus = exchange * delta / 256;
    int ratioBonus = exchange * (val[side]-val[xside]) / 256;
#endif

#ifdef _DEBUG
    if( Alpha == -20000 ) {
        printf("# val=%d ratioBonus=%d => %d (hkey=%I64x) centerpawn#=%d\n", delta, ratioBonus, 
            delta + ratioBonus, pste->hashkPc, phashp->c_center_pawns );
    	for (int co = coBLACK; co <= coWHITE; co++) {
            int dvlp_ratio = notpawn[co] > 4 ? 100 * tempoval[co] / (notpawn[co]*4) : 100;
            printf("# (%c) val=%d finalval=%d posval=%d tacval=%d endgval=%d other=%d tempo=%d\n",
                "BW"[co], val[co], finalVal[co], posval[co], tacval[co], endgval[co], otherval[co], tempoval[co]);
            printf("# dynamics  dvlp=%d attack=%d move=%d offense=%d king=%d c_mob=%d mobility=%d\n", 
                dvlp_ratio, attackval[co], dynamicVal[co].c_move, dynamicVal[co].c_offense, 
                dynamicVal[co].c_king, c_mobility[co], mobility[co]);
            printf("# others : space in side=%d\n", defSpace[co]);
            printf("# pawns : val=%d  passed#=%d candidate#=%d backwards=%d connected=%d\n", phashp->side[co].val, 
                popCount(phashp->side[co].passed), popCount(phashp->side[co].candidate), 
                phashp->side[co].c_backward, phashp->side[co].c_connected );
#ifdef _DEBUG
            printf("#  mobility for pieces\n# ");
            for(int i = 0; i < mobDebugIdx; i ++)
                if( (mobDebugTable[i].piece & 1) == co ) 
                    printf("%c (%d)  ", stringPc[mobDebugTable[i].piece], mobDebugTable[i].mobility);
            printf("\n");
#endif
        }
    }
#endif
    if( drawish )
        return (delta + ratioBonus) / 2;
    else
        return (delta ? delta : -2 ) + ratioBonus;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

/*
//  a   b   c   d   e   f   g   h
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,
	//
	//	White rook.
	//
//  a   b   c   d   e   f   g   h
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	4,  4,  4,  4,  4,  4,  4,  4,
	//

*/

U64 kposit[8];
int pieceToPatBit[16];

static INLINE unsigned int squarePattern(const CON *pcon, int n, int f, int r, int sq) {
        kposit[n] = IdxToU64(sq);
        if(f < 0 || f > 7 || r < 0 || r > 7)
                return 1;
        else if(pcon->pieces[B_OCCU] & IdxToU64(sq))
                return 1;
        return 0;
}
static INLINE unsigned int squarePattern2(const CON *pcon, int n, int sq) {
        kposit[n] = IdxToU64(sq);
//      if(pieces[B_OCCU] & IdxToU64(sq))
//              return 1;
//      return 0;
        return pieceToPatBit[ pcon->pos[sq] ];
}
static INLINE unsigned int squarePattern3(const CON *pcon, int n, int sq) {
        kposit[n] = ((1LL)<<sq);
//      if(pieces[B_OCCU] & (C64(1)<<sq))
//              return 1;
//      return 0;
        // checking for white pieces is ok
//      if(pos[sq] & 1)
//              return 1;
        // checking for black pieces is not ok
        return pieceToPatBit[ pcon->pos[sq] ];
//      if(pos[sq] && !(pos[sq] & 1))
//              return 1;
//      return 0;
}

unsigned int checkPattern(const CON *pcon) {
        int kpos = pcon->king[B];
        unsigned int pattern = 0;
        int f = File(kpos);
        int r = Rank(kpos);
//      pattern = (pattern << 1) | squarePattern(f,r,kpos);
        pattern = (pattern << 1) | squarePattern(pcon, 0, f-1,r,kpos-1);
        pattern = (pattern << 1) | squarePattern(pcon, 1, f-1,r+1,kpos+7);
        pattern = (pattern << 1) | squarePattern(pcon, 2, f,r+1,kpos+8);
        pattern = (pattern << 1) | squarePattern(pcon, 3, f+1,r+1,kpos+9);
        pattern = (pattern << 1) | squarePattern(pcon, 4, f+1,r,kpos+1);
        pattern = (pattern << 1) | squarePattern(pcon, 5, f-1,r-1,kpos-9);
        pattern = (pattern << 1) | squarePattern(pcon, 6, f,r-1,kpos-8);
        pattern = (pattern << 1) | squarePattern(pcon, 7, f+1,r-1,kpos-7);
        return pattern;
}

unsigned int checkPattern2(const CON *pcon) {
        int kpos = pcon->king[B];
        unsigned int pattern = 0;
        int f = File(kpos);
        int r = Rank(kpos);
        if( f == 0 || f == 7 || r == 0 || r == 7 )
                return checkPattern(pcon);

        pattern = (pattern << 1) | squarePattern2(pcon, 0, kpos-1);
        pattern = (pattern << 1) | squarePattern2(pcon, 1, kpos+7);
        pattern = (pattern << 1) | squarePattern2(pcon, 2, kpos+8);
        pattern = (pattern << 1) | squarePattern2(pcon, 3, kpos+9);
        pattern = (pattern << 1) | squarePattern2(pcon, 4, kpos+1);
        pattern = (pattern << 1) | squarePattern2(pcon, 5, kpos-9);
        pattern = (pattern << 1) | squarePattern2(pcon, 6, kpos-8);
        pattern = (pattern << 1) | squarePattern2(pcon, 7, kpos-7);
        return pattern;
}

unsigned int checkPattern3(const CON *pcon) {
        int kpos = pcon->king[B];
        unsigned int pattern = 0;
        int f = File(kpos);
        int r = Rank(kpos);

        if( kpos == G8 ) {
                pattern = (pattern << 1) | squarePattern3(pcon, 0, F8);
                pattern = (pattern << 1) | 1;
                pattern = (pattern << 1) | 1;
                pattern = (pattern << 1) | 1;
                pattern = (pattern << 1) | squarePattern3(pcon, 4, H8);
                pattern = (pattern << 1) | squarePattern3(pcon, 5, F7);
                pattern = (pattern << 1) | squarePattern3(pcon, 6, G7);
                pattern = (pattern << 1) | squarePattern3(pcon, 7, H7);
        }
        return pattern;

}

struct bitPat {
        unsigned char bishop_attack : 1;
        unsigned char queen_attack : 1;
        unsigned char pawn_attack : 1;
        unsigned char knight_attack : 1;
        unsigned char square_defended : 1;
} ;

typedef struct {
        struct bitPat   atkdef[8];
        unsigned char   score;
        unsigned char   atk_count;
        unsigned char   def_count;
} patEntry;

patEntry        chunks[256];

int checkKingAtk(const CON *pcon) {
        int idx = checkPattern3(pcon);
        patEntry        pat = chunks[idx];
        U64 bishop_attack = rand();
        U64 queen_attack = rand();
        U64 knight_attack = rand();
        U64 all_defense = rand();
        int atk_nok = 0;
        int def_nok = 0;
        for(int i = 0; i < 8 ; i++) {
                U64 sqMask = kposit[i];
                if( pat.atkdef[i].bishop_attack )
                        if( (bishop_attack & sqMask) == 0 )
                                atk_nok++;
                if( pat.atkdef[i].queen_attack )
                        if( (queen_attack & sqMask) == 0 )
                                atk_nok++;
                if( pat.atkdef[i].knight_attack )
                        if( (knight_attack & sqMask) == 0 )
                                atk_nok++;
                if( pat.atkdef[i].square_defended == 0 )
                        if( (all_defense & sqMask) != 0 )
                                def_nok++;
        }

        int atk_ok = (pat.atk_count - atk_nok);
        int def_ok = 8 - def_nok;
//      return (atk_ok / pat.atk_count) * (def_ok) / 8;
        return (32*atk_ok*def_ok) / pat.atk_count;
} 


// From Gnu Chess 4.x
/*
 * nextpos[piece][from-square] , nextdir[piece][from-square] gives vector of
 * positions reachable from from-square in ppos with piece such that the
 * sequence	ppos = nextpos[piece][from-square]; pdir =
 * nextdir[piece][from-square]; u = ppos[sq]; do { u = ppos[u]; if(color[u]
 * != neutral) u = pdir[u]; } while (sq != u); will generate the sequence of
 * all squares reachable from sq.
 *
 * If the path is blocked u = pdir[sq] will generate the continuation of the
 * sequence in other directions.
 */

/*
 * ptype is used to separate white and black pawns, like this; ptyp =
 * ptype[side][piece] piece can be used directly in nextpos/nextdir when
 * generating moves for pieces that are not black pawns.
 */
// Base piece/move types
//const int KING		= 0;
//const int QUEEN		= 2;
//const int ROOK		= 4;
//const int BISHOP	= 6;
//const int KNIGHT	= 8;
//const int PAWN		= 10;

const int ptype[16] =
{ APC_king, APC_king, APC_queen, APC_queen, APC_rook, APC_rook, APC_bishop, APC_bishop, 
    APC_knight, APC_knight, APC_bpawn, APC_pawn,   
    APC_no_piece, APC_no_piece, APC_no_piece, APC_no_piece};

/* data used to generate nextpos/nextdir */
static const int direc[8][8] =
{
    { 0, 0, 0, 0, 0, 0, 0, 0 },             // no piece
    { 10, 9, 11, 0, 0, 0, 0, 0 },           // wpawn
    { 8, -8, 12, -12, 19, -19, 21, -21 },   // knight
    { 9, 11, -9, -11, 0, 0, 0, 0 },         // bishop
    { 1, 10, -1, -10, 0, 0, 0, 0 },         // rook
    { 1, 10, -1, -10, 9, 11, -9, -11 },     // queen
    { 1, 10, -1, -10, 9, 11, -9, -11 },     // king
    { -10, -9, -11, 0, 0, 0, 0, 0 } };      // bpawn
static const int max_steps[8] =
{0, 2, 1, 7, 7, 7, 1, 2};
static const int nunmap[120] =
{
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, 0, 1, 2, 3, 4, 5, 6, 7, -1,
  -1, 8, 9, 10, 11, 12, 13, 14, 15, -1,
  -1, 16, 17, 18, 19, 20, 21, 22, 23, -1,
  -1, 24, 25, 26, 27, 28, 29, 30, 31, -1,
  -1, 32, 33, 34, 35, 36, 37, 38, 39, -1,
  -1, 40, 41, 42, 43, 44, 45, 46, 47, -1,
  -1, 48, 49, 50, 51, 52, 53, 54, 55, -1,
  -1, 56, 57, 58, 59, 60, 61, 62, 63, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1};

int InitFlag = false;
static void Initialize_moves (void) {

/*
 * This procedure pre-calculates all moves for every piece from every square.
 * This data is stored in nextpos/nextdir and used later in the move
 * generation routines.
 */

    for (int ptyp = 0; ptyp < 8; ptyp++)
        for (int po = 0; po < 64; po++)
            for (int pZero = 0; pZero < 64; pZero++) {
                PosDirTab[ptyp][po][pZero].dir = (U8) po;
                PosDirTab[ptyp][po][pZero].pos = (U8) po;
//                nextpos[ptyp][po][pZero] = (U8) po;
//                nextdir[ptyp][po][pZero] = (U8) po;
            }

  for (int ptyp = 1; ptyp < 8; ptyp++)
    for (int po = 21; po < 99; po++)
      if (nunmap[po] >= 0) {
        int dest[8][8];
        int steps[8];
        int sorted[8];
	    /* dest is a function of direction and steps */
	    for (int d = 0; d < 8; d++) {
	        dest[d][0] = nunmap[po];
	        int delta = direc[ptyp][d];
            int s = 0;
	        if (delta != 0) {
		        int pZero = po;
		        for (s = 0; s < max_steps[ptyp]; s++) {
		            pZero = pZero + delta;
		            /*
		            * break if (off board) or (pawns only move two
		            * steps from home square)
		            */
		            if ( nunmap[pZero] < 0 )
			            break;
		            if ( ((ptyp == APC_pawn) || (ptyp == APC_bpawn)) && (s > 0))
			            break;
		            else
			            dest[d][s] = nunmap[pZero];
		        }
		    }
            /*
            * sort dest in number of steps order currently no sort
            * is done due to compability with the move generation
            * order in old gnu chess
            */
            steps[d] = s;
            for (int di = d; s > 0 && di > 0; di--)
                if (steps[sorted[di - 1]] == 0)	/* should be: < s */
                    sorted[di] = sorted[di - 1];
                else
                    break;
            sorted[di] = d;
	    }
        posdir *nextPosDir = PosDirTab[ptyp][nunmap[po]];
        /*
        * update nextpos/nextdir, pawns have two threads (capture
        * and no capture)
        */
        int pZero = nunmap[po];
        if (ptyp == APC_pawn || ptyp == APC_bpawn) {
            for (int s = 0; s < steps[0]; s++) {
                nextPosDir[pZero].pos = (U8) dest[0][s];
                pZero = dest[0][s];
            }
            pZero = nunmap[po];
            for (int d = 1; d < 3; d++) {
                nextPosDir[pZero].dir = (U8) dest[d][0];
                pZero = dest[d][0];
            }
        } else {
            nextPosDir[pZero].dir = (U8) dest[sorted[0]][0];
            for (int d = 0; d < 8; d++)
                for (int s = 0; s < steps[sorted[d]]; s++) {
                    nextPosDir[pZero].pos = (U8) dest[sorted[d]][s];
                    pZero = dest[sorted[d]][s];
                    if (d < 7)
                        nextPosDir[pZero].dir = (U8) dest[sorted[d + 1]][0];
                    /*
                    * else is already initialized
                    */
                }
        }
    }
}

#if 0
void ataks (int side, short *a) {

/*
 * Fill array atak[][] with info about ataks to a square.  Bits 8-15 are set
 * if the piece (king..pawn) ataks the square.  Bits 0-7 contain a count of
 * total ataks to the square.
 */
    short u, c;
    short piece, *PL;

    PL = PieceList[side];
    for (int i = PieceCnt[side]; i >= 0; i--) {
        int sq = PL[i];
        piece = board[sq];
        c = control[piece];
        if (sweep[piece]) {
            U8 *ppos, *pdir;
            ppos = nextpos[piece][sq];
            pdir = nextdir[piece][sq];
            u = ppos[sq];
            do {
                a[u] = ((a[u]+1) | c);
                u = ((color[u] == neutral) ? ppos[u] : pdir[u]);
            } while (u != sq);
        } else {
            U8 *pdir;
            pdir = nextdir[ptype[side][piece]][sq];
            u = pdir[sq];		/* follow captures thread for pawns */
            do {
                a[u] = ((a[u]+1) | c);
                u = pdir[u];
            } while (u != sq);
        }
    }
}
#endif

