// Bitboard Perft by Jacob Hales (using Pradyumna Kannan's magic move generator)
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h> 
#include <limits.h>

#include "magicmoves.h"
#include "engine.hpp"
#include "genmagic.hpp"
#include "hash.hpp"
#include "moves.hpp"
#include "eval.hpp"
#include "deprec.hpp"

// Define 64-bit integer
#ifndef __64_BIT_INTEGER_DEFINED__
   #define __64_BIT_INTEGER_DEFINED__
   #if defined(_MSC_VER) && _MSC_VER<1300
      typedef unsigned __int64 U64; //For the old microsoft compilers
   #else
      typedef unsigned long long  U64; //Supported by MSC 13.00+ and GCC
   #endif //defined(_MSC_VER) && _MSC_VER<1300
#endif //__64_BIT_INTEGER_DEFINED__


#ifdef	DEBUG
static char const s_aszModule[] = __FILE__;
#endif


const char pc_char_map[12] = {
	'k', 'K', 'q', 'Q', 'r', 'R', 'b', 
	'B', 'n', 'N', 'p', 'P'
};

// http://supertech.csail.mit.edu/papers/debruijn.pdf
const U64 debruijn64 =	0x07EDD5E59A4E28C2ULL;

int index64[64] = {
	63,  0, 58,  1, 59, 47, 53,  2,
	60, 39, 48, 27, 54, 33, 42,  3,
	61, 51, 37, 40, 49, 18, 28, 20,
	55, 30, 34, 11, 43, 14, 22,  4,
	62, 57, 46, 52, 38, 26, 32, 41,
	50, 36, 17, 19, 29, 10, 13, 21,
	56, 45, 25, 31, 35, 16,  9, 12,
	44, 24, 15,  8, 23,  7,  6,  5
};

const unsigned char castle_spoiler[64] = { 
	 13, 15, 15, 15, 15, 15, 15, 14,
	 15, 15, 15, 15, 15, 15, 15, 15,
	 15, 15, 15, 15, 15, 15, 15, 15,
	 15, 15, 15, 15, 15, 15, 15, 15,
	 15, 15, 15, 15, 15, 15, 15, 15,
	 15, 15, 15, 15, 15, 15, 15, 15,
	 15, 15, 15, 15, 15, 15, 15, 15,
	  7, 15, 15, 15, 15, 15, 15, 11
};

enum RANKS	{ 
	RANK_1,		RANK_2,		RANK_3,		RANK_4,
	RANK_5,		RANK_6,		RANK_7,		RANK_8 
};

enum FILES	{ 
	FILE_A,		FILE_B,		FILE_C,		FILE_D, 
	FILE_E,		FILE_F,		FILE_G,		FILE_H 
};

//const U64 CLEAR_LEFT	= 0xFEFEFEFEFEFEFEFEULL;
//const U64 CLEAR_RIGHT	= 0x7F7F7F7F7F7F7F7FULL;

// Castling masks
const unsigned char castle_mask[2]		=	{ 12, 3 };
const unsigned char perm_mask_00[2]		=	{  4, 1 };
const unsigned char perm_mask_000[2]	=	{  8, 2 };

const U64 mask_00[2]	= { 0x6000000000000000ULL, 0x0000000000000060ULL };
const U64 mask_000[2]	= { 0x0E00000000000000ULL, 0x000000000000000EULL };

// Pawn move generation data
const int xoccu_wp[2]	= { X_OCCU, WP };
const int xoccu_bp[2]	= { X_OCCU, BP };
const int xoccu_tmp[2]	= { X_OCCU, TEMP };
const int woccu_wp[2]	= { W_OCCU, WP };
const int boccu_bp[2]	= { B_OCCU, BP };
const U64 mask_5_3[2]	= { 0x000000FF00000000ULL, 0x0000000000FF0000ULL};

const int p_forw[2]			= { -8, 8 };
const int inter_shift[2]	= {  8, 0 };
const int inter_shift2[2]	= { 16, 0 };

const U64 prom_mask[2]		= { 0x000000000000FF00ULL, 0xFF00000000000000ULL };
const U64 xprom_mask[2]		= { 0xFFFFFFFFFFFF00FFULL, 0x00FFFFFFFFFFFFFFULL };

const U64 ev_prom_mask[2]	= { 0x000000000000FF00ULL, 0x00FF000000000000ULL };
const U64 ev_xprom_mask[2]	= { 0xFFFFFFFFFFFF00FFULL, 0xFF00FFFFFFFFFFFFULL };





// Macros to simplify adding moves to a list (array of encoded unsigned integers)
// from {0-5}, to {6-11}, piece {12-15}, capt {16-19}, prom {20-21}
#define AddMove(from, to, pc) \
    list++->m = ((from) | ((to) << 6) | ((pc) << 12))

#define AddCapt(from, to, type, capt) \
    list++->m = ((from) | ((to) << 6) | ((type) << 12) | ((capt) << 16))

#define AddProm(from, to, type, capt) \
	list++->m = ((from) | ((to) << 6) | ((type) << 12) | ((capt) << 16) | (PROM_Q << 20))

#define AddUnderProm(from, to, type, capt) \
	list++->m = ((from) | ((to) << 6) | ((type) << 12) | ((capt) << 16) | (PROM_N << 20)); \
	list++->m = ((from) | ((to) << 6) | ((type) << 12) | ((capt) << 16) | (PROM_R << 20)); \
	list++->m = ((from) | ((to) << 6) | ((type) << 12) | ((capt) << 16) | (PROM_B << 20))



// Engine data
//
U64 king_attacks[64];
U64 knight_attacks[64];
U64 pawn_attacks[2][64];
U64 squares_between[64][64];

U64 king_base[64];
U64 _IdxToU64[64];

U64 nodes;

// Function declarations
//
unsigned int PopLSB(U64 * val);
unsigned int LSB(const U64 val);

void SetFromFEN(PCON pcon, PSTE pste, char* fen);
//void AddPiece(int idx, int piece);

U64 AttacksTo(const CON *pcon, int sqr);
static void Perft(PCON pcon, PSTE pste, int depth);




/* de Bruijn */ 

const int       lsz64_tbl[64] = 
{ 
    0, 31,  4, 33, 60, 15, 12, 34, 
   61, 25, 51, 10, 56, 20, 22, 35, 
   62, 30,  3, 54, 52, 24, 42, 19, 
   57, 29,  2, 44, 47, 28,  1, 36, 
   63, 32, 59,  5,  6, 50, 55,  7, 
   16, 53, 13, 41,  8, 43, 46, 17, 
   26, 58, 49, 14, 11, 40,  9, 45, 
   21, 48, 39, 23, 18, 38, 37, 27, 
}; 


/*popCount() 
 *a noniterative population count of 1 bits in a quadword 
 * 
 *@param b - the quadword to be counted 
 *@returns the number of 1 bits in b 
 */ 
#define m1 0x5555555555555555ULL 
#define m2 0x3333333333333333ULL 
unsigned popCount1(bitboard b)
{ 
    unsigned        n; 
    const bitboard  a = b - ((b >> 1) & m1); 
    const bitboard  c = (a & m2) + ((a >> 2) & m2); 
    n = (unsigned) ((c & 0xffffffff) + (c >> 32)); 
    n = (n & 0x0F0F0F0F) + ((n >> 4) & 0x0F0F0F0F); 
    n = (n & 0xFFFF) + (n >> 16); 
    n = (n & 0xFF) + (n >> 8); 
    return n; 
} 

/* BitBoard population count   */ 
/* (R. Scharnagl, 2005/Jul/20) */ 

#define msk1 0xEEEEEEEEUL 
#define msk2 0xCCCCCCCCUL 
#define msk3 0x88888888UL 
#define msk4 0x0F0F0F0FUL 

unsigned popCount2(const bitboard b)
{ 
  unsigned buf; 
  unsigned acc; 

  buf  = (unsigned)b; 
  acc  = buf; 
  acc -= ((buf &= msk1)>>1); 
  acc -= ((buf &= msk2)>>2); 
  acc -= ((buf &= msk3)>>3); 
  buf  = (unsigned)(b>>32); 
  acc += buf; 
  acc -= ((buf &= msk1)>>1); 
  acc -= ((buf &= msk2)>>2); 
  acc -= ((buf &= msk3)>>3); 
  acc = (acc & msk4)   + ((acc >> 4) & msk4); 
  acc = (acc & 0xFFFF) + (acc >> 16); 
  acc = (acc & 0xFF)   + (acc >> 8); 
  return acc; 
} 

#define USE_32_BIT_MULTIPLICATIONS

int transform(bitboard b, bitboard magic, int bits) {
#if defined(USE_32_BIT_MULTIPLICATIONS)
  return
    (unsigned)((int)b*(int)magic ^ (int)(b>>32)*(int)(magic>>32)) >> (32-bits);
#else
  return (int)((b * magic) >> (64 - bits));
#endif
}

unsigned horizontal_attack_index(bitboard occ) { 
   unsigned u = (unsigned)occ | (unsigned)(occ >> 32) << 0; 
   return u * 0x02020202 >> 26; 
} 

unsigned vertical_attack_index(bitboard occ) { 
   unsigned u = (unsigned)occ | (unsigned)(occ >> 32) << 3; 
   return u * 0x01041041 >> 26; 
} 


void AddPiece(PCON pcon, int idx, int piece)
{
	int col = piece & 1;
	if ((piece & 14) == KING) {
		pcon->king[col] = U8(idx);
	}

	pcon->pos[idx] = U8(piece);
	pcon->pieces[piece] |= IdxToU64(idx);
	pcon->pieces[B_OCCU|col] |= IdxToU64(idx);
	pcon->pieces[OCCU] = pcon->pieces[B_OCCU] | pcon->pieces[W_OCCU];
	pcon->pieces[X_OCCU] = ~pcon->pieces[OCCU];
}

void bGenMoves(PCON pcon, PSTE pste) {
	if (Attacked(pcon, pcon->king[pste->side], pste->xside)) {
		bGenEvasions(pcon, pste, pste->mlist_start);
	}
	else {
		CM * p_moves = bGenCaptures(pcon, pste, pste->mlist_start);
		bGenNonCaptures(pcon, pste, p_moves);
	}
}

static void Perft(PCON pcon, PSTE pste, int depth)
{
	if (depth == 0) { nodes++; return; }

    bGenMoves(pcon, pste);

    foreachmove(pste,pcm) {
		bMove(pcon, pste, pcm->m );
		Perft(pcon, pste + 1, depth - 1);
		bUndoMove(pcon, pste, pcm->m);
	}
}



void Print(const CON *pcon, const STE *pste)
{
	int i, j;

    printf("# +---------------+\n");

	for (i = 7; i >= 0; i--) {	
        printf("# %d|", i + 1);

		for (j = 0; j < 8; j++) {

			if (!IsSet(Idx(i, j), pcon->pieces[OCCU])) 
				printf(".");
			else 
				printf("%c", pc_char_map[pcon->pos[Idx(i, j)]]);
			if (j < 7) printf(" ");
		}

		printf("|\n");
	}

    printf("# +---------------+\n");
    printf("#  a b c d e f g h\n");
    printf("# side=%d\n", pste->side );
    fflush(stdout);
}
/*
static void checkKing(const CON *pcon) {
    if( popCount(pcon->pieces[KING]) != 1 || popCount(pcon->pieces[KING|1]) != 1 ||
        (pcon->pieces[KING] & pcon->pieces[PAWN]) !=0) {
        Print(pcon, pcon->argste );
        Assert(0);
    }
}

static void checkPos(const CON *pcon, bool init) {
static U8 pos_save[64];
    if(init) {
        for(int i = 0; i < 64 ; i++ )
            pos_save[i] = pcon->pos[i];
    } else {
        for(int i = 0; i < 64 ; i++ )
            if( pos_save[i] != pcon->pos[i] ) {
                Assert(0);
            }

    }
}
*/
void InitBitboards(void)
{
	int i, j, k, file, rank;
	U64 temp;
    assert(UINT_MAX >= 0xffffffff); 

	initmagicmoves();

	// king/knight movement offsets
	int knight_off[] = { 15, 17, 10, 6, -6, -10, -17, -15 };
	int king_off[] = { 1, -1, 8, -8, 7, -7, 9, -9 };

	memset(squares_between, 0, sizeof(squares_between));

	for (i = 0; i < 64; i++) {
        // a table access is faster than a call to __allshl(1,idx)
        _IdxToU64[i] = (1ULL << (i));

		// Initialize king attacks
		for (j = 0; j < 8; j++) {
			int king_idx = i + king_off[j];
			int change = abs(File(i) - File(king_idx))
				+ abs(Rank(i) - Rank(king_idx));

			if (king_idx < 0 || king_idx > 63 || change > 2) {
				continue;
			}

			Set(king_idx, king_attacks[i]);
		}
        int kr = Rank(i);
        int kf = File(i);
        for(int r = -2; r <= 2; r++)
            for(int f = -2; f <= 2; f++) {
                int br = kr + r;
                int bf = kf + f;
                if( br >= 0 && br <= 7 && bf >= 0 && bf <= 7 )
                    Set( Idx(br,bf), king_base[i]);
            }

		// Initialize knight attacks
		for (j = 0; j < 8; j++) {
			int knight_idx = i + knight_off[j];
			int change = abs(File(i) - File(knight_idx))
				+ abs(Rank(i) - Rank(knight_idx));

			if (knight_idx < 0 || knight_idx > 63 || change != 3) {
				continue;
			}

			Set(knight_idx, knight_attacks[i]);
		}

		// Initialize squares_between
		for (j = 0; j < 64; j++) {
			int off = 0;
			int r_c = Rank(i) - Rank(j);
			int f_c = File(i) - File(j);

			if (j == i) continue;

			// Set an offset based on rank & file change
			if (f_c == 0) {
				off = (r_c < 0 ? 8 : -8);
			}
			else if (r_c == 0) {
				off = (f_c < 0 ? 1 : -1);
			}
			else if (abs(r_c) == abs(f_c)) {
				off =	(r_c < 0 ? 8 : -8);
				off +=	(f_c < 0 ? 1 : -1);				
			}
			else continue;

			// Set bits along the offset
			for (k = 1; ((k * off) + i) != j; k++) {
				Set((k * off + i), squares_between[i][j]);
			}
		}

		// Initialize the pawn attacks
		file = File(i), rank = Rank(i);
		temp = 0;

		if ((file - 1) >= 0) Set((file - 1), temp);
		if ((file + 1) < 8) Set((file + 1), temp);

		pawn_attacks[W][i] = temp << ((rank + 1) * 8);
		pawn_attacks[B][i] = temp << ((rank - 1) * 8);
	}
}



void SetFromFEN(PCON pcon, PSTE pste, char* fen)
{
	int i, j, idx = 0;

	// Reset the board

	PSTE status = pste;
	status[0].castle	= 0;
	status[0].ep_t		= 0;
    status[0].checkf    = 0;

	memset(pcon->pos, 0, sizeof(pcon->pos));
	memset(pcon->pieces, 0, sizeof(pcon->pieces));
	pste->hashkPc = 0;		// Going to XOR the all pieces/pawns into this.
	pste->hashkPn = 0;		// Going to XOR the pawns only into this.

	// Parse board description
	for (i = 7; i >=0; i--) {
		for (j = 0; j <= 7; j++) {
			if (isdigit(fen[idx])) {
				j += atoi(&fen[idx]) - 1;
			}
			if (isalpha(fen[idx])) {
				int col = B;

				if (isupper(fen[idx])) {
					col = W;
				}
				switch (tolower(fen[idx])) {
					case 'p':
			            rehashBB( pste->hashkPc, BP|col, Idx(i, j));
			            rehashBB( pste->hashkPn, BP|col, Idx(i, j));
						AddPiece(pcon, Idx(i, j), PAWN|col);
						break;
					case 'n':
			            rehashBB( pste->hashkPc, BN|col, Idx(i, j));
						AddPiece(pcon, Idx(i, j), KNIGHT|col);
						break;
					case 'b':
			            rehashBB( pste->hashkPc, BB|col, Idx(i, j));
						AddPiece(pcon, Idx(i, j), BISHOP|col);
						break;
					case 'r':
			            rehashBB( pste->hashkPc, BR|col, Idx(i, j));
						AddPiece(pcon, Idx(i, j), ROOK|col);
						break;
					case 'q':
			            rehashBB( pste->hashkPc, BQ|col, Idx(i, j));
						AddPiece(pcon, Idx(i, j), QUEEN|col);
						break;
					case 'k':
			            rehashBB( pste->hashkPc, BK|col, Idx(i, j));
						AddPiece(pcon, Idx(i, j), KING|col);
						break;
				}

			}
			idx++;
		}
		idx++;
	}

	// Determine side to move
	if (fen[idx++] == 'w') 
		pste->side = W, pste->xside = B;
    else {
		pste->side = B, pste->xside = W;
        pste->hashkPc = HashkSwitch(pste->hashkPc);
    }


	// Parse castle field
	if (fen[++idx] == '-') {
		idx++;
	}
	else {
		while (isalpha(fen[idx])) {
			switch (fen[idx++]) {
				case 'K':
					status[0].castle |= 1; 
					break;
				case 'k': 
					status[0].castle |= 4; 
					break;
				case 'Q': 
					status[0].castle |= 2; 
					break;
				case 'q': 
					status[0].castle |= 8; 
					break;
			}
		}
	}

	// Parse En Passant target
	if (fen[++idx] != '-') {
		int file = fen[idx] - 'a';
		int rank = fen[idx + 1] - '1';
		pste->ep_t = U8(Idx(rank, file));
	}

	// Initialize pin/check data
	PinCheckUpdate(pcon, pste, pste->side);

	// Contact checks
	if (pawn_attacks[pste->side][pcon->king[pste->side]] & pcon->pieces[PAWN|pste->xside]) {
		pste->checkf = 1;
	}
	if (knight_attacks[pcon->king[pste->side]] & pcon->pieces[KNIGHT|pste->xside]) {
		pste->checkf = 1;
	}
}

void convMailboxtoBB(PCON pcon, PSTE pste) {
	// Reset the board

//	pste->castle	= 0;
    pste->checkf    = 0;
#define RnkFromIsq(isq)				((isq) / filIBD)
#define FilFromIsq(isq)				((isq) & (filIBD - 1))
	pste->ep_t		= U8(Idx(RnkFromIsq(pste->isqEnP), FilFromIsq(pste->isqEnP)));

	memset(pcon->pos, 0, sizeof(pcon->pos));
	memset(pcon->pieces, 0, sizeof(pcon->pieces));
	pste->hashkPc = 0;		// Going to XOR the all pieces/pawns into this.
	pste->hashkPn = 0;		// Going to XOR the pawns only into this.

    for(int co = B; co <= W ; co++) {
        int col = co;
        for (int ipi = 0; ipi < pcon->argcpi[co]; ipi++) {

		    PPI ppiFrom = &pcon->argpi[co][ipi];
		    if (ppiFrom->fDead)	// Ignore pieces that don't exist anymore.
			    continue;
		    int isq = ppiFrom->isq;
            int rnk = RnkFromIsq( isq );
            int fil = FilFromIsq( isq );
		    switch (ppiFrom->pc) {
			    case pcPAWN:
				    AddPiece(pcon, Idx(rnk, fil), PAWN|col);
			        rehashBB( pste->hashkPc, BP|col, Idx(rnk, fil));
			        rehashBB( pste->hashkPn, BP|col, Idx(rnk, fil));
				    break;
			    case pcKNIGHT:
		            rehashBB( pste->hashkPc, BN|col, Idx(rnk, fil));
				    AddPiece(pcon, Idx(rnk, fil), KNIGHT|col);
				    break;
			    case pcBISHOP:
		            rehashBB( pste->hashkPc, BB|col, Idx(rnk, fil));
				    AddPiece(pcon, Idx(rnk, fil), BISHOP|col);
				    break;
			    case pcROOK:
		            rehashBB( pste->hashkPc, BR|col, Idx(rnk, fil));
				    AddPiece(pcon, Idx(rnk, fil), ROOK|col);
				    break;
			    case pcQUEEN:
		            rehashBB( pste->hashkPc, BQ|col, Idx(rnk, fil));
				    AddPiece(pcon, Idx(rnk, fil), QUEEN|col);
				    break;
			    case pcKING:
		            rehashBB( pste->hashkPc, BK|col, Idx(rnk, fil));
				    AddPiece(pcon, Idx(rnk, fil), KING|col);
				    break;
		    }
	    }
    }
	// side to move
//    pste->side = pste->side == coWHITE ? W : B;
    pste->xside = pste->side ^ 1;

    pste->ep_t = pste->isqEnP == isqNIL ? 0 : U8(Idx(RnkFromIsq(pste->isqEnP), FilFromIsq(pste->isqEnP)));

	// Initialize pin/check data
	PinCheckUpdate(pcon, pste, pste->side);

	// Contact checks
//	if (pawn_attacks[pste->xside][pcon->king[pste->side]]) {
	if (pawn_attacks[pste->side][pcon->king[pste->side]] & pcon->pieces[PAWN|pste->xside]) {
		pste->checkf = 1;
	}
	if (knight_attacks[pcon->king[pste->side]] & pcon->pieces[KNIGHT|pste->xside]) {
		pste->checkf = 1;
	}
#undef RnkFromIsq
#undef FilFromIsq

}


// Based on Dr. Hyatt's algorithm (Crafty)
U64 AttacksTo(const CON *pcon, int sqr)
{
	return (king_attacks[sqr] & (pcon->pieces[BK] | pcon->pieces[WK]))
		  | (pawn_attacks[B][sqr] & pcon->pieces[WP])
		  | (pawn_attacks[W][sqr] & pcon->pieces[BP])
		  | (knight_attacks[sqr] & (pcon->pieces[WN] | pcon->pieces[BN]))
		  | (Bmagic(sqr, pcon->pieces[OCCU]) & 
		  (pcon->pieces[WB] | pcon->pieces[BB] | pcon->pieces[WQ] | pcon->pieces[BQ]))
		  | (Rmagic(sqr, pcon->pieces[OCCU]) & 
		  (pcon->pieces[WR] | pcon->pieces[BR] | pcon->pieces[WQ] | pcon->pieces[BQ]));
}



// Based on Dr. Hyatt's algorithm (Crafty)
int Attacked(const CON *pcon, int sqr, int col)
{
	if (Rmagic(sqr, pcon->pieces[OCCU]) & 
		(pcon->pieces[ROOK|col] | pcon->pieces[QUEEN|col]))
		return 1;
	if (Bmagic(sqr, pcon->pieces[OCCU]) & 
		(pcon->pieces[BISHOP|col] | pcon->pieces[QUEEN|col]))
		return 1;
	if (knight_attacks[sqr] & pcon->pieces[KNIGHT|col])
		return 1;
	if (king_attacks[sqr] & pcon->pieces[KING|col])
		return 1;
	if (pawn_attacks[col ^ 1][sqr] & pcon->pieces[PAWN|col])
		return 1;
	return 0;
}



// Generates a bitboard of pinned pieces for the side to move,
// and determines if there are any distance checks (sliders)
void PinCheckUpdate(const CON *pcon, PSTE pste, int col)
{
	U64 sliders, pinned;
	int xcol = col ^ 1;

	pste->pinner = 0;
	pste->pinned = 0;

	sliders = (Rmagic(pcon->king[col], 0) & (pcon->pieces[ROOK|xcol] | pcon->pieces[QUEEN|xcol]))
		| (Bmagic(pcon->king[col], 0) & (pcon->pieces[BISHOP|xcol] | pcon->pieces[QUEEN|xcol]));
			
	while (sliders) {
		int from = PopLSB(&sliders);

		pinned = squares_between[from][pcon->king[col]] & pcon->pieces[OCCU];

		// If it's neither a pin nor a check
		if (pinned & pcon->pieces[W_OCCU - col] || pinned & (pinned - 1))
			continue;

		if (pinned) {
//            if( pinned & (pcon->pieces[KING] | pcon->pieces[KING|1]) ) {
//                Print(pcon, pste);
//                Assert(0);
//            }
			pste->pinned |= pinned;
			Set(from, pste->pinner);
		}
		else 
            pste->checkf += 2;
	}
}



CM* bGenCaptures(CON *pcon, PSTE pste, CM * list)
{
	U64 pc, moves;
	int sqr[2];
    const U8 *pos = pcon->pos;
    int side = pste->side;
    int xside = pste->xside;
    const unsigned char *king = pcon->king;
//    checkKing(pcon);
//    checkPos(pcon, true);

	// Remove pinned pieces from their piece boards
	for (pc = pste->pinner; pc; ) {
		U64 pin_ray;
		sqr[1] = PopLSB(&pc);

		pin_ray = squares_between[king[side]][sqr[1]];
		sqr[0] = LSB(pcon->pieces[B_OCCU|side] & pin_ray);
//        if( pos[sqr[0]] == 0 ) {
//            Print(pcon, pste);
//            Assert(0);
//        }
		Clear(sqr[0], pcon->pieces[pos[sqr[0]]]);

		switch (pcon->pos[sqr[0]]) {
			case BQ:	
				moves = Qmagic(sqr[0], pcon->pieces[OCCU]) & IdxToU64(sqr[1]);
				if (moves)	AddCapt(sqr[0], sqr[1], BQ, pos[sqr[1]]);
				break;
			case WQ:	
				moves = Qmagic(sqr[0], pcon->pieces[OCCU]) & IdxToU64(sqr[1]);
				if (moves)	AddCapt(sqr[0], sqr[1], WQ, pos[sqr[1]]);
				break;
			case BR:	
				moves = Rmagic(sqr[0], pcon->pieces[OCCU]) & IdxToU64(sqr[1]);
				if (moves)	AddCapt(sqr[0], sqr[1], BR, pos[sqr[1]]);
				break;
			case WR:	
				moves = Rmagic(sqr[0], pcon->pieces[OCCU]) & IdxToU64(sqr[1]);
				if (moves)	AddCapt(sqr[0], sqr[1], WR, pos[sqr[1]]);
				break;
			case BB:	
				moves = Bmagic(sqr[0], pcon->pieces[OCCU]) & IdxToU64(sqr[1]);
				if (moves)	AddCapt(sqr[0], sqr[1], BB, pos[sqr[1]]);
				break;
			case WB:
				moves = Bmagic(sqr[0], pcon->pieces[OCCU]) & IdxToU64(sqr[1]);
				if (moves)	AddCapt(sqr[0], sqr[1], WB, pos[sqr[1]]);
				break;
			case BP:	
				if (sqr[0] - 7 == sqr[1] || sqr[0] - 9 == sqr[1]) {
                    if (Rank(sqr[1]) > 0)	{AddCapt(sqr[0], sqr[1], BP, pos[sqr[1]]);}
					else					AddProm(sqr[0], sqr[1], BP, pos[sqr[1]]);
				}

				// Try the rare but possible pinnned e.p. capture
				if (pste->ep_t) {
					if (IdxToU64(pste->ep_t) & pin_ray) {
						if (sqr[0] - 7 == pste->ep_t
							|| sqr[0] - 9 == pste->ep_t) {
							AddCapt(sqr[0], pste->ep_t, BP, B_EP);
						}
					}
				}
				break;
			case WP:		
				if (sqr[0] + 7 == sqr[1] || sqr[0] + 9 == sqr[1]) {
                    if (Rank(sqr[1]) < 7)	{AddCapt(sqr[0], sqr[1], WP, pos[sqr[1]]);}
					else					AddProm(sqr[0], sqr[1], WP, pos[sqr[1]]);
				}

				// Try the rare but possible pinnned e.p. capture
				if (pste->ep_t) {
					if (IdxToU64(pste->ep_t) & pin_ray) {
						if (sqr[0] + 7 == pste->ep_t
							|| sqr[0] + 9 == pste->ep_t) {
							AddCapt(sqr[0], pste->ep_t, WP, W_EP);
						}
					}
				}
				break;
		}
	}

	// - Knights -
	for (pc = pcon->pieces[KNIGHT|side]; pc; ) {
		sqr[0] = PopLSB(&pc);
		moves = knight_attacks[sqr[0]] & pcon->pieces[B_OCCU|xside];

		while (moves) {
			sqr[1] = PopLSB(&moves);
			AddCapt(sqr[0], sqr[1], KNIGHT|side, pos[sqr[1]]);
		}
	}

	// - Bishop -
	for (pc = pcon->pieces[BISHOP|side]; pc; ) {
		sqr[0] = PopLSB(&pc);
		moves = Bmagic(sqr[0], pcon->pieces[OCCU]) & pcon->pieces[B_OCCU|xside];

		while (moves) {
			sqr[1] = PopLSB(&moves);
			AddCapt(sqr[0], sqr[1], BISHOP|side, pos[sqr[1]]);
		}
	}

	// - Rook -
	for (pc = pcon->pieces[ROOK|side]; pc; ) {
		sqr[0] = PopLSB(&pc);
		moves = Rmagic(sqr[0], pcon->pieces[OCCU]) & pcon->pieces[B_OCCU|xside];

		while (moves) {
			sqr[1] = PopLSB(&moves);
			AddCapt(sqr[0], sqr[1], ROOK|side, pos[sqr[1]]);
		}
	}

	// - Queen -
	for (pc = pcon->pieces[QUEEN|side]; pc; ) {
		sqr[0] = PopLSB(&pc);
		moves = Qmagic(sqr[0], pcon->pieces[OCCU]) & pcon->pieces[B_OCCU|xside];

		while (moves) {
			sqr[1] = PopLSB(&moves);
			AddCapt(sqr[0], sqr[1], QUEEN|side, pos[sqr[1]]);
		}
	}

	// - King -
	moves = king_attacks[king[side]] & pcon->pieces[B_OCCU|xside];
	while (moves) {
		sqr[1] = PopLSB(&moves);
		if (Attacked(pcon, sqr[1], xside))
			continue;
		AddCapt(king[side], sqr[1], KING|side, pos[sqr[1]]);
	}

	// - Pawns forward one -
	pcon->pieces[TEMP] = (pcon->pieces[xoccu_wp[side]] << 8) 
		& pcon->pieces[xoccu_bp[xside]];

	moves = pcon->pieces[TEMP] & prom_mask[side];
	while (moves) {
		sqr[side] = PopLSB(&moves);
		sqr[xside] = sqr[side] - 8;
		AddProm(sqr[0], sqr[1], PROM|side, 0);
	}

	// - Pawns seven-shift -
	pc = ((pcon->pieces[woccu_wp[side]] & CLEAR_LEFT) << 7)
		& pcon->pieces[boccu_bp[xside]];

	moves = pc & xprom_mask[side];
	while (moves) {
		sqr[side] = PopLSB(&moves);
		sqr[xside] = sqr[side] - 7;
		AddCapt(sqr[0], sqr[1], PAWN|side, pos[sqr[1]]);
	}

	moves = pc & prom_mask[side];
	while (moves) {
		sqr[side] = PopLSB(&moves);
		sqr[xside] = sqr[side] - 7;
		AddProm(sqr[0], sqr[1], PROM|side, pos[sqr[1]]);
	}

	// - Pawns nine-shift -
	pc = ((pcon->pieces[woccu_wp[side]] & CLEAR_RIGHT) << 9)
		& pcon->pieces[boccu_bp[xside]];

	moves = pc & xprom_mask[side];
	while (moves) {
		sqr[side] = PopLSB(&moves);
		sqr[xside] = sqr[side] - 9;
		AddCapt(sqr[0], sqr[1], PAWN|side, pos[sqr[1]]);
	}

	moves = pc & prom_mask[side];
	while (moves) {
		sqr[side] = PopLSB(&moves);
		sqr[xside] = sqr[side] - 9;
		AddProm(sqr[0], sqr[1], PROM|side, pos[sqr[1]]);
	}

	// - En Passant -
	if (pste->ep_t) {
		pc = (pcon->pieces[PAWN|side] & ~pste->pinned)
			& pawn_attacks[xside][pste->ep_t];

		while (pc) {
			U64 move_ep;
			sqr[0] = PopLSB(&pc);

			move_ep = IdxToU64(sqr[0]) | IdxToU64(pste->ep_t)
				| IdxToU64(pste->ep_t - p_forw[side]);

			pcon->pieces[OCCU] ^= move_ep;

			if (!(Rmagic(king[side], pcon->pieces[OCCU]) & (pcon->pieces[QUEEN|xside] | pcon->pieces[ROOK|xside]))) {
				AddCapt(sqr[0], pste->ep_t, PAWN|side, EP_CAPT|side);
			}

			pcon->pieces[OCCU] ^= move_ep;
		}
	}
//    checkKing(pcon);

	// Add pinned pieces back to their piece boards
	for (pc = pste->pinned; pc; ) {
		sqr[0] = PopLSB(&pc);
		Set(sqr[0], pcon->pieces[pos[sqr[0]]]);
	}
//    checkKing(pcon);

//    checkPos(pcon, false);
    (pste+1)->mlist_start = list;
	return list;
}



CM* bGenNonCaptures(PCON pcon, PSTE pste, CM * list)
{
	U64 pc, moves;
	int sqr[2];
    const U8 *pos = pcon->pos;
    int side = pste->side;
    int xside = pste->xside;
    const unsigned char *king = pcon->king;
//    checkPos(pcon, true);

	// Remove pinned pieces from their piece boards
	for (pc = pste->pinner; pc; ) {
		U64 pin_ray;
		sqr[1] = PopLSB(&pc);

		pin_ray = squares_between[king[side]][sqr[1]];
		sqr[0] = LSB(pcon->pieces[B_OCCU|side] & pin_ray);

		Clear(sqr[0], pcon->pieces[pos[sqr[0]]]);

		switch (pos[sqr[0]]) {
			case BQ:	
				moves = Qmagic(sqr[0], pcon->pieces[OCCU]) & pin_ray;
				while (moves) AddMove(sqr[0], PopLSB(&moves), BQ);
				break;
			case WQ:	
				moves = Qmagic(sqr[0], pcon->pieces[OCCU]) & pin_ray;
				while (moves) AddMove(sqr[0], PopLSB(&moves), WQ);
				break;
			case BR:	
				moves = Rmagic(sqr[0], pcon->pieces[OCCU]) & pin_ray;
				while (moves) AddMove(sqr[0], PopLSB(&moves), BR);
				break;
			case WR:	
				moves = Rmagic(sqr[0], pcon->pieces[OCCU]) & pin_ray;
				while (moves) AddMove(sqr[0], PopLSB(&moves), WR);
				break;
			case BB:	
				moves = Bmagic(sqr[0], pcon->pieces[OCCU]) & pin_ray;
				while (moves) AddMove(sqr[0], PopLSB(&moves), BB);
				break;
			case WB:
				moves = Bmagic(sqr[0], pcon->pieces[OCCU]) & pin_ray;
				while (moves) AddMove(sqr[0], PopLSB(&moves), WB);
				break;
			case BP:	
				if (Rank(sqr[0]) == RANK_2) {
					if (sqr[0] - 7 == sqr[1] || sqr[0] - 9 == sqr[1]) {
						AddUnderProm(sqr[0], sqr[1], BP, pos[sqr[1]]);
					}
				}

				// Straight moves
				if ((IdxToU64(sqr[0]) >> 8) & pin_ray) {
//					if (Rank(sqr[0]) == RANK_7 && pos[sqr[0] - 16] == 0) { 
					if (Rank(sqr[0]) == RANK_7 && pos[sqr[0] - 16] == 0
                        && (sqr[0] - 16) != king[B]) { 
						AddMove(sqr[0], sqr[0] - 16, BP);
					}
					AddMove(sqr[0], sqr[0] - 8, BP);
				}
				break;
			case WP:		
				if (Rank(sqr[0]) == RANK_7) {
					if (sqr[0] + 7 == sqr[1] || sqr[0] + 9 == sqr[1]) {
						AddUnderProm(sqr[0], sqr[1], WP, pos[sqr[1]]);
					}
				}

				// Straight moves
				if ((IdxToU64(sqr[0]) << 8) & pin_ray) {
//					if (Rank(sqr[0]) == RANK_2 && pos[sqr[0] + 16] == 0) { 
					if (Rank(sqr[0]) == RANK_2 && pos[sqr[0] + 16] == 0
                        && (sqr[0] + 16) != king[B]) { 
						AddMove(sqr[0], sqr[0] + 16, WP);
					}
					AddMove(sqr[0], sqr[0] + 8, WP);
				}
				break;
		}
	}

	// - Knights -
	for (pc = pcon->pieces[KNIGHT|side]; pc; ) {
		sqr[0] = PopLSB(&pc);
		moves = knight_attacks[sqr[0]] & pcon->pieces[X_OCCU];

		while (moves) {
			AddMove(sqr[0], PopLSB(&moves), KNIGHT|side);
		}
	}

	// - Bishop -
	for (pc = pcon->pieces[BISHOP|side]; pc; ) {
		sqr[0] = PopLSB(&pc);
		moves = Bmagic(sqr[0], pcon->pieces[OCCU]) & pcon->pieces[X_OCCU];

		while (moves) {
			AddMove(sqr[0], PopLSB(&moves), BISHOP|side);
		}
	}

	// - Rook -
	for (pc = pcon->pieces[ROOK|side]; pc; ) {
		sqr[0] = PopLSB(&pc);
		moves = Rmagic(sqr[0], pcon->pieces[OCCU]) & pcon->pieces[X_OCCU];

		while (moves) {
			AddMove(sqr[0], PopLSB(&moves), ROOK|side);
		}
	}

	// - Queen -
	for (pc = pcon->pieces[QUEEN|side]; pc; ) {
		sqr[0] = PopLSB(&pc);
		moves = Qmagic(sqr[0], pcon->pieces[OCCU]) & pcon->pieces[X_OCCU];

		while (moves) {
			AddMove(sqr[0], PopLSB(&moves), QUEEN|side);
		}
	}

	// - King -
	moves = king_attacks[king[side]] & pcon->pieces[X_OCCU];
	
	while (moves) {
		sqr[1] = PopLSB(&moves);
		if (Attacked(pcon, sqr[1], xside))
			continue;
		AddMove(king[side], sqr[1], KING|side);
	}

	// - Castling -
	if (pste->castle & castle_mask[side]) {
		if (pste->castle & perm_mask_00[side]
			&& (pcon->pieces[OCCU] & mask_00[side]) == 0
			&& !Attacked(pcon, king[side] + 1, xside)
			&& !Attacked(pcon, king[side] + 2, xside)){
				AddMove(king[side], king[side] + 2, CASTLE|side);
		}
		if (pste->castle & perm_mask_000[side]
			&& (pcon->pieces[OCCU] & mask_000[side]) == 0
			&& !Attacked(pcon, king[side] - 1, xside)
			&& !Attacked(pcon, king[side] - 2, xside)){
				AddMove(king[side], king[side] - 2, CASTLE|side);
		}
	}

	// - Pawns forward one -
	pcon->pieces[TEMP] = (pcon->pieces[xoccu_wp[side]] << 8) 
		& pcon->pieces[xoccu_bp[xside]];

	moves = pcon->pieces[TEMP] & xprom_mask[side];
	while (moves) {
		sqr[side] = PopLSB(&moves);
		sqr[xside] = sqr[side] - 8;
		AddMove(sqr[0], sqr[1], pos[sqr[0]]);
	}

	moves = pcon->pieces[TEMP] & prom_mask[side];
	while (moves) {
		sqr[side] = PopLSB(&moves);
		sqr[xside] = sqr[side] - 8;
		AddUnderProm(sqr[0], sqr[1], PROM|side, 0);
	}

	// - Pawns forward two -
	moves = ((pcon->pieces[xoccu_tmp[side]] & mask_5_3[side]) << (8 << xside)) 
		& pcon->pieces[xoccu_tmp[xside]];

	while (moves) {
		sqr[side] = PopLSB(&moves);
		sqr[xside] = sqr[side] - 16;
		AddMove(sqr[0], sqr[1], pos[sqr[0]]);
	}

	// - Pawns seven-shift -
	pc = ((pcon->pieces[woccu_wp[side]] & CLEAR_LEFT) << 7)
		& pcon->pieces[boccu_bp[xside]];

	moves = pc & prom_mask[side];
	while (moves) {
		sqr[side] = PopLSB(&moves);
		sqr[xside] = sqr[side] - 7;
		AddUnderProm(sqr[0], sqr[1], PROM|side, pos[sqr[1]]);
	}

	// - Pawns nine-shift -
	pc = ((pcon->pieces[woccu_wp[side]] & CLEAR_RIGHT) << 9)
		& pcon->pieces[boccu_bp[xside]];

	moves = pc & prom_mask[side];
	while (moves) {
		sqr[side] = PopLSB(&moves);
		sqr[xside] = sqr[side] - 9;
		AddUnderProm(sqr[0], sqr[1], PROM|side, pos[sqr[1]]);
	}

	// Add pinned pieces back to their piece boards
	for (pc = pste->pinned; pc; ) {
		sqr[0] = PopLSB(&pc);
		Set(sqr[0], pcon->pieces[pos[sqr[0]]]);
	}
//    checkPos(pcon, false);

    (pste+1)->mlist_start = list;
	return list;
}



// Generate moves that move out of check. Check flags:
// [1: 0001] Contact check		[3: 0011] Contact/Distance check
// [2: 0010] Distance check		[4: 0100] Distance double-check
CM* bGenEvasions(PCON pcon, PSTE pste, CM * list)
{
	U64 pc, moves, x_pin = ~pste->pinned;
	int sqr[2];
    const U8 *pos = pcon->pos;
    int side = pste->side;
    int xside = pste->xside;
    const unsigned char *king = pcon->king;
    if( list == NULL )
        list = pste->mlist_start;

//    checkPos(pcon, true);
//    checkKing(pcon);
    pc = king_attacks[king[side]];

	// Prevent the king from hiding behind himself
	Clear(king[side], pcon->pieces[OCCU]);
	
	// - King Captures -
	for (moves = pc & pcon->pieces[B_OCCU|xside]; moves; ) {
		sqr[1] = PopLSB(&moves);
		if (Attacked(pcon, sqr[1], xside)) 
			continue;
		AddCapt(king[side], sqr[1], KING|side, pos[sqr[1]]);
	}

	// - King Moves -
	for (moves = pc & pcon->pieces[X_OCCU]; moves; ) {
		sqr[1] = PopLSB(&moves);
		if (Attacked(pcon, sqr[1], xside))
			continue;
		AddMove(king[side], sqr[1], KING|side);
	}
	
	// Add the king back onto the occupancy
	Set(king[side], pcon->pieces[OCCU]);

	// If it's not double check, other moves are possible
	if (pste->checkf <= 2) {
		U64 interpose = 0, targ = AttacksTo(pcon, king[side]) & pcon->pieces[B_OCCU|xside];
		sqr[1] = LSB(targ);

		// Set the interpose board
		if (pste->checkf == 2) {
			interpose = squares_between[king[side]][sqr[1]];
		}

		// - Knights -
		for (pc = pcon->pieces[KNIGHT|side] & x_pin; pc; ) {
			sqr[0] = PopLSB(&pc);
			moves = knight_attacks[sqr[0]];

			if (moves & targ) AddCapt(sqr[0], sqr[1], KNIGHT|side, pos[sqr[1]]);
			moves &= interpose;
			while (moves) AddMove(sqr[0], PopLSB(&moves), KNIGHT|side);
		}

		// - Bishops -
		for (pc = pcon->pieces[BISHOP|side] & x_pin; pc; ) {
			sqr[0] = PopLSB(&pc);
			moves = Bmagic(sqr[0], pcon->pieces[OCCU]);

			if (moves & targ) AddCapt(sqr[0], sqr[1], BISHOP|side, pos[sqr[1]]);
			moves &= interpose;
			while (moves) AddMove(sqr[0], PopLSB(&moves), BISHOP|side);
		}

		// - Rooks -
		for (pc = pcon->pieces[ROOK|side] & x_pin; pc; ) {
			sqr[0] = PopLSB(&pc);
			moves = Rmagic(sqr[0], pcon->pieces[OCCU]);

			if (moves & targ) AddCapt(sqr[0], sqr[1], ROOK|side, pos[sqr[1]]);
			moves &= interpose;
			while (moves) AddMove(sqr[0], PopLSB(&moves), ROOK|side);
		}

		// - Queens -
		for (pc = pcon->pieces[QUEEN|side] & x_pin; pc; ) {
			sqr[0] = PopLSB(&pc);
			moves = Qmagic(sqr[0], pcon->pieces[OCCU]);

			if (moves & targ) AddCapt(sqr[0], sqr[1], QUEEN|side, pos[sqr[1]]);
			moves &= interpose;
			while (moves) AddMove(sqr[0], PopLSB(&moves), QUEEN|side);
		}

		// - Pawn Captures -
		pc = pawn_attacks[xside][sqr[1]] & (pcon->pieces[PAWN|side] & x_pin);

		moves = pc & ev_xprom_mask[side];
		while (moves) {
			sqr[0] = PopLSB(&moves);
			AddCapt(sqr[0], sqr[1], PAWN|side, pos[sqr[1]]);
		}

		moves = pc & ev_prom_mask[side];
		while (moves) {
			sqr[0] = PopLSB(&moves);
			AddProm(sqr[0], sqr[1], PROM|side, pos[sqr[1]]);
			AddUnderProm(sqr[0], sqr[1], PROM|side, pos[sqr[1]]);
		}

		// - En Passant Evasion -
		if (pste->ep_t) {
			// Try to move attacker to e.p. square
			sqr[1] += p_forw[side];

			if (sqr[1] == pste->ep_t) {
				moves = (pawn_attacks[xside][sqr[1]] & pcon->pieces[PAWN|side]) & x_pin;
				while (moves) {
					AddCapt(PopLSB(&moves), sqr[1], PAWN|side, EP_CAPT|side);
				}
			}
		}

		// After e.p. evasions, attacker (sqr[1]) has been lost

		// - Pawns forward one -
		pc = pcon->pieces[PAWN|side]; // Back-up pawn bitboard
		pcon->pieces[PAWN|side] &= x_pin;

		pcon->pieces[TEMP] = (pcon->pieces[xoccu_wp[side]] << 8)
			& pcon->pieces[xoccu_bp[xside]];

		moves = pcon->pieces[TEMP] & xprom_mask[side] & (interpose << inter_shift[side]);
		while (moves) {
			sqr[side] = PopLSB(&moves);
			sqr[xside] = sqr[side] - 8;
			AddMove(sqr[0], sqr[1], pos[sqr[0]]);
		}

		moves = pcon->pieces[TEMP] & prom_mask[side] & (interpose << inter_shift[side]);
		while (moves) {
			sqr[side] = PopLSB(&moves);
			sqr[xside] = sqr[side] - 8;
			AddUnderProm(sqr[0], sqr[1], PROM|side, 0);
		}

		// - Pawns forward two -
		moves = ((pcon->pieces[xoccu_tmp[side]] & mask_5_3[side]) << (8 << xside)) 
			& pcon->pieces[xoccu_tmp[xside]];

		moves &= (interpose << inter_shift2[side]);
		while (moves) {
			sqr[side] = PopLSB(&moves);
			sqr[xside] = sqr[side] - 16;
			AddMove(sqr[0], sqr[1], pos[sqr[0]]);
		}

		pcon->pieces[PAWN|side] = pc; // Restore pawn bitboard
	}
//    checkKing(pcon);
//    checkPos(pcon, false);

    (pste+1)->mlist_start = list;
	return list;
}



void bMove(PCON pcon, PSTE pste, U32 move)
{
    int from, to;
    U64 to_64, from_64, move_64;
    U8 *pos = pcon->pos;
    unsigned char *king = pcon->king;
    PSTE npste = pste + 1;  // next stack element

    pste->lastMove = move;  // to display the current variation
#ifdef _DEBUG
    npste->lastMove = 0;    // to display the current variation
#endif

//    checkKing(pcon);

    // Update the current status and ply
    npste->castle   = pste->castle;
    npste->ep_t     = 0;
    npste->checkf   = 0;
    npste->side     = pste->side ^ 1;
    npste->xside    = pste->xside ^ 1;
    npste->fNull    = false;        // By default, null-move is allowed.
	//
	//	Initialize various values for the next ply.
	//
	npste->plyFifty = pste->plyFifty + 1;
	npste->hashkPc  = HashkSwitch(pste->hashkPc);
	npste->hashkPn  = pste->hashkPn;
	npste->valPcUs  = pste->valPcThem;
	npste->valPcThem= pste->valPcUs;
	npste->valPnUs  = pste->valPnThem;
	npste->valPnThem= pste->valPnUs;

    pste++;
	// Decompress the move
	from	= From(move);
	to		= To(move);
	
	to_64	= IdxToU64(to);
	from_64 = IdxToU64(from);
	move_64	= to_64 | from_64;	
	
    int piece = Pc(move);
	// Perform captures, if any
    int captured = Capt(move);
//    if( (pos[from] != piece && ((piece|1) != (B_PROM|1))) || (pos[to] != captured) ) {
//        Print(pcon, pste);
//        Assert(0);
//    }
	// Move the piece on the array board
	pos[to]		= pos[from];
	pos[from]	= 0;

    // note : king can not be captured
    if(captured) {
        npste->plyFifty = 0;			    	// Capture resets this.
        npste->prevCaptValue = b_valPc[ captured ];
        npste->prevCaptSqr   = to;
    }

	switch (captured) {
        case 0:
            npste->prevCaptValue    = 0;
            break;
		case BQ:
			pcon->pieces[B_OCCU]	^= to_64;
			pcon->pieces[BQ]		^= to_64;
            npste->valPcUs          -= valQUEEN;
            // XOR the piece out of the hash key
            rehashBB(npste->hashkPc, captured, to);
			break;
		case WQ:
			pcon->pieces[W_OCCU]	^= to_64;
			pcon->pieces[WQ]		^= to_64;
            npste->valPcUs          -= valQUEEN;
            rehashBB(npste->hashkPc, captured, to);
			break;
		case BR:
			pcon->pieces[B_OCCU]	^= to_64;
			pcon->pieces[BR]		^= to_64;
			npste->castle           &= castle_spoiler[to];
            npste->valPcUs          -= valROOK;
            rehashBB(npste->hashkPc, captured, to);
			break;
		case WR:
			pcon->pieces[W_OCCU]	^= to_64;
			pcon->pieces[WR]		^= to_64; 
			npste->castle           &= castle_spoiler[to];
            npste->valPcUs          -= valROOK;
            rehashBB(npste->hashkPc, captured, to);
			break;
		case BB:
			pcon->pieces[B_OCCU]	^= to_64;
			pcon->pieces[BB]		^= to_64; 
            npste->valPcUs          -= valBISHOP;
            rehashBB(npste->hashkPc, captured, to);
			break;
		case WB:
			pcon->pieces[W_OCCU]	^= to_64;
			pcon->pieces[WB]		^= to_64;
            npste->valPcUs          -= valBISHOP;
            rehashBB(npste->hashkPc, captured, to);
			break;
		case BN:
			pcon->pieces[B_OCCU]	^= to_64;
			pcon->pieces[BN]		^= to_64;
            npste->valPcUs          -= valKNIGHT;
            rehashBB(npste->hashkPc, captured, to);
			break;
		case WN:
			pcon->pieces[W_OCCU]	^= to_64;
			pcon->pieces[WN]		^= to_64;
            npste->valPcUs          -= valKNIGHT;
            rehashBB(npste->hashkPc, captured, to);
			break;
		case BP:
			pcon->pieces[B_OCCU]	^= to_64;
			pcon->pieces[BP]		^= to_64;
            npste->valPnUs          -= valPAWN;
            rehashBB(npste->hashkPc, captured, to);
			rehashBB( npste->hashkPn, captured, to);
			break;
		case WP:
			pcon->pieces[W_OCCU]	^= to_64;
			pcon->pieces[WP]		^= to_64;
            npste->valPnUs          -= valPAWN;
            rehashBB(npste->hashkPc, captured, to);
			rehashBB( npste->hashkPn, captured, to);
			break;
		case B_EP:
			pos[to + 8]		        = 0;
			pcon->pieces[WP]		^= (to_64 << 8);
			pcon->pieces[W_OCCU]	^= (to_64 << 8);
            npste->valPnUs          -= valPAWN;
            rehashBB(npste->hashkPc, captured, to + 8);
            rehashBB(npste->hashkPn, captured, to + 8);
			break;
		case W_EP:
			pos[to - 8]		        = 0;
			pcon->pieces[BP]		^= (to_64 >> 8);
			pcon->pieces[B_OCCU]	^= (to_64 >> 8);
            npste->valPnUs          -= valPAWN;
            rehashBB(npste->hashkPc, captured, to - 8);
            rehashBB(npste->hashkPn, captured, to - 8);
			break;
	}
	rehashBB( npste->hashkPc, piece, from );
	rehashBB( npste->hashkPc, piece, to );

	// Move the piece on the bitboards
	switch (Pc(move)) {
		case BK:
			npste->castle	        &= 3;
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BK]		^= move_64;
			king[B]				    = U8(to);
			break;
		case WK:	
			npste->castle	        &= 12;
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WK]		^= move_64;
			king[W]				    = U8(to);
			break;
		case BQ:	
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BQ]		^= move_64;
			break;
		case WQ:	
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WQ]		^= move_64;
			break;
		case BR:	
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BR]		^= move_64;
			npste->castle           &= castle_spoiler[from];
			break;
		case WR:	
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WR]		^= move_64; 	
			npste->castle           &= castle_spoiler[from];
			break;
		case BB:	
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BB]		^= move_64;
			break;
		case WB:	
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WB]		^= move_64;
			break;
		case BN:	
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BN]		^= move_64;
			if (knight_attacks[to] & pcon->pieces[WK])
				npste->checkf = 1;
			break;
		case WN:	
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WN]		^= move_64;
			if (knight_attacks[to] & pcon->pieces[BK])
				npste->checkf = 1;
			break;
		case BP:	
            npste->plyFifty         = 0;
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BP]		^= move_64;
			if (from - to == 16 && (pcon->pieces[WP] & pawn_attacks[B][from - 8]))
				npste->ep_t = U8(from - 8);
			if (pawn_attacks[B][to] & pcon->pieces[WK])
				npste->checkf = 1;
	        rehashBB( npste->hashkPn, piece, from );
	        rehashBB( npste->hashkPn, piece, to );
			break;
		case WP:
            npste->plyFifty         = 0;
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WP]		^= move_64;
			if (to - from == 16 && (pcon->pieces[BP] & pawn_attacks[W][from + 8]))
				npste->ep_t = U8(from + 8);
			if (pawn_attacks[W][to] & pcon->pieces[BK])
				npste->checkf = 1;
	        rehashBB( npste->hashkPn, piece, from );
	        rehashBB( npste->hashkPn, piece, to );
			break;
		case B_CASTLE:
			npste->castle	        &= 3;
			pcon->pieces[BK]		^= move_64;
			king[B]				    = U8(to);

			if (to == G8) {
				pcon->pieces[B_OCCU]	^= 0xF000000000000000ULL;
				pcon->pieces[BR]		^= 0xA000000000000000ULL;
				pos[F8]			= BR;
				pos[H8]			= 0;
	            rehashBB( npste->hashkPc, BR, H8 );
	            rehashBB( npste->hashkPc, BR, F8 );
			}
			else {
				pcon->pieces[B_OCCU]	^= 0x1D00000000000000ULL;
				pcon->pieces[BR]		^= 0x0900000000000000ULL;
				pos[A8]			= 0;
				pos[D8]			= BR;
	            rehashBB( npste->hashkPc, BR, A8 );
	            rehashBB( npste->hashkPc, BR, D8 );
			}
			break;
		case W_CASTLE:
			npste->castle	        &= 12;
			pcon->pieces[WK]		^= move_64;
			king[W]				    = U8(to);

			if (to == G1) {				
				pcon->pieces[W_OCCU]	^= 0x00000000000000F0ULL;
				pcon->pieces[WR]		^= 0x00000000000000A0ULL;
				pos[F1]			= WR;
				pos[H1]			= 0;
	            rehashBB( npste->hashkPc, WR, H1 );
	            rehashBB( npste->hashkPc, WR, F1 );
			}
			else {
				pcon->pieces[W_OCCU]	^= 0x000000000000001DULL;
				pcon->pieces[WR]		^= 0x0000000000000009ULL;
				pos[A1]			= 0;
				pos[D1]			= WR;
	            rehashBB( npste->hashkPc, WR, A1 );
	            rehashBB( npste->hashkPc, WR, D1 );
			}
			break;
		case B_PROM:
			pcon->pieces[BP]			^= from_64;
			pcon->pieces[B_OCCU]		^= move_64;
            npste->plyFifty             = 0;
            npste->valPnThem            -= valPAWN;
            // delete the pawn
            rehashBB( npste->hashkPc, BP, from );
            rehashBB( npste->hashkPn, BP, from );

			switch (Prom(move)) {
				case PROM_Q:
					pcon->pieces[BQ]	^= to_64;
					pos[to]		        = BQ;
                    npste->valPcThem      += valQUEEN;
                    // add a piece
                    rehashBB( npste->hashkPc, BQ, to);
					break;
				case PROM_R:
					pcon->pieces[BR]	^= to_64;
					pos[to]		        = BR;
                    npste->valPcThem      += valROOK;
                    rehashBB( npste->hashkPc, BR, to);
					break;
				case PROM_B:
					pcon->pieces[BB]	^= to_64;
					pos[to]		        = BB;
                    npste->valPcThem      += valBISHOP;
                    rehashBB( npste->hashkPc, BB, to);
					break;
				case PROM_N:
					pcon->pieces[BN]	^= to_64;
					pos[to]		        = BN;
                    npste->valPcThem      += valKNIGHT;
                    rehashBB( npste->hashkPc, BN, to);
					if (knight_attacks[to] & pcon->pieces[WK])
						npste->checkf = 1;
					break;
			}
			break;
		case W_PROM:
			pcon->pieces[WP]			^= from_64;
			pcon->pieces[W_OCCU]		^= move_64;
            npste->plyFifty             = 0;
            npste->valPnThem              -= valPAWN;
            // delete the pawn
            rehashBB( npste->hashkPc, WP, from );
            rehashBB( npste->hashkPn, WP, from );

			switch (Prom(move)) {
				case PROM_Q:
					pcon->pieces[WQ]	^= to_64;
					pos[to]		        = WQ;
                    npste->valPcThem      += valQUEEN;
                    // add a piece
                    rehashBB( npste->hashkPc, WQ, to);
					break;
				case PROM_R:
					pcon->pieces[WR]	^= to_64;
					pos[to]		        = WR;
                    npste->valPcThem      += valROOK;
                    rehashBB( npste->hashkPc, WR, to);
					break;
				case PROM_B:
					pcon->pieces[WB]	^= to_64;
					pos[to]		        = WB;
                    npste->valPcThem      += valBISHOP;
                    rehashBB( npste->hashkPc, WB, to);
					break;
				case PROM_N:
					pcon->pieces[WN]	^= to_64;
					pos[to]		        = WN;
                    npste->valPcThem      += valKNIGHT;
                    rehashBB( npste->hashkPc, WN, to);
					if (knight_attacks[to] & pcon->pieces[BK])
						npste->checkf = 1;
					break;
			}
			break;
	}
	pcon->pieces[OCCU] = pcon->pieces[B_OCCU] | pcon->pieces[W_OCCU];
	pcon->pieces[X_OCCU] = ~pcon->pieces[OCCU];

	PinCheckUpdate(pcon, npste, npste->side);
//    checkKing(pcon);
//    if( pos[to] == 0 && piece) {
//        Print(pcon, npste);
//        Assert(0);
//    }
}

void bNullMove(PCON pcon, PSTE pste)
{
    PSTE npste = pste + 1;  // next stack element
    pste->lastMove = 0;     // to display the current variation
#ifdef _DEBUG
    npste->lastMove = 0;    // to display the current variation
#endif
//    checkKing(pcon);

    // Update the current status and ply
    npste->castle   = pste->castle;     // Castling flags are passed.
    npste->ep_t     = 0;
    npste->checkf   = 0;
    npste->side     = pste->side ^ 1;
    npste->xside    = pste->xside ^ 1;
    npste->fNull    = true;             // Null move not allowed.
	npste->plyFifty = pste->plyFifty + 1;	//  little strange but probably okay.

	npste->mlist_start = pste->mlist_start;
	//
	//	Initialize various values for the next ply.
	//
	npste->plyFifty = pste->plyFifty + 1;
	npste->hashkPc  = HashkSwitch(pste->hashkPc);
	npste->hashkPn  = pste->hashkPn;
	npste->valPcUs  = pste->valPcThem;
	npste->valPcThem= pste->valPcUs;
	npste->valPnUs  = pste->valPnThem;
	npste->valPnThem= pste->valPnUs;
    npste->prevCaptValue    = 0;

	PinCheckUpdate(pcon, npste, npste->side);
//    checkKing(pcon);

}


void bUndoMove(PCON pcon, PSTE pste, U32 move)
{
	U64 to_64, from_64, move_64;
	int from, to, pc, capt;
    U8 *pos = pcon->pos;
    unsigned char *king = pcon->king;

//    checkKing(pcon);
	// Decompress the move
	from	= From(move);
	to		= To(move);
	pc		= Pc(move);
	capt	= Capt(move);

	to_64	= IdxToU64(to);
	from_64 = IdxToU64(from);
	move_64 = to_64 | from_64;

	// Move the piece on the array board
	pos[from]	= U8(pc);

	// Perform captures, if any
	switch (capt) {
		case 0:
			pos[to] = 0;
			break;
		case BQ:	
			pos[to]			= BQ;
			pcon->pieces[B_OCCU]	^= to_64;
			pcon->pieces[BQ]		^= to_64;
			break;
		case WQ:	
			pos[to]			= WQ;
			pcon->pieces[W_OCCU]	^= to_64;
			pcon->pieces[WQ]		^= to_64;
			break;
		case BR:
			pos[to]			= BR;
			pcon->pieces[B_OCCU]		^= to_64;
			pcon->pieces[BR]		^= to_64;
			break;
		case WR:	
			pos[to]			= WR;
			pcon->pieces[W_OCCU]	^= to_64;
			pcon->pieces[WR]		^= to_64; 
			break;
		case BB:	
			pos[to]			= BB;
			pcon->pieces[B_OCCU]	^= to_64;
			pcon->pieces[BB]		^= to_64; 
			break;
		case WB:	
			pos[to]			= WB;
			pcon->pieces[W_OCCU]	^= to_64;
			pcon->pieces[WB]		^= to_64;
			break;
		case BN:	
			pos[to]			= BN;
			pcon->pieces[B_OCCU]	^= to_64;
			pcon->pieces[BN]		^= to_64;
			break;
		case WN:	
			pos[to]			= WN;
			pcon->pieces[W_OCCU]	^= to_64;
			pcon->pieces[WN]		^= to_64;
			break;
		case BP:	
			pos[to]			= BP;
			pcon->pieces[B_OCCU]	^= to_64;
			pcon->pieces[BP]		^= to_64;
			break;
		case WP:		
			pos[to]			= WP;
			pcon->pieces[W_OCCU]	^= to_64;
			pcon->pieces[WP]		^= to_64;
			break;
		case B_EP:
			pos[to]			= 0;
			pos[to + 8]		= WP;
			pcon->pieces[WP]		^= (to_64 << 8);
			pcon->pieces[W_OCCU]	^= (to_64 << 8);
			break;
		case W_EP:
			pos[to]			= 0;
			pos[to - 8]		= BP;
			pcon->pieces[BP]		^= (to_64 >> 8);
			pcon->pieces[B_OCCU]	^= (to_64 >> 8);
	}

	// Move the piece on the bitboards
	switch (pc) {
		case BK:	
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BK]		^= move_64;
			king[B]			= U8(from);
			break;
		case WK:	
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WK]		^= move_64;
			king[W]			= U8(from);
			break;
		case BQ:	
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BQ]		^= move_64;
			break;
		case WQ:	
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WQ]		^= move_64;
			break;
		case BR:	
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BR]		^= move_64;
			break;
		case WR:	
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WR]		^= move_64; 
			break;
		case BB:	
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BB]		^= move_64; 
			break;
		case WB:	
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WB]		^= move_64;
			break;
		case BN:	
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BN]		^= move_64;
			break;
		case WN:	
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WN]		^= move_64;
			break;
		case BP:	
			pcon->pieces[B_OCCU]	^= move_64;
			pcon->pieces[BP]		^= move_64;
			break;
		case WP:		
			pcon->pieces[W_OCCU]	^= move_64;
			pcon->pieces[WP]		^= move_64;	
			break;
		case B_CASTLE:
			pcon->pieces[BK]		^= move_64;
			king[B]			= U8(from);

			if (to == G8) {
				pcon->pieces[B_OCCU]	^= 0xF000000000000000ULL;
				pcon->pieces[BR]		^= 0xA000000000000000ULL;
				pos[E8]			= BK;
				pos[F8]			= 0;
				pos[H8]			= BR;
			}
			else {
				pcon->pieces[B_OCCU]	^= 0x1D00000000000000ULL;
				pcon->pieces[BR]		^= 0x0900000000000000ULL;
				pos[A8]			= BR;
				pos[D8]			= 0;
				pos[E8]			= BK;
			}
			break;
		case W_CASTLE:
			pcon->pieces[WK]			^= move_64;
			king[W]				= U8(from);

			if (to == G1) {				
				pcon->pieces[W_OCCU]	^= 0xF0ULL;
				pcon->pieces[WR]		^= 0xA0ULL;
				pos[E1]			= WK;
				pos[F1]			= 0;
				pos[H1]			= WR;	
			}
			else {
				pcon->pieces[W_OCCU]	^= 0x1DULL;
				pcon->pieces[WR]		^= 0x9ULL;
				pos[A1]			= WR;
				pos[D1]			= 0;
				pos[E1]			= WK;
			}
			break;
		case B_PROM:
			pcon->pieces[BP]		^= from_64;
			pcon->pieces[B_OCCU]	^= move_64;
			pos[from]		= BP;

			switch (Prom(move)) {
				case PROM_Q:
					pcon->pieces[BQ]	^= to_64;
					break;
				case PROM_R:
					pcon->pieces[BR]	^= to_64;
					break;
				case PROM_B:
					pcon->pieces[BB]	^= to_64;
					break;
				case PROM_N:
					pcon->pieces[BN]	^= to_64;
					break;
			}
			break;
		case W_PROM:
			pcon->pieces[WP]			^= from_64;
			pcon->pieces[W_OCCU]		^= move_64;
			pos[from]			= WP;

			switch (Prom(move)) {
				case PROM_Q:
					pcon->pieces[WQ]	^= to_64;
					break;
				case PROM_R:
					pcon->pieces[WR]	^= to_64;
					break;
				case PROM_B:
					pcon->pieces[WB]	^= to_64;
					break;
				case PROM_N:
					pcon->pieces[WN]	^= to_64;
					break;
			}
			break;
	}

	pcon->pieces[OCCU] = pcon->pieces[B_OCCU] | pcon->pieces[W_OCCU];
	pcon->pieces[X_OCCU] = ~pcon->pieces[OCCU];
//    checkKing(pcon);
}



// Initializes and begins an endless perft loop. It can accept a fen string
// (no quotations, etc.) as an argument, and _will_ crash on invalid fens
int testMAGIC(PCON pcon, PSTE pste, int depth)
{
	int i;
	clock_t start, end;

//	SetFromFEN("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");

//	SetFromFEN("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");

//	if (fen != 0) {
//		SetFromFEN(fen);
//	}

	Print(pcon, pste);

	// Begin an infinite perft loop
	for (i = 1; i <= depth; i++) {
		nodes = 0;

		start = clock();

		Perft(pcon, pste, i);

		end = clock();

		float dtime = (float(end - start) / CLOCKS_PER_SEC);

		printf("%2d. %10d %7.2f\n", i, int(nodes), dtime);
	}

	return 0;
}
void computeMobility(CON *pcon, PSTE pste, int side, bitboard xpawns, _dynamicVal & dynamicVal) {
    bitboard offenseMapSide[2] = {0x00000000ffffffffULL, 0xffffffff00000000ULL};

	U64 pc, moves;
	int sqr;
    int xside = side^1;
    const unsigned char *king = pcon->king;
    bitboard offenseMap = offenseMapSide[side];
    bitboard kingSquare = king_base[king[xside]];
    dynamicVal.c_move = 0;
    dynamicVal.c_king = 0;
    dynamicVal.c_offense = 0;
//    dynamicVal.c_bishop = 0;
    dynamicVal.attacks = 0;
    // we will compute mobility thru our own pieces but not thru our pawns
    // and not thru king so trapped rook does not think it has some mobility
    bitboard xrayblocker = pcon->pieces[PAWN|side] | pcon->pieces[KING|side];

	// - Knights -
	for (pc = pcon->pieces[KNIGHT|side]; pc; ) {
		sqr = PopLSB(&pc);
		moves = knight_attacks[sqr];

        dynamicVal.attacks |= moves;
//        dynamicVal.c_move += popCount(moves & xpawns);
        dynamicVal.c_offense += popCount(moves & offenseMap);
        dynamicVal.c_king += popCount(moves & kingSquare);
	}

	// - Bishop -
	for (pc = pcon->pieces[BISHOP|side]; pc; ) {
		sqr = PopLSB(&pc);
		moves = Bmagic(sqr, pcon->pieces[B_OCCU|xside]|xrayblocker) &~ xrayblocker;
        dynamicVal.attacks |= moves;
//        dynamicVal.c_move += popCount(moves & xpawns);
        dynamicVal.c_offense += popCount(moves & offenseMap);
        dynamicVal.c_king += popCount(moves & kingSquare);
	}

	// - Rook -
	for (pc = pcon->pieces[ROOK|side]; pc; ) {
		sqr = PopLSB(&pc);
		moves = Rmagic(sqr, pcon->pieces[B_OCCU|xside]|xrayblocker) &~ xrayblocker;
        dynamicVal.attacks |= moves;
//        dynamicVal.c_move += popCount(moves & xpawns);
        dynamicVal.c_offense += popCount(moves & offenseMap);
        dynamicVal.c_king += popCount(moves & kingSquare);
	}

	// - Queen -
	for (pc = pcon->pieces[QUEEN|side]; pc; ) {
		sqr = PopLSB(&pc);
		moves = Qmagic(sqr, pcon->pieces[B_OCCU|xside]|xrayblocker) &~ xrayblocker;
        dynamicVal.attacks |= moves;
//        dynamicVal.c_move += popCount(moves & xpawns);
        dynamicVal.c_offense += popCount(moves & offenseMap);
        dynamicVal.c_king += popCount(moves & kingSquare);
	}

	// - King -
	moves = king_attacks[king[side]] & pcon->pieces[B_OCCU|xside];
    dynamicVal.attacks |= moves;
//    dynamicVal.c_move += popCount(moves & xpawns);
    dynamicVal.c_offense += popCount(moves & offenseMap);

}

