
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	Gerbil
//
//	Copyright (c) 2001, Bruce Moreland.  All rights reserved.
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
//
//	This file is part of the Gerbil chess program project.
//
//	Gerbil is free software; you can redistribute it and/or modify it under
//	the terms of the GNU General Public License as published by the Free
//	Software Foundation; either version 2 of the License, or (at your option)
//	any later version.
//
//	Gerbil is distributed in the hope that it will be useful, but WITHOUT ANY
//	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
//	FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
//	details.
//
//	You should have received a copy of the GNU General Public License along
//	with Gerbil; if not, write to the Free Software Foundation, Inc.,
//	59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#include "engine.hpp"
#include "gproto.hpp"
#include "eval.hpp"
#include "moves.hpp"
#include "genmagic.hpp"

#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

#ifdef	DEBUG
static char const s_aszModule[] = __FILE__;
#endif

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

// the 16 center pawns
const U64 centerMask = 0x00003C3C3C3C0000ULL;

static HASHPawn *s_rghashp;			// Pawn hash table.
static unsigned	s_chashpMac = 0;	// Number of elements in the table, minus one.  I
							//  do the minus one thing in order to avoid an
							//  inner-loop subtract, which is no big deal, but
							//  every little bit helps.

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

#ifdef _DEBUG

//	I sat here and thought about how to write this routine, for about five
//	minutes.  I could have written something that would have required
//	modification any time the pawn structure stuff is messed with, but I
//	didn't like that idea, since the code I've used in the past for this
//	routine gets kind of annyong to change.

//	On the other hand, this program is supposed to be simple and direct.

//	Finally, I could not resist doing something I have never done -- write
//	something that works like "printf".

//	The format string is comprised of commands.  The commands are as follows:

//		digit ('0'..'9')		The next argument is an array of "digit"
//								bitmaps, to be displayed vertically, with
//								"*" for any set bit and "." for any zero bit.
//								The argument after this is a string pointing
//								to the column header.

//		'!'						The next argument is an array of two bitmaps,
//								which should be displayed as one bitmap with
//								"P" if a bit from the first map is set, and
//								"p" if a bit from the second map is set, and
//								"." if neither bit is set.  There is nothing
//								special if both maps are set, that comes up
//								"P".  The argument after this is a string
//								pointing to the column header.

//	The column header strings should be <= 8 characters long.

//	Here's an example call:

//		VDumpPawnMaps("!1", argbmLoc, "Loc", &bmDoubled, "Doubled");

//	This would be somewhat different, but would work:

//		VDumpPawnMaps("21", argbmLoc, "Loc", &bmDoubled, "Doubled");

//	There's an example line somewhere in the code below, unless I accidentally
//	deleted it.  You can comment it out and mess with it, if you decide you
//	want to add more pawn bitmaps and dump them.

//	I'm glad I wrote this routine, rather than just assuming that my code
//	worked.  It found a bunch of bugs.



void VDumpPawnMaps(char * szFmt, ...)
{
	va_list	pvaArg;
	char * sz;

	//	Display headers and figure out how many bitmaps in the largest array.
	//
	int cbmMax = 0;
	va_start(pvaArg, szFmt);
	for (sz = szFmt; *sz != '\0'; sz++) {
		char *	szArg;
		bitboard *pbmArg;
		int	cbm;

		if (isdigit(*sz))
			cbm = *sz - '0';
		else {
			Assert(*sz == '!');
			cbm = 1;
		}
		if (cbm > cbmMax)
			cbmMax = cbm;					// <-- New largest.
		pbmArg = va_arg(pvaArg, bitboard *);		// Pull bitmap pointer.
		szArg = va_arg(pvaArg, char *);		// Pull column header.
		if (sz != szFmt)
			putchar(' ');
		printf("%-8s", szArg);				// Display arg.
	}
	putchar('\n');
	//
	//	Display the bitmaps.  The largest argument I got was "cbmMax", so I
	//	will loop that many times, displaying one row of bitmaps each time.
	//
	for (int ibm = 0; ibm < cbmMax; ibm++) {

		if (ibm != 0)
			putchar('\n');
		//
		//	Display a row of bitmaps.
		//
		for (int rnk = rnk8; rnk >= rnk1; rnk--) {
			//
			//	Display one rank from all the bitmaps.
			//
			va_start(pvaArg, szFmt);
			for (sz = szFmt; *sz != '\0'; sz++) {
				char *	szArg;

				if (sz != szFmt)
					putchar(' ');
				int cbm = (isdigit(*sz)) ? *sz - '0' : 1;
				bitboard *pbmArg = va_arg(pvaArg, bitboard *);
				szArg = va_arg(pvaArg, char *);
				if (ibm >= cbm)			// This sometimes leaves trailing
					printf("        ");	//  spaces, but it's not worth fixing.
				else		// Display one rank from one bitmap.
					for (int fil = filA; fil <= filH; fil++) {
						bitboard bm;

						int isq = Idx(rnk, fil);
						bm = IdxToU64(isq);
						if (*sz != '!')
							putchar((bm & pbmArg[ibm]) ? '*' : '.');
						else if (bm & pbmArg[coWHITE])
							putchar('P');
						else
							putchar((bm & pbmArg[coBLACK]) ? 'p' : '.');
					}
			}
			putchar('\n');
		}
	}
}
#endif

static const int pawnDir[coMAX] = {16,-16};

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Penalties for pawn defects.

#define	valDOUBLED_ISO	20	// Doubled isolated pawn.
#define	valDOUBLED		10	// Normal doubled pawn, not strictly isolated.
#define	valISOLATED		15	// Strictly isolated pawn (no pawns of this color
							//  on either adjacent file).
#define valISOLATED_OPEN 15 // isolated pawn on semi-open file (*added* to other penalities)

#define valBACKWARD     5
#define valBACKWARDweak 5

// Bonus for pawns

#define valCONNECTED    3   // this will be counted twice (once for each pawn)

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
static int tempTable[64];

// SetDistance/SetNextDistance is used to compute a path for a king from a square to another square
// The distance between a square and the destination square is stored in a square array
// so the distance can be known for each square of the board with a simple table lookup
//
void setNextDistance(int rank, int file, const CON *pcon, HASHPawn *phashp, int side, int distance) {
    // check if on board
    if( rank < 0 || rank > 7 || file < 0 || file > 7 )
        return;
    int from = Idx(rank, file);
    if( distance ) {
        // exit if square is unavailable (pawn or attacked by a pawn)
        if( (pcon->pos[from]|1) == (PAWN|1) )
            return;
        bitboard thisSqr = IdxToU64(from);
        if( phashp->side[side^1].pawnatk & thisSqr )
            return;
    }
    // update distance (depending on path taken)
    // exit if a shorter was leading to this square
    if( distance >= tempTable[from] )
        return;
    tempTable[from] = distance;
    if( distance < 15 )
        distance ++;
    setNextDistance(rank+1, file+0, pcon, phashp, side, distance);
    setNextDistance(rank-1, file+0, pcon, phashp, side, distance);
    setNextDistance(rank+0, file+1, pcon, phashp, side, distance);
    setNextDistance(rank+0, file-1, pcon, phashp, side, distance);
    setNextDistance(rank+1, file+1, pcon, phashp, side, distance);
    setNextDistance(rank+1, file-1, pcon, phashp, side, distance);
    setNextDistance(rank-1, file+1, pcon, phashp, side, distance);
    setNextDistance(rank-1, file-1, pcon, phashp, side, distance);
}
void setDistance(const CON *pcon, HASHPawn *phashp, int from, squareDist *keySq, int winningSide) {
    keySq->square = from;
    keySq->winningSide = winningSide;
    for(int side = 0; side <= 1; side ++) {
        for(int i = 0; i <= 63; i++)
            tempTable[i] = 15;
        setNextDistance(Rank(from), File(from), pcon, phashp, side, 0);
        memcpy(keySq->distance[side], tempTable, sizeof(keySq->distance[side]));
#ifdef DEBUG
        printf("\n");
	    for (int rnk = rnk8; rnk >= rnk1; rnk--) {
		    for (int fil = filA; fil <= filH; fil++) {
			    int isq = Idx(rnk, fil);
                if( tempTable[isq] == 15)
                    printf(" -");
                else if( tempTable[isq] )
			        printf(" %d", tempTable[isq]);
                else
                    printf(" o");
		    }
		    printf("\n");
	    }
#endif
    }
}

// from Steffan Westcott / Gerd Isenberg
// http://chessprogramming.wikispaces.com/All+Shortest+Paths
/////////////////////////////////////////////////////////////////////////
//
// Returns length of shortest path of set bits present in 'path' that
// 8-way connect any set bit in sq1 to any set bit of sq2. 0 is returned
// if no such path exists. Also fills a sequence of bitboards asp[length]
// that describes all such shortest paths.
 
int allShortestPaths(U64 sq1, U64 sq2, U64 path, U64* asp)
{
    // Do an 8-way flood fill with sq1, masking off bits not in path and
    // storing the fill frontier at every step. Stop when fill reaches
    // any set bit in sq2 or quit if fill cannot progress any further.
    // Then do 8-way flood fill from reached bits in sq2, ANDing the
    // frontiers with those from the first fill in reverse order.
 
    if (!(sq1 &= path) || !(sq2 &= path)) return 0;
                        // Drop bits not in path
                        // Early exit if sq1 or sq2 not on any path
    int i = 1;
    asp[0] = sq1;
 
    while(1)  // Fill from all set bits in sq1, to any set bit in sq2
    {
        if (sq1 & sq2) break;                     // Found good path
        U64 temp = sq1;
        sq1 |= eastOne(sq1) | westOne(sq1);       // Set all 8 neighbours
        sq1 |= soutOne(sq1) | nortOne(sq1);
        sq1 &= path;                              // Drop bits not in path
        if (sq1 == temp) return 0;                // Fill has stopped
        asp[i++] = sq1 & ~temp;                   // Store fill frontier
    }
 
    int length = i;                               // Remember path length
    asp[--i] = (sq2 &= sq1);                      // Drop unreached bits
 
    while(i)  // Fill from reached bits in sq2
    {
        U64 temp = sq2;
        sq2 |= eastOne(sq2) | westOne(sq2);       // Set all 8 neighbours
        sq2 |= soutOne(sq2) | nortOne(sq2);
        sq2 &= path;                              // Drop bits not in path
        asp[--i] &= sq2 & ~temp;                  // Intersect frontiers
    }
    return length;
}

// This function returns the length of the shortest path, not the path.
// This is for pawn endgames.
// The initial path should contain the whole board minus occupied squares, minus attacked squares
// also it is not called for pawns defended by other pawns (sq2 is not in the path in this case)
int pathLength(U64 sq1, U64 sq2, U64 path)
{
    // Do an 8-way flood fill with sq1, masking off bits not in path and
    // storing the fill frontier at every step. Stop when fill reaches
    // any set bit in sq2 or quit if fill cannot progress any further.
    // Then do 8-way flood fill from reached bits in sq2, ANDing the
    // frontiers with those from the first fill in reverse order.

    if (!(sq1 &= path) || !(sq2 &= path)) return 99;
                        // Drop bits not in path
                        // Early exit if sq1 or sq2 not on any path
    int i = 0; //1;

    while(1)  // Fill from all set bits in sq1, to any set bit in sq2
    {
        if (sq1 & sq2) break;                     // Found good path
        U64 temp = sq1;
        sq1 |= eastOne(sq1) | westOne(sq1);       // Set all 8 neighbours
        sq1 |= soutOne(sq1) | nortOne(sq1);
        sq1 &= path;                              // Drop bits not in path
        if (sq1 == temp) return 99;                // Fill has stopped
        i++;
    }
//    VDumpPawnMaps("1", &sq1, "path");
    int length = i;                               // Remember path length
    return length;
}

//	This gets the pawn structure info from the hash table if it's there,
//	otherwise it generates it.

//	In most cases, the stuff is in the hash table.  The hash table is big
//	enough, and the pawn structures don't change that much.

//	Since the pawn computation stuff isn't done that much, I don't take
//	special pains to make sure that it's fast.  It's better to be clear and
//	accurate.  It would be very possible to generate the defect values via
//	less complicated data structures, and fewer passes over the data, but
//	the excess time shoudn't matter.

//	On the other hand, the whole thing stays flexible so it should be easy
//	to add to.

//	This routine returns values from the perspective of the side to move,
//	and stores values in the hash table from white's perspective.

//	There are a lot of other pawn features that can be detected, but doubled
//	and isolated pawns should be good enough for now.  The trick is to play
//	solidly enough that the program doesn't drop pawns a few moves over its
//	horizon.

//	An obvious addition is *anything* that would help with the passed pawn
//	problem, which is always severe in a simple program.  The program could
//	also benefit from an understanding of backwardness and artificial
//	isolation.  Additionally, the values given for doubled and isolated pawns
//	could vary depending upon where the pawns are on the board, and the
//	condition of the opponent's pawns (a doubled isolated pawn on an open file
//	is worse than one on a closed file).

//	In a heavier implementation, some of the pawn maps might end up being
//	stored in the hash table itself, which is another level of complication.

static void ValPawns(const CON *pcon, const STE *pste)
{
	HASHPawn *phashp = &s_rghashp[pste->hashkPn & s_chashpMac];

	int	argval[coMAX];				// Total value of pawn features.
	bitboard argbmLoc[coMAX];			// Bits set for pawns that exist.
	BM	bmDoubled;					// Bit set indicates doubled pawn.
	BM	bmIsolated;					// Bit set indicates isolated pawn.
    bitboard bmInfluence[coMAX] = {0,0};
    bitboard    blocked[coMAX];

#ifdef KEYSQUARE
    phashp->keySquare[0].square = 0;
    phashp->keySquare[2].square = 0;
    phashp->keySquare[3].square = 0;
#endif
    argbmLoc[W] = pcon->pieces[WP];
    argbmLoc[B] = pcon->pieces[BP];
    phashp->pawnon6or7 = false;
    phashp->c_center_pawns = popCount( (pcon->pieces[WP] | pcon->pieces[BP]) & centerMask);

	bmDoubled.qw = 0;
	bmIsolated.qw = 0;
	for (int coUs = coBLACK; coUs <= coWHITE; coUs++) {
        phashp->side[coUs].c_doubled = 0;
        phashp->side[coUs].c_islands = 0;
        blocked[coUs]                = 0;
        phashp->side[coUs].pawnatk   = 0;
        phashp->side[coUs].candidate = 0;
        phashp->side[coUs].c_backward= 0;
        phashp->side[coUs].c_connected=0;

        // count the number of pawns on this file
        for(int f = 0; f < 8 ; f++)
            phashp->side[coUs].c_pawn[f] = 0;

		argval[coUs] = 0;
        // build the influence bitmap, squares that are attacked by pawns
        // on their position or if they move up/down
        // note that pawns can move from one file to another so this computation
        // can be wrong in some situations
	    for (bitboard pawns = pcon->pieces[PAWN|coUs]; pawns; ) {
		    int sqr = PopLSB(&pawns);
            phashp->side[coUs].pawnatk |= pawn_attacks[coUs][sqr];
            int rank = Rank(sqr);
            int dir = coUs == W ? 8 : -8;
            if( coUs == W ) {
                if( rank == 5 || rank == 6 )
                    phashp->pawnon6or7 = true;
            } else {
                if( rank == 2 || rank == 1 )
                    phashp->pawnon6or7 = true;
            }
            bitboard xpawns = pcon->pieces[PAWN|(coUs^1)];
            int guard = 0;
            for(int sq = sqr; sq < 64 && sq >= 0; sq += dir ) {
                guard += popCount( pawn_attacks[coUs][sq] & xpawns );
            	bmInfluence[coUs] |= pawn_attacks[coUs][sq];
                if( xpawns & IdxToU64(sq) ) {   // there is an opponent pawn
                    blocked[coUs] |= IdxToU64(sqr);
                    guard = 10;
                }
            }
            if( guard <= 1 )
                phashp->side[coUs].candidate |= IdxToU64(sqr);
            phashp->side[coUs].c_pawn[File(sqr)]++;
        }

        phashp->side[coUs].location = argbmLoc[coUs];
		//	Collect doubled pawn info.  I'm see if there is more than one
		//	pawn of a certain color on the a-file, if there is I will OR
		//	all of those pawns into the "bmDoubled" bitmap, then I'll try
		//	the b-file, and so on.
		//
		bitboard bmI = (U64)0x0101010101010101;	// Start with a1...a8.
		for (;;) {
			//	Get all pawns of this color on this file.
			//
			bitboard bmS = argbmLoc[coUs] & bmI;
			//
			//	If more than one bit set in this map, all the pawns of
			//	this color on this file are doubled.
			//
			if (bmS & (bmS - 1))
				bmDoubled.qw |= bmS;
			//
			//	I'm going to make a map of all the pawns on files adjacent
			//	to this file.  If there are no pawns in that map, all the
			//	pawns on this file are isolated.
			//
			bitboard bmT = 0;
			if (!(bmI & 0x01))
				bmT |= bmI >> 1;
			if (!(bmI & 0x80))
				bmT |= bmI << 1;
			if (!(bmT & argbmLoc[coUs]))
				bmIsolated.qw |= bmS;
			//
			//	Set up to examine the next file.  0x80 is the h-file, so
			//	if we just examined it, bomb out of the loop, otherwise
			//	shift to the next file.
			//
			if (bmI & 0x80)
				break;
			bmI += bmI;	// Examine file one to the right.
		}
	}
    phashp->side[W].influence = bmInfluence[W];
    phashp->side[B].influence = bmInfluence[B];

	for (int coUs = coBLACK; coUs <= coWHITE; coUs++) {
        phashp->side[coUs].passed = phashp->side[coUs].location & 
            (~blocked[coUs]) & (~phashp->side[coUs ^ 1].influence) ;
        // control of the center
        argval[coUs] += 5 * popCount(phashp->side[coUs].pawnatk & centerMask);
        // for each pawn of the board...
	    for (bitboard pawns = pcon->pieces[PAWN|coUs]; pawns; ) {
		    int sqr = PopLSB(&pawns);
            int nSqr = remapSq[coUs][sqr];
			bitboard bmLoc;
			bmLoc = IdxToU64(sqr);
            int f = File(sqr);
			if (bmLoc & bmIsolated.qw) {
                if( phashp->side[coUs^1].c_pawn[f] == 0 )
                    argval[coUs] -= valISOLATED_OPEN;
				if (bmLoc & bmDoubled.qw)
					argval[coUs] -= valDOUBLED_ISO;
				else
					argval[coUs] -= valISOLATED;
            } else if (bmLoc & bmDoubled.qw) {
					argval[coUs] -= valDOUBLED;
            } else {
                // non isolated pawn on an open file
                if( phashp->side[coUs^1].c_pawn[f] == 0 ) {
                    int dir = coUs == W ? -1 : 1;
                    int r = Rank(sqr);
                    bool hasFriends = false;
                    while(r > 0 && r < 7 ) {
                        if( f > 0 && pcon->pos[Isq64FromRnkFil(r,f - 1)] == (PAWN|coUs)) {
                            hasFriends = true;
                            break;
                        }
                        if( f < 7 && pcon->pos[Isq64FromRnkFil(r,f + 1)] == (PAWN|coUs)) {
                            hasFriends = true;
                            break;
                        }
                        r += dir;
                    }
                    if( hasFriends ) {
                        if( r == Rank(sqr) ) {
                            phashp->side[coUs].c_connected++;
                            argval[coUs] += valCONNECTED;
                        }
                    } else {
                        // if there is no friendly pawn behind this one then it's backward
                        // c_backward is used for debug only
                        phashp->side[coUs].c_backward ++;
                        argval[coUs] -= valBACKWARD;
                        dir = -dir;
                        r = Rank(sqr) + dir;
                        while(r > 0 && r < 7 ) {
                            // check for ennemy pawn
                            if( f > 0 && pcon->pos[Isq64FromRnkFil(r + dir,f - 1)] == (PAWN|coUs^1)) {
                                argval[coUs] -= valBACKWARDweak;
                                phashp->side[coUs].c_backward += 10;
                                break;
                            }
                            if( f < 7 && pcon->pos[Isq64FromRnkFil(r + dir,f + 1)] == (PAWN|coUs^1)) {
                                argval[coUs] -= valBACKWARDweak;
                                phashp->side[coUs].c_backward += 10;
                                break;
                            }
                            if( f > 0 && pcon->pos[Isq64FromRnkFil(r,f - 1)] == (PAWN|coUs))
                                break;
                            if( f < 7 && pcon->pos[Isq64FromRnkFil(r,f + 1)] == (PAWN|coUs))
                                break;
                            r += dir;
                        }
                    }
                }
            }
		}
	}
#ifdef KEYSQUARE
//    setDistance(pcon, phashp, G5, &phashp->keySquare[0], W);
//    setDistance(pcon, phashp, B5, &phashp->keySquare[1], W);
    setDistance(pcon, phashp, A5, &phashp->keySquare[0], W);
    setDistance(pcon, phashp, F5, &phashp->keySquare[1], W);
#endif

#ifdef _DEBUG
    static int dump_count = 0;
    dump_count++;
    if( dump_count <= 1 ) {

//        bitboard blocked[2] = {phashp->side[0].blocked,phashp->side[1].blocked};
        bitboard passed[2] = {phashp->side[0].passed, phashp->side[1].passed };
        VDumpPawnMaps("!111111",
            argbmLoc, "Loc",
            &bmDoubled.qw, "Doubled",
            &bmIsolated.qw, "Isolated",
            &phashp->side[W].influence, "WInfl.",
            &phashp->side[B].influence, "BInfl.",
//            &blocked, "blocked",
            &phashp->side[W].pawnatk, "WAtk.",
            &phashp->side[B].pawnatk, "BAtk."
        );
    }
#endif
    phashp->hashk = pste->hashkPn;
    // keep a score for each side
    phashp->side[W].val = argval[coWHITE];
    phashp->side[B].val = argval[coBLACK];

    // now we know if a pawn is
    // - isolated, doubled, blocked/candidate, passed
    // we also have the count of pawn per file
    // we can do rook_file_attack & location(xside) to know if a rook is on a semi open file
}

void EvalPawns(const CON *pcon, const STE *pste, PHASHPawn & phashp) {
	phashp = &s_rghashp[pste->hashkPn & s_chashpMac];

	if (phashp->hashk != pste->hashkPn) {
        ValPawns(pcon, pste);
    }
}


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Zero all pawn hash memory.

void VClearHashp(void)
{
    if( s_chashpMac )
	    memset(s_rghashp, 0, (s_chashpMac + 1) * sizeof(HASHPawn));
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This allocates the pawn hash memory according to the max value passed in
//	by the user, then clears the table.

bool FInitHashp(PCON pcon, int cbMaxPawnHash)
{
    if( cbMaxPawnHash <= 0 || cbMaxPawnHash > 256*1024*1024 )
        cbMaxPawnHash = 1*1024*1024;

    int chashpMax = 1;
	for (;;) {
		if (chashpMax * 2 * (int)sizeof(HASHPawn) > cbMaxPawnHash)
			break;
		chashpMax *= 2;
	}
    int Hsize = chashpMax * sizeof(HASHPawn);
	VPrSendComment("%d Kbytes pawn hash memory (%d entries of %d bytes)", Hsize/1024, chashpMax, sizeof(HASHPawn));
    s_rghashp = (HASHPawn *) malloc(Hsize);
	if ( s_rghashp == NULL) {
		VPrSendComment("Can't allocate pawn hash memory: %d bytes", Hsize);
        chashpMax = 1024;
        Hsize = chashpMax * sizeof(HASHPawn);
	    VPrSendComment("%d Kbytes pawn hash memory", Hsize/1024);
        s_rghashp = (HASHPawn *) malloc(Hsize);
	    if ( s_rghashp == NULL) {
		    VPrSendComment("Can't allocate pawn hash memory: %d bytes", Hsize);
		    return false;
        }
	}
	s_chashpMac = chashpMax - 1;
	VClearHashp();
	return true;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-
