//
// Cyrano Chess engine
//
// Copyright (C) 2007  Harald JOHNSEN
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
//
//

#include "magicmoves.h"
#include "engine.hpp"
#include "gproto.hpp"
#include "hash.hpp"
#include "board.hpp"
#include "moves.hpp"
#include "stats.hpp"
#include "search.hpp"
#include "eval.hpp"
#include "genmagic.hpp"
#include <stdlib.h>

#ifdef	_DEBUG
static char const s_aszModule[] = __FILE__;
#endif

// enables the PVS search, undefine to have standard alpha beta search (for bug hunting)
// reduce the number of nodes
#define NULL_WINDOW

// reduce the number of nodes
#define NULL_MOVE

// allways extend depth by one ply when in check
// increase the number of nodes, find mates in less plies
#define EXTEND_CHECK

// reduce the number of nodes
#define USE_HASH

// change move ordering => reduce the number of nodes
#define USE_KILLERS

#define QS_BEFORE_HASH      // note:probing hash first gives worse results

// MULTI_CUT is buggy here, I'll check that later
//#define MULTI_CUT

// while it appears to work, fhr can not be implemented so simply
// I disable it for now
#define FHR_THREAT  (300)
//#define FHR

// late move reduction : ok it's not so bad even with my poor implementation
#define USE_LMR // disabled to test things

// change move ordering
#define HISTORY

// skip descending the tree for some useless xchange in the quiescent search
// reduce the number of nodes ?
#define PRUNE_QS

// ...
#define FUTILITY_PRUNING

// this will stop search when we have a score good enought (mate)
#define DISTANCE_PRUNING

#define RECAPTURE_EXTENSION

// does not work, need more testing one day
//#define KING_ATTACK

//#define MOVE_SAFETY_SORT  // this is not so bad
//#define MOVE_PST_SORT     // this is not so good
//#define MOVE_FILTER       // not implemented, first tests were not good (don't know why)

// this was supposed to reduce the depth of search in qs but it produces less cutoffs
// so i get more nodes at the end (more tests needed later)
//#define SEE_IN_QS

#define USE_IID

#define moveisCapture(pcm) (Capt((pcm)->m) != 0)
//#define moveisGoodCapture(pcm) (Capt(pcm->m) != 0)
#define moveisGoodCapture(pcm) (((pcm)->cmk & (cmkGOODCAPT|cmkEQUCAPT)) != 0)
#define moveisPromotion(pcm) ((Pc((pcm)->m)|1) == (B_PROM|1))
#define moveisPawnMove(pcm) ((Pc((pcm)->m)|1) == (PAWN|1))
#define moveto7thRank(pcm)  ((Rank(To((pcm)->m)) == 6) || (Rank(To((pcm)->m)) == 1))


static int SearchQ(PCON pcon, PSTE pste, int Alpha, int Beta);
static int Search(PCON pcon, PSTE pste, int Alpha, int Beta, nodeType nt);
static bool inline capture_is_good(PCON pcon, const STE *pste, const CM *pcm);


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

static void SortKillers(PSTE pste, int depth)
{
    U32 Killer1 = Killers[depth].Killer1.m;
    U32 Killer2 = Killers[depth].Killer2.m;
    U32 Killer3 = Killers[depth-2].Killer1.m;
    U32 Killer4 = Killers[depth-2].Killer2.m;
    foreachmove(pste, pcm)
        if( (pcm->m == MateKillers[depth].Killer1.m) || (pcm->m == MateKillers[depth].Killer2.m)) {
			pcm->cmk |= cmkKILLER | 5;
            stats.Killers++;
        } else if ( (pcm->m == Killer1) || (pcm->m == Killer2) ) {
			pcm->cmk |= cmkKILLER | 3;
            stats.Killers++;
		} else if ( (pcm->m == Killer3) || (pcm->m == Killer4) ) {
			pcm->cmk |= cmkKILLER | 1;
            stats.Killers++;
		}
}

static void SortKingAttack(PSTE pste)
{
    foreachmove(pste, pcm) {
    }
}

//	This will be used to figure out which move to search first.  It is very
//	important to try to search the best move first, since this will produce
//	a huge reduction in tree size.

//	Best moves are typically winning captures and moves that have been shown
//	to be good here in the past (PV move and hash table move).

static void PrepareSort(CON *pcon, const STE *pste, int depth, PHASHPawn phashp) {
    // bitmap of squares attacked by the opponent pawns
    bitboard notSafe = phashp->side[pste->side ^1 ].pawnatk;
    U32 lm = 0;
    if(depth >= 2)
        lm = (pste-1)->lastMove;

    foreachmove(pste, pcm) {
        pcm->cmk = 0;
        if(Capt(pcm->m)) {
            pcm->cmk = MVV_LVA[Pc(pcm->m)][Capt(pcm->m)];
            if( pcm->cmk & cmkBADCAPT ) {
                // RxN, QxP
                if( capture_is_good(pcon, pste, pcm ) ) {
                    pcm->cmk = cmkGOODCAPT | 2000;
                }
            }
        }
        else if( moveisPawnMove(pcm) && moveto7thRank(pcm)) {
            // only checking the dest rank because we know it's a pawn
            pcm->cmk = (cmkKILLER | cmkGOODCAPT) + 10000;
        } else if( moveisPromotion(pcm) ) {
            pcm->cmk = cmkQUEEN | (100 - Prom(pcm->m));
        } else {
            // having 'bad' captures after non capture moves gives more nodes
            // ie having non captures after 'bad' captures
//            pcm->cmk = cmkNONE;
#ifdef HISTORY
            pcm->cmk |= History.sort_key( pcm );
#endif

#ifdef KEYSQUARE
            // perhaps one day...
            // this is for the move ordering of pawn endgames with corresponding/key squares
            if( (pste->valPcThem + pste->valPcUs == 0) && ((Pc(pcm->m)|1) == (KING|1)) ) {
                PHASHPawn phashp;
                EvalPawns(pcon, pste, phashp);
                int side = pste->side;
                int to = To(pcm->m);
                for(int i = 0; i < 3 ; i ++ )
                    if( phashp->keySquare[i].square ) {
                        int winner = phashp->keySquare[i].winningSide;
                        pcm->cmk += 50 - phashp->keySquare[i].distance[side][ to ] + phashp->keySquare[i].distance[side][ pcon->king[side] ];
                        if( side == winner ) {
                            // dist is the number of moves needed to go on the key square
                            int dist = phashp->keySquare[i].distance[winner^1][ pcon->king[winner^1] ] - 
                                phashp->keySquare[i].distance[winner][ to ];
                            // side is side to move next
                            // if we can go in less moves than the oppopent then it's a win
                            if( dist >= 1 )
                                pcm->cmk |= cmkCAPTURE + dist * 10;
                        }
                    }
            }
#endif
        }
#ifdef MOVE_SAFETY_SORT
        bitboard to = IdxToU64(To(pcm->m));
        // keep key positive
        pcm->cmk += 100;
        // moving to a square attacked by a pawn
        if( notSafe & to )
            pcm->cmk -= 20;
        bitboard from = IdxToU64(From(pcm->m));
        // leaving a square attacked by a pawn
        if( notSafe & from )
            pcm->cmk += 20;

//        if( Attacked(pcon, To(pcm->m), pste->side ^ 1 ) )
//            pcm->cmk -= 10;

#ifdef MOVE_FILTER
        // not implemented, ideas are
        //  malus is moving the same piece twice in a row, moving back to the last position
        //  bonus if the move can be done because another piece moved from To(m) or was
        //  between From(m) & To(m)
        if(lm) {
            if( Pc(pcm->m) == Pc(lm) ) {
                //pcm->cmk -= 4;
                if( To(pcm->m) == From(lm) )
                    pcm->cmk -= 8;
            } else {
            }
        }
#endif

#ifdef MOVE_PST_SORT
        // set cmk to delta positional value
        int pc = Pc(pcm->m);
        int neutralPc = pc & (~1);
        int co = pc & 1;
        // neutral color square for positional values
        int nSqr = remapSq[co][To(pcm->m)];
        // basic positional value
        pcm->cmk += staticPosVal[neutralPc >> 1][nSqr];
        nSqr = remapSq[co][From(pcm->m)];
        // basic positional value
        pcm->cmk -= staticPosVal[neutralPc >> 1][nSqr];
#endif
#endif
    }
}

static void PrepareSortQS(PCON pcon, const STE *pste, int depth) {
    foreachmove(pste, pcm) {
        U32 m = pcm->m;
        pcm->cmk = 0;
        if(Capt(m)) {
            pcm->cmk = MVV_LVA[Pc(m)][Capt(m)];
            if( pcm->cmk & cmkBADCAPT ) {
                // RxN, QxP
//                if( capture_is_good(pcon, pste, pcm ) ) {
//                    pcm->cmk = cmkGOODCAPT | 2000;
//                }
            }
        }
    }
}

static inline void SortNext(const STE *pste, PCM pcm) {
	PCM	pcmBest = pcm;

	for (PCM pcmI = pcm + 1; pcmI < (pste + 1)->mlist_start ; pcmI++)
		if (pcmI->cmk > pcmBest->cmk)
			pcmBest = pcmI;
	CM cm = *pcm;
	*pcm = *pcmBest;
	*pcmBest = cm;
}

static inline int MaterialValue(const STE *pste) {
    return (pste->valPcUs + pste->valPnUs) - (pste->valPcThem + pste->valPnThem);
}

static inline int MaterialGain(const CON *pcon, const CM *pcm) {
    // Capt(move) == 0 => no capture, array return 0 in this case
    return b_valPc[ Capt(pcm->m) ];
}

static inline int recapMaterialValue(const CON *pcon, const CM *pcm) {
    // Capt(move) == 0 => no capture, array return 0 in this case
    return exchangePieceValue[ Capt(pcm->m) ];
}

static inline void copyPV(PSTE pste, const CM &cm) {
    int how_many = (pste + 1)->ccmPv;
    if( how_many > 0 )
        memcpy(pste->argcmPv + 1, (pste + 1)->argcmPv, how_many * sizeof(CM));
    pste->argcmPv[0] = cm;
    pste->ccmPv = how_many + 1;
}
static inline void copyPVQS(PSTE pste, const CM &cm) {
#if 0
    int how_many = (pste + 1)->ccmPv;
    if( how_many > 0 )
        memcpy(pste->argcmPv + 1, (pste + 1)->argcmPv, how_many * sizeof(CM));
    pste->argcmPv[0] = cm;
    pste->argcmPv[0].cmk = cmkHASH | 123456;
    pste->ccmPv = how_many + 1;
#endif
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	This does a null-move search to see if the side not to move has a threat.
//	If they can't drive the score above alpha, even if they get to move twice
//	in a row, their positions probably sucks, and we'll prune this variation.
//	It's not called in endgames, because it has problems detecting zugzwang.

//  We don't need to do a null move if we know allready that there is threats around
//  In endgames any pawn move is a potential threat (score will vary a lot)

static int FThreat(PCON pcon, PSTE pste, int Alpha, int Beta, const nodeType nt)
{
//	if ((pste->plyRem <= 0) || (pste->fNull))
//		return true;
	bNullMove(pcon, pste);
    // Reduced-depth search
    int R = 2;
    // we can reduce the depth further when there is lot of material on the board
    // this does not work, I get a lot more nodes
    // R = 3 if (rem_ply > 8) || (rem_ply > 6 (and pieces per side >= 3))
#if 1
    if( (pste->plyRem >= 5) || (pste->valPcThem + pste->valPcUs >= 6*valKNIGHT) 
                                && (pste->plyRem >= 4) ) {
        R = 3;
    }
#endif
//    R = 3; // set to 3 since we do a verification search
    (pste + 1)->plyRem = pste->plyRem - 1 - R;
    int val = -Search(pcon, pste + 1, - Beta, - Beta + 1, nt);
	return val;
}


//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Quiescent search.  This lets captures play out at the end of the search,
//	so I don't end up aborting a line at a stupid place.

//	The quiescent search doesn't see checks, so it will not detect checkmates,
//	and might actually end up making decisions based upon an illegal line.

//	This is one reason I don't record the PV while in quiescent search -- I
//	don't want people telling me the program is doing illegal stuff.

// 200 gives better results than nothing or 100 or 150 (on wac-141)
//#define QS_MARGIN   200//100
//#define QS_MARGIN   100
//#define QS_MARGIN   125
//#define QS_MARGIN   200
#define QS_MARGIN   75

static bool inline capture_is_good(PCON pcon, const STE *pste, const CM *pcm) {
    int pc = Pc(pcm->m);
    int took = Capt(pcm->m);
    if( (pc|1) == (KING|1) )
        return true;
    if( moveisPromotion( pcm ) )
        return true;
    if( exchangePieceValue[ took ] >= exchangePieceValue[ pc ] )
        return true;
    // en passant (note that this is allready handled by the line above)
    // because took == 0 == black king value
    if( (pc|1) == (B_EP|1) )
        return true;

    // compute see value...
    int val = see(pcon, pste, From(pcm->m), To(pcm->m));
    return val >= 0;
}

static char sz_score[128];

static int SearchQ(PCON pcon, PSTE pste, int Alpha, int Beta)
{
	pcon->ss.nodes++;
    stats.QS_count ++;
	CM cmBest(0,0);
	pste->ccmPv = 0;
	bool fFound = false;
    bool reply_to_check = false;
    bool isPV = Beta != (Alpha+1);
    int iaminCheck = pste->checkf;
    int  f_score = 0;
    int  f_max   = valMAX;
    CM * p_moves;
	//
	//	See if static eval will cause a cutoff or raise alpha.
	//
    // this will be the return value if king evasion does no capture
	int posVal = ValEval(pcon, pste, Alpha, Beta);
    // score is valMIN when in check or else it won't play moves that lead to loss of material when nearly mated
    int bestVal = valMIN; //posVal;


#ifdef PRUNE_QS
    if( !iaminCheck ) {
        f_score = posVal + QS_MARGIN ;
        f_max   = f_score;
    }
#endif
    if( !iaminCheck ) {
        bestVal = posVal;
	    if (bestVal >= Beta)
		    return bestVal;
        if (bestVal > Alpha)
            Alpha = bestVal;
//        bestVal = Alpha;      // WTH ???
    }
    // generate captures, if we are in check then generate all moves (evade)
	if (iaminCheck) {
		bGenEvasions(pcon, pste);
#if 1
        if( legalMoveCount( pste ) == 1 )
            reply_to_check = true;
        else
            bestVal = posVal;
#endif
	}
	else {
		p_moves = bGenCaptures(pcon, pste, pste->mlist_start);
	}

    int depth = int(pste - pcon->argste) + 1;
    if( depth > pcon->ss.plyMaxDepth )
        pcon->ss.plyMaxDepth = depth;
    PrepareSortQS( pcon, pste, depth );
    //
	//	Iterate through capturing moves, to see if I can improve upon alpha
	//	(which may be the static eval).
	//

    foreachmove(pste, pcm) {

        SortNext(pste, pcm);
        fFound = true;
#ifdef SEE_IN_QS
        bool good_capture = iaminCheck || capture_is_good(pcon, pste, pcm);
#endif
        bMove(pcon, pste, pcm->m);
#if 1
        // we've generated all moves if in check, but we wont play them
        // we just look if we have a legal move
        if( iaminCheck && !reply_to_check && (Capt(pcm->m) == 0) ) {
            (pste+1)->ccmPv = 0;
		    bUndoMove(pcon, pste, pcm->m);
            continue;
        }
#endif
#ifdef SEE_IN_QS
        if( !good_capture && (pste[1].checkf == 0) ) {
            // allways do a true eval if we give check
            // doing a check can change things a lot in an exchange
            // an apparently bad capture can be good at the end
            // (each side has an hanging piece, one hanging piece takes a pawn and give check for example)
            // the bad thing is that now we must do/undo the move to have the check flag
		    bUndoMove(pcon, pste, pcm->m);
            continue;
        }
#endif
#ifdef PRUNE_QS
        if( !iaminCheck && (pste[1].checkf == 0 || (pste->plyRem != -1) )
//            && (!isPV)
            && !moveisPromotion(pcm) && !moveisPawnMove(pcm)
            && (f_max + MaterialGain(pcon, pcm) <= Alpha)
            ) {
                stats.FP_QS++;
    		    bUndoMove(pcon, pste, pcm->m);
                // check that
                int approx = f_max + MaterialGain(pcon, pcm);
                if( approx > bestVal ) {
                    cmBest.m = 0;
	                pste->ccmPv = 0;
                    // is that needed ?
                    bestVal = approx;
                }
                continue;
        }
#endif

		(pste + 1)->plyRem = pste->plyRem - 1;
		int val = -SearchQ(pcon, pste + 1, -Beta, -Alpha);
		bUndoMove(pcon, pste, pcm->m);
#ifdef _DEBUG
        if( pcon->ss.plyDepth == 1 ) {
            sprintf(sz_score, " QS v=%d [%d,%d] ", val, Alpha, Beta);
            VDumpLine(pcon, pste, sz_score);
        }
#endif
        if (val >= Beta) {
            copyPVQS(pste, *pcm );
			return val;
        }
        if (val > bestVal) {
			bestVal = val;
            cmBest = *pcm;
            copyPVQS(pste, *pcm );
            if (val > Alpha)
			    Alpha = val;
        }
	}
    if( !iaminCheck && (pste->plyRem == -1) && (bestVal < (valKNIGHT/2))
        && (pste->valPcUs >= valROOK)) {
        // gen checks
        // this is of course not optimal
        bGenNonCaptures(pcon, pste, p_moves);

        foreachmove(pste, pcm) {

            fFound = true;
            bMove(pcon, pste, pcm->m);
            // we only want moves that do check
            if( (pste[1].checkf == 0) ) {
		        bUndoMove(pcon, pste, pcm->m);
                continue;
            }
		    (pste + 1)->plyRem = pste->plyRem - 1;
		    int val = -SearchQ(pcon, pste + 1, -Beta, -Alpha);
		    bUndoMove(pcon, pste, pcm->m);
            if (val >= Beta) {
                copyPVQS(pste, *pcm );
			    return val;
            }
            if (val > bestVal) {
			    bestVal = val;
                cmBest = *pcm;
                copyPVQS(pste, *pcm );
                if (val > Alpha)
			        Alpha = val;
            }
	    }
    }

	if (!fFound) {
        // We have no move !!
		if (iaminCheck) {
            // Checkmate
			bestVal = mateValue(depth);
		}
            // perhaps draw but we did not gen all moves
	}
	return bestVal;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

static int potentialKingAttack(PCON pcon, PSTE pste) {
    for(int side = B; side <= W; side ++) {

        int xside = side ^ 1, sqr;
        if( side != W )
            continue;
        if( side != pste->side  )
            continue;
        bitboard pc;
        bitboard valid_atk = (pcon->pieces[B_OCCU|xside]|pcon->pieces[X_OCCU]);
	    // - Knights -
        bitboard knight_atk = 0;
        bitboard rook_atk = 0;
        bitboard bishop_atk = 0;
        bitboard queen_atk = 0;
	    for (pc = pcon->pieces[KNIGHT|side]; pc; ) {
		    sqr = PopLSB(&pc);
		    knight_atk |= knight_attacks[sqr] & valid_atk;
	    }

	    // - Bishop -
	    for (pc = pcon->pieces[BISHOP|side]; pc; ) {
		    sqr = PopLSB(&pc);
		    bishop_atk |= Bmagic(sqr, pcon->pieces[OCCU]) & valid_atk;
	    }

	    // - Rook -
	    for (pc = pcon->pieces[ROOK|side]; pc; ) {
		    sqr = PopLSB(&pc);
		    rook_atk |= Rmagic(sqr, pcon->pieces[OCCU]) & valid_atk;
	    }

	    // - Queen -
	    for (pc = pcon->pieces[QUEEN|side]; pc; ) {
		    sqr = PopLSB(&pc);
		    queen_atk |= Qmagic(sqr, pcon->pieces[OCCU]) & valid_atk;
	    }
        bitboard king_space = king_attacks[pcon->king[xside]];
        bishop_atk  &= king_space;
        rook_atk    &= king_space;
        queen_atk   &= king_space;
        if( queen_atk & (bishop_atk | knight_atk | rook_atk) )
            return side;
        if( rook_atk & (bishop_atk | knight_atk | queen_atk) )
            return side;
        if( bishop_atk & (knight_atk | queen_atk | rook_atk) )
            return side;
    }

    return -1;
}

//	Full-width search.  This is the heart of the "chess playing" part of this
//	program.  It performs a recursive alpha-beta search.

#define LOWERBOUND  a
//#define LOWERBOUND  Alpha // ouch, so wrong

#define NCMove(from, to, pc) ((from) | ((to) << 6) | ((pc) << 12))
#define CMove(from, to, type, capt) ((from) | ((to) << 6) | ((type) << 12) | ((capt) << 16))

static int Search(PCON pcon, PSTE pste, int Alpha, int Beta, nodeType nt)
{
	int	val;
    pste->ccmPv = 0;
    pste->reduced = false;
	CM cmBest(0,0);
//    bool isPV = (nt == nodePV);
    bool king_attak = false;
    int king_attacker = -1;
    int preExtend = 0;
    int bestVal = valMIN;
    int iaminCheck = pste->checkf;
    bool nm_threat = false;

    if (false ) {
        if( pcon->argste[0].lastMove == NCMove(A8,A1,QUEEN|B)
            && pcon->argste[1].lastMove == CMove(E4,D4,ROOK|W,BISHOP|B)
            && pcon->argste[2].lastMove == NCMove(A1,G1,QUEEN|B)
            && pcon->argste[3].lastMove == NCMove(G3,F3,KING|W)
            && pcon->argste[4].lastMove == NCMove(G1,F1,QUEEN|B)
            && pcon->argste[5].lastMove == NCMove(F3,E3,KING|W)
            && pcon->argste[6].lastMove == CMove(F1,H3,QUEEN|B,PAWN|W)
            ) {
            // Qa1 in root
            sprintf(sz_score, " Qa1 ...");
            VDumpLine(pcon, pste, sz_score);
        }
    }
#ifdef QS_BEFORE_HASH
    if (pste->plyRem < 0) {
#ifdef KING_ATTACK
        if( !iaminCheck && extCount <= 2 && (king_attacker=potentialKingAttack(pcon, pste))>=0 ) {
            king_attak = true;
            // keep plyrem == 0 because of hash probe later
            pste->plyRem = 0;
            extCount += 4;
            if( king_attacker == pste->side )
                bestVal = SearchQ(pcon, pste,  Alpha, Beta);
//            preExtend = 1;
            // increment extCount so we don't do null move
//            Print(pcon, pste);
        } else
#endif
        return SearchQ(pcon, pste,  Alpha, Beta);
    }
#endif

    int depth = int(pste - pcon->argste) + 1;
    // extCount is the net number of extension done (all extensions - all reductions done)
    int extCount = depth - pcon->ss.plyDepth + pste->plyRem; // extCount < 0 => reduced or wanted depth not reached yet
    int maxExt = depth - extCount;//pcon->ss.plyDepth;
    maxExt = maxExt * 3 / 4;

    if( depth >= 30 ) {
        depth = depth;
    }
    if( depth >= (csteMAX-50) ) {
        return SearchQ(pcon, pste,  Alpha, Beta);
    }

	//	Increment global node counter, and figure out if it is time to call
	//	the interface callback, which is what allows for an interruptible
	//	search.  It's crucial that this be called now and then.
	//
	if (++pcon->ss.nodes >= pcon->ss.nodesNext) {
		//
		//	2000 nodes is maybe 1/50to 1/100 of a second.
		//
		pcon->ss.nodesNext = pcon->ss.nodes + 20000;
		VPrCallback(pcon);
		VCheckTime(pcon);
		if ((pcon->fAbort) || (pcon->fTimeout))
			return 0;
	}
	//
	//	I can return draws and cut off if I'm somewhere other than the
	//	first ply of depth.
	//
	//	Return draw score if this is a repeated node.
	//
	if (FRepSet(pcon, pste->hashkPc)) {
        stats.repCount++;
		return 0;
	}
	//	Check hash table, in order to get "best" moves, and try to cut
	//	off.  Don't cut off within two plies of the root, otherwise the
	//	program might allow stupid rep draws.
	//
    PHASH hash_entry = 0;
#ifdef USE_HASH
    hash_entry = ProbeHashValue(pcon, pste, &val, Alpha, Beta, depth);
    if( hash_entry && (nt != nodePV) && (hash_entry->plyRem >= pste->plyRem) ) {
        PHASH phash = hash_entry;
        int val = phash->val;
#if 0
        if( val >= valMATEscore ) {
            val -= depth - 1;
        } else if( val <= -valMATEscore ) {
            val += depth - 1;
        }
#endif
		switch(phash->getBound())
        {
			case hashfEXACT:
                stats.Hcutoff ++;
                return val;
                break;
			case hashfBETA:
				{
                    if( val >= Beta ) {
                        stats.Hcutoff ++;
						return val;
                    }
// this can not work as is
                    if( val > Alpha ) {
                        stats.Halpha ++;
//                        Alpha = val;
                    }

					break;
				}
			case hashfALPHA:
				{
                    if( val <= Alpha ) {
                        stats.Hcutoff ++;
						return val;
                    }
                    // the next lines are sometimes good, sometimes bad...
                    if( val < Beta) {
                        stats.Hbeta++;
//                        Beta = val;
                    }
					break;
				}
            default:
                Assert(0);
        }
	}
#endif

    if( hash_entry )
        nm_threat = hash_entry->getThreat();

#ifndef QS_BEFORE_HASH
    if (pste->plyRem < 0)
        return SearchQ(pcon, pste,  Alpha, Beta);
#endif

	//
	// Check alpha/beta window against a mate score
    // Thx for the idea
	//
#ifdef DISTANCE_PRUNING
#if 1
    // this is the lowest score I can get from here
    int i_am_mate_score = mateValue(depth + 1);
    if( i_am_mate_score > Alpha ) {
        Alpha = i_am_mate_score;
        // Beta is mate in n, best here is mate in n+m, so return
        if(i_am_mate_score >= Beta)
            return Beta; //i_am_mate_score;
//            return i_am_mate_score;
    }
#endif
#if 1
    // this is the best score we can get from here
    int i_will_mate_score =  -mateValue(depth);
    if( i_will_mate_score < Beta ) {
        Beta = i_will_mate_score;
        // if I set + 0 then I don't see the final move (
            // I will not see something better than Alpha so just return
        // Alpha is mate in n, best here is mate in n+m, so return
        if( i_will_mate_score <= Alpha )
            return Alpha;
//            return i_will_mate_score;
    }
#endif
#endif
	//	Mark this move so if I find it in a sub-tree I can detect a rep
	//	draw.
	//
//	VSetRep(pcon, pste->hashkPc);

    PHASHPawn phashp;
    EvalPawns(pcon, pste, phashp);

#ifdef NULL_MOVE
	//
	//	Do null-move search, which is a reduced depth that tries to
	//	figure out if the opponent has no threats.
	//
    // Note :
    // - if Beta is a mate score then it is not probable that I can do better without doing anything
    // - only done if not a PV (what about branches that will become a pv later ?)
    // - nm_threat caching does not work for an unknown reason (more nodes sometimes)
    // - material constraint reduced, I hope this works well with the verification search
    // - reductions are now allowed after an extension (so no test on extCount)
    // - if the stand pat score is < Beta then doing nothing will not raise it to Beta or more

    if ( // (extCount == 0) && 
        (!(nt == nodePV)) 
        && (!iaminCheck)
//        && (pste->plyRem >= 3) 
//        && (pste->plyRem >= 1)  // ok
//        && (pste->plyRem <= 3)    //=========
//        && (pste->plyRem <= 5)    //========= ok
//        && (pste->valPcUs >= valROOK)
//        && (pste->valPcThem >= valROOK)
        && (pste->valPcUs > 0)
        && (pste->valPcThem > 0)
        && (MaterialValue(pste) + valPAWN > Beta)
        && (Beta < valMATEscore)    // && (Beta > -valMATEscore)
//        && (!phashp->pawnon6or7)
//        && (!nm_threat)             // if we allready know that there is a threat...
        && (!pste->fNull)) {
        if( true ) {
            stats.NM1++;
            // compute successor node type since I'm only doing history counters on cut nodes
            nodeType succNType;
            if(nt == nodeCUT )
                succNType = nodeALL;
            else
                succNType = nodeCUT;

            int val = FThreat(pcon, pste, Alpha, Beta, succNType);
            if( val >= Beta ) {
                if( false ) {
                    U32 Qa8a1 = NCMove(A8,A1,QUEEN|B);
                    if( pcon->argste[0].lastMove == Qa8a1 ) {
                        // Qa1 in root
                        sprintf(sz_score, " Qa1 was prunned in NM v=%d beta=%d ", val, Beta);
                        VDumpLine(pcon, pste, sz_score);
                    }
                }
                // why not store result in hash ?
                // TODO check this: this does not work well on some position (more nodes)
                stats.NM2++;
                // verification search (for end game mostly)
                // still need to do NM at lower depth (higher remaining depth)
                const int Rv = 1 + 4;
                if(pste->plyRem >= Rv) {
                    pste->fNull = true; // set or else we will loop :)
                    pste->plyRem -= Rv;
                    val = Search(pcon, pste, Alpha, Beta, nodeCUT /*nt*/);
                    pste->plyRem += Rv;
                    if( val >= Beta ) {
                        // wait...that is allready in hash with a lower plyrem
                        // note that now we store the result with a bigger plyrem
//                        RecordHashValue(pcon, pste, cmBest, Beta, Alpha, Beta, depth, false);
                        stats.NM3++;
                        return Beta;
                    }
                } else {
                    // store because we don't want to do this same search *again*
//                    RecordHashValue(pcon, pste, cmBest, Beta, Alpha, Beta, depth, false);
                    stats.NM3++;
                    return Beta;
                }
            }
            if( val < Alpha && (pste-1)->reduced ) {
                // there will be a re-search
                return valMIN;
            } else if( val <= -valMATEscore ) {
            // extend one ply if there is a mate threat
                // ok storing threat in hash does not work for some reason...
//                if( pste->plyRem >= 4 )
//                    nm_threat = true;
    #if 0
                sprintf(sz_score, "MATE threat in NM(%d)", val);
                VDumpLine(pcon, pste, sz_score);
                val = val;
    #endif
                // only setting king_attack so that we don't reduce the search
                // if we extend then we can search a lot deeper for no reason when
                // there is a mate but winning side does not play the winnin move
                king_attak = true;
                // threat flag should be stored in hash too for next search
            }
        }
	}
#endif

#ifdef FHR
    // we need a real threat value, so don't use now
    if( !((nt == nodePV)||king_attak||iaminCheck) && (MaterialValue(pste) - FHR_THREAT >= Beta) 
        && (pste->plyRem > 0)
        && ((pste->valPcUs + pste->valPcThem) >= 1*valQUEEN)
        && !phashp->pawnon6or7
        ) {
        pste->plyRem --;
    }
#endif


//        king_attak |= nm_threat;

//    if( pste->hashkPc == 0x00001720c10998d8 ) {
//        Print(pcon, pste);
//        val = val;
//    }

#ifdef USE_HASH
#ifdef USE_IID
    // Note : we are doing IID on every kind of nodes, not doing that will result in more total nodes
    // Does this means that my search is sub optimal ?
    if( hash_entry == NULL && pste->plyRem >= 4 ) {
        // depth /2 gives worse results (more nodes)
//        int reduc = pste->plyRem / 2;
        int reduc = 2;
        pste->plyRem -= reduc;
        int iidval = Search(pcon, pste, Alpha, Beta, nt);
#if 0
        /* frombeowulf : Re-search properly if the previous search failed low, so that we know we're getting
         * a good move, not just the move with the highest upper bound (which is essentially
         * random and depends on the search order.) */
        if( iidval <= Alpha )
            iidval = Search(pcon, pste, valMIN, Alpha+1, nt);
#endif
        pste->plyRem += reduc;
        hash_entry = ProbeHashValue(pcon, pste, &val, Alpha, Beta, depth);
        // Is it possible to not have a hash move here ? (how many times ?)
    }
#endif
#endif

	//	Mark this move so if I find it in a sub-tree I can detect a rep
	//	draw.
	//
	VSetRep(pcon, pste->hashkPc);
	//	Generate moves.
	//
    bGenMoves(pcon, pste);
	//
	//	I know how many reversible moves have been made up until now.
	//	I just tried to make another move.  If I am here, I didn't
	//	leave myself in check, so I succeeded.
	//
	//	If the number of reversible moves made before this is at
	//	least 50 for each side, this was a draw.
	//
    int lmc = legalMoveCount(pste);
#ifdef DISTANCE_PRUNING
#if 0
//    int i_am_mate_score = mateValue(depth + 1);
//    int i_am_mate_score = mateValue(depth);
    int i_am_mate_score = mateValue(depth + 2);
    if( iaminCheck && lmc == 0)
        if( i_am_mate_score > Alpha )
            i_am_mate_score = mateValue(depth);
    if( i_am_mate_score > Alpha ) {
        Alpha = i_am_mate_score;
        if(i_am_mate_score >= Beta) {
		    VClearRep(pcon, pste->hashkPc);
//            return Beta; //i_am_mate_score;
            return i_am_mate_score;
        }
    }
#endif
#endif

	if ( (pste->plyFifty >= 50 * 2) && (lmc > 0) ) {
		pste->ccmPv = 0;
		VClearRep(pcon, pste->hashkPc);
		return 0;
	}
    // single move extension
    if( (lmc == 1) )
        preExtend = 1;

#if 1
    Killers[depth+4].raz();
    MateKillers[depth+4].raz();
#endif

    PrepareSort( pcon, pste, depth, phashp );
	//	Mark the PV and hash moves so they sort first.
	//
	if (lastPV[depth].m != 0) {
		SortPV(pste, depth);
		lastPV[depth].m = 0;
	}
#ifdef USE_HASH
	if (hash_entry != NULL)
		SortHash(pste, hash_entry);
#endif
#ifdef USE_KILLERS
    SortKillers(pste, depth);
#endif

    int a = Alpha;
    int b = Beta;


#ifdef MULTI_CUT
    const int R = 2;
    if( (nt == nodeCUT) && (!(iaminCheck||king_attak)) && (pste->plyRem >= R) ) {
        int c = 0;
        int m = 0;
        const int C = 3;
        const int M = 10;
        foreachmove(pste,pcm) {
            SortNext(pste, pcm);
		    bMove(pcon, pste, pcm->m);
		    (pste + 1)->plyRem = pste->plyRem - 1 - R;
#ifdef EXTEND_CHECK
            if( (pste + 1)->checkf )
                (pste + 1)->plyRem++;
#endif
            int val = -Search(pcon, pste + 1, - b, - a, nodeALL);
            bUndoMove(pcon, pste, pcm->m);
            if( val >= Beta ) {
                c++;
                if( c >= C ) {
				    VClearRep(pcon, pste->hashkPc);
                    return Beta;
                }
            }
            m++;
            if( m >= M )
                break;
        }
    }
#endif

#ifdef RECAPTURE_EXTENSION
    // margin to include Knight, Bishop and rook
    int prevCaptureValue    = pste->prevCaptValue;
    int prevCaptureValueMin = pste->prevCaptValue - 200;
    int prevCaptureValueMax = pste->prevCaptValue + 200;
    int prevCaptureSqr = pste->prevCaptSqr;
#endif

    bool fFound = false;
    int  c_move = 0;

    bool f_prune = false;
    int  f_max   = valMAX;
    int  f_score = 0;

#define FUTILITY_MARGIN (200)
#define FUTILITY_MARGIN_PRE (300)
//#define FUTILITY_MARGIN (200)
//#define FUTILITY_MARGIN_PRE (300)
#define FRONTIER    0

#ifdef FUTILITY_PRUNING
    if( !iaminCheck 
//        && ((pste->valPcUs + pste->valPcThem) >= 2*valQUEEN)
        && ((pste->valPcUs + pste->valPcThem) >= 1*valQUEEN)
        && !phashp->pawnon6or7
        ) {
        if( pste->plyRem == FRONTIER ) {
    //        int posVal = ValEval(pcon, pste, Alpha, Beta);
            f_score = MaterialValue( pste ) + FUTILITY_MARGIN ;
            f_prune = true;
            f_max   = f_score;
        } else  if( pste->plyRem == (FRONTIER+1) ) {
            f_score = MaterialValue( pste ) + FUTILITY_MARGIN_PRE ;
            f_prune = true;
            f_max   = f_score;
        }
    }
#endif

	//
	//	Iterate through the move list, trying everything out.
	//
    foreachmove(pste,pcm) {
		//
		//	All moves are sorted on the cmk field
		//	cmk is set in PrepareSort() and in other sort functions

        SortNext(pste, pcm);
		//
		//	Make the move

		bMove(pcon, pste, pcm->m);

//        unsigned int old_cmk = pcm->cmk;
//        History.seen( pcm, nt );
        int extension = preExtend;
        // don't try to reduce if we just extended the search
        // Note : now I allow the reduction after the extension, 
        // this is needed for the check extension
//        bool can_reduce = (extCount + extension) > 0 ? false : true;
        bool can_reduce = (extension) > 0 ? false : true;
        can_reduce = can_reduce && !king_attak;
//        bool ok_to_extend = (nt != nodeALL);
        // only do some kind of extension on PV nodes
        bool ok_to_extend = (nt == nodePV);

        if( (pste + 1)->checkf ) {
            can_reduce = false;
#ifdef EXTEND_CHECK
            // extend everywhere in the tree
            if( pste->plyRem == FRONTIER || extCount < 77) { // 5,7,77
                // stop extending if I lead the score
                if( a < valKNIGHT )
                    extension = 1;
            }
#else
            // only extend on check at the frontier
            if( pste->plyRem == FRONTIER ) {
                // stop extending if I lead the score
                if( a < valKNIGHT )
                    extension = 1;
            }
#endif
        }
        // Pawn push to 7th
        if( ok_to_extend && moveisPawnMove(pcm) && moveto7thRank(pcm) ) {
            extension = 1;
            can_reduce = false;
        }

#ifdef RECAPTURE_EXTENSION
        if( ok_to_extend && prevCaptureValue && Capt(pcm->m) && (To(pcm->m) == prevCaptureSqr) 
//            && moveisGoodCapture(pcm) 
            ) { 
            int matValue = recapMaterialValue(pcon, pcm);
            if( true || (matValue >= prevCaptureValueMin) && (matValue <= prevCaptureValueMax)) {
                stats.recapture++;
                extension = 1;
                can_reduce = false;
            }
        }
#endif
#if 0
        // They gave material to give check ?
        if( (pste->plyRem == FRONTIER) && iaminCheck && Capt(pcm->m ) ) {
            can_reduce = false;
            extension = 1;
        }
        // I gave material to give check in the last move
        if( (pste->plyRem == FRONTIER) && (prevCaptureValue > 0) && (pste - 1)->checkf ) {
            can_reduce = false;
            extension = 1;
        }
#endif
        //  ==========================================================================
		//	Set up for recursion.
		//
		(pste + 1)->plyRem = pste->plyRem - 1;
        if( extension && (extCount < maxExt) ) {
            (pste + 1)->plyRem++;
        }
        //
		//	Recurse into normal search or quiescent search, as appropriate.
		//
        nodeType succNType;
        if(nt == nodeCUT )
            succNType = nodeALL;
        else if(nt == nodeALL || fFound)
            succNType = nodeCUT;
        else 
            succNType = nodePV;
        // the above does not work like the old code below, it gives the double of nodes ???
        // nodePV is only checked for null move so wth ?
// lets try a bit without the next line 
//        succNType = fFound ? nodeCUT : nodePV;

#ifdef FUTILITY_PRUNING

        if( fFound && (nt != nodePV) && can_reduce && f_prune
            && !moveisPromotion(pcm) && !moveisCapture(pcm) && !moveisPawnMove(pcm)
//            && (f_max + MaterialGain(pcon, pcm) <= a)
//            && History.can_reduce(pcm)
            ) 
            if (f_max  <= a) {
                stats.FP_frontier++;
    		    bUndoMove(pcon, pste, pcm->m);
                // check that
//                int approx = f_max + MaterialGain(pcon, pcm);
                int approx = f_max;
                if( approx > bestVal ) {
                    // is that needed ?
                    bestVal = approx;
			        cmBest = *pcm;
			        pste->argcmPv[0] = cmBest;
			        pste->ccmPv = 1;
                }
                continue;
            } else if( false && (c_move >= 10) && History.can_prune( pcm ) ) {
                stats.FP_frontier++;
    		    bUndoMove(pcon, pste, pcm->m);
                continue;
            }
#endif

//        extension += extension;
#ifdef _DEBUG
        if( false && (pste)->hashkPc == 0x437875d1e38 
//            && (nt != nodePV)
            ) {
            unsigned int st = History.get_stats( pcm );
            int count = History.seen_count( pcm );
            unsigned int fh = (st >> 16) & 0xffff;
            unsigned int fl = (st) & 0xffff;
            char c = '.';
//            sprintf(sz_score, " ... (%d,%d/%d) r=%3.2f [%d,%d] %c", fh, fl, count, 100.0f*fl/count, Alpha, Beta, c);
//            VDumpLine(pcon, pste, sz_score);
        }
#endif
#ifdef USE_LMR
        // Note : LMR not activated until I get non random history counters
        //
        // move order is [H] [good capt] [equ capt] [killers] [bad capt] others
        if( 
//            (c_move >= 7)
            (c_move >= 3)
            && (nt != nodePV) && (!iaminCheck) && can_reduce 
            && (pste->plyRem >= 1) 
//            && (pste->plyRem <= 5)
//            && (pste->plyRem >= 3)  // results ok
//            && (pste->plyRem >= 4)
            && (a > -valMATEscore)
            && !moveisPromotion(pcm) && !moveisCapture(pcm) && !moveisPawnMove(pcm)
            && History.can_reduce(pcm)
//            && ((pcm->cmk & cmkMASK) == 0)
            ) {
            // search with a reduced depth
            stats.LMR++;
            pste->reduced = true;
		    (pste + 1)->plyRem --;
            val = -Search(pcon, pste + 1, - Beta, - a, succNType);
            pste->reduced = false;
            if( val >= Beta ) {
//            if( val > a ) {
                // re-search at normal depth
                stats.LMRreSearch++;
    		    (pste + 1)->plyRem ++;
                val = -Search(pcon, pste + 1, - Beta, - a, succNType);
            }
        }
        else
#endif
#ifdef KING_ATTACK
        // only play moves that gives check
        if( (pste + 1)->checkf == 0 && king_attacker == pste->side )
            val = valMIN;
        else
#endif
#if 0
        {
            val = -Search(pcon, pste + 1, - b, - a, succNType);
        }

#ifdef NULL_WINDOW
        if( (fFound) && (val > a) && ( val < Beta)) {
//        if( (fFound) && (val > a)) {
            // this is a new PV
//            static char spaces[30] = "                             ";
            int old_val = val;
//            sprintf(sz_score, "%s 1)old search=%d [%d %d]", spaces+(30-depth), val, a, b);
//            VDumpLine(pcon, pste, sz_score);
            val = -Search(pcon, pste + 1, - Beta, - a, nodePV);
//            sprintf(sz_score, "%s 2)new search=%d [%d %d]", spaces+(30-depth), val, a, Beta);
//            VDumpLine(pcon, pste, sz_score);
//            val = -Search(pcon, pste + 1, - Beta, - val, nodePV);
//            if( val < a || val < bestVal ) {
//                printf(" Oups in re-search %d %d\n", old_val, val);
//                fflush(stdout);
//            }
            stats.reSearch++;
//            nt = nodePV;
        }
#endif
#else
        if (nt != nodePV || !fFound) {
            val = -Search(pcon, pste + 1, - Beta, - a, succNType);
        } else { // PV moves
            val = -Search(pcon, pste + 1, - b, - a, succNType); // cut nodes
            if( val > a ) { // re-search
                stats.reSearch++;
                val = -Search(pcon, pste + 1, - Beta, - a, nodePV); // new PV
            }
        }
#endif

#ifdef _DEBUG
        if( false && depth <= 3 ) {
            sprintf(sz_score, " ... %d [%d %d] " , val, Alpha, Beta);
            VDumpLine(pcon, pste, sz_score);
        }
#endif

		bUndoMove(pcon, pste, pcm->m);
		if ((pcon->fAbort) || (pcon->fTimeout)) {
			VClearRep(pcon, pste->hashkPc);
			return 0;
		}
		//	We got a value back.  We unmade the move.  We're not dead.  Let's
		//	see how good this move was.  If it was >= "beta", it was so good
		//	that we don't need to search for anything better, so we'll leave.
		//
		//	If it was not >= "beta", but it was > "alpha", this is better than
		//	anything else we've found before, but not so good we have to
		//	leave.  These kinds of moves are actually quite rare.  If I find
		//	one of these, I have to store it in the PV (main-line) that I'm
		//	constructing.  This might end up being the main-line for the whole
		//	search, if it gets backed up all the way to the root.
		//
#ifdef _DEBUG
#if 0
        if( pste->side == W && val < -300 ) {
            sprintf(sz_score, " ... %d [%d,%d]", val, Alpha, Beta);
            VDumpLine(pcon, pste, sz_score);
        }
#endif
#endif
#ifdef HISTORY
//        if( val <= Alpha )
//            History.failed_low( pcm, nt );
#endif
        if( val > bestVal ) {
            bestVal = val;
		    if (val > a) {
                if (val >= Beta) {
				    //
				    //	This move failed high
				    //
                    //  Set the new killer move only if it was not marked as good allready
                    //  'Good' moves are allready sorted first so storing them in the killer
                    //  array is counter productive
                    // This was added to see mate search ; with the mate distance bound
                    // no pv was recorded 
			        cmBest = *pcm;
//                    cmBest.cmk = iMove;
#ifdef _DEBUG
                    cmBest.cmk = History.get_ratio( pcm );
#endif
                    copyPV( pste, cmBest );
				    VClearRep(pcon, pste->hashkPc);
                    goto cutNode;
			    }
			    //	This move is between alpha and beta, which is actually pretty
			    //	rare.  If this happens I have to add a PV move, and append the
			    //	returned PV to it
			    //
			    cmBest = *pcm;
//                cmBest.cmk = iMove;
#ifdef _DEBUG
                cmBest.cmk = History.get_ratio( pcm );
#endif
                copyPV( pste, cmBest );
			    a = bestVal;

            } else {
                if( !fFound ) {
                    copyPV( pste, *pcm );
                }
            }
        }
		fFound = true;
        c_move++;

#ifdef NULL_WINDOW
        b = a + 1;              // set the new null window
#endif
//        if( nt == nodeCUT )
//            nt = nodeALL;
	}
	VClearRep(pcon, pste->hashkPc);

	if (!fFound) {
        // We have no move !!
		if (iaminCheck) {
            // Checkmate
			bestVal = mateValue(depth);
            if( depth >= 10 )
                bestVal = bestVal;
        } else {
            // Draw
            bestVal = 0;
//            char sz_score[50];
//            sprintf(sz_score, "stalemate");
//            VDumpLine(pcon, pste, sz_score);
        }
        // ==================================
        return bestVal;
        // ==================================
	}
    // the node type is not accurate, so I can not increment my (bad/good) move counters
    // in the move loop :(
    // if I have a move, then this is not an all nodes
cutNode:;

//    if( (iMove < moveCount) || (nt == nodePV) ) {
    if( bestVal > Alpha ) {
//    if( nt != nodeALL ) {
        if( cmBest.m == 0 )
            nt = nt;
        if( bestVal >= Beta ) {
            stats.cCUT++;
            if(iMove == 0)
                stats.cCUTrank1++;
            if(iMove < 5)
                stats.cCUTrankn++;
        }

#ifdef HISTORY
        if( bestVal >= Beta 
            && !moveisCapture(&cmBest) ) {   // don't update on good captures
            // increment the 'seen' count, if I don't increment the FH hit then it's
            // a FL move
            // ==> no need to do that if its is a capture ?
            int c = 0;
            // caution if the sortNext function changes
            // played moved are at the begining of pste->mlist_start
            // all the moves are bad or only the last played moves ?
            // we don't have any info on the not played moves...
//            foreachmove(pste,pcm) {
//                History.seen( pcm, nt );
//            }
            for(CM *badcm = pste->mlist_start ; c < iMove ; c++, badcm++ )
                History.seen( badcm, nt );
            // increment FH counter and classic history counter
            History.update( &cmBest, depth, nt );
            if( iMove < moveCount )   // Beta cut
                History.seen( &cmBest, nt );
        }
#endif
#ifdef USE_KILLERS
        if( bestVal >= valMATEscore )
            MateKillers[depth].set(&cmBest);
        else if( !moveisGoodCapture(&cmBest) )
            Killers[depth].set(&cmBest);
#endif
    }

    RecordHashValue(pcon, pste, cmBest, bestVal, LOWERBOUND, Beta, depth, nm_threat);
//    if( nt == nodePV )
//        History.update( pcmBest, depth );
	return bestVal;
}

//	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-	-

//	Full-width search.  This is the heart of the "chess playing" part of this
//	program.  It performs a recursive alpha-beta search.

static struct {
    char    move[12];
    int     c_nodes;
    int     hval;
} sortDebug[200];
static int c_sortDebug;

#define maxRootMoves    200
static struct _rootMove {
    char szmove[8];
    int score;
    int hint;
    U64 nodes;
    bool    evaluated;
    CM cm;
} rootMoves[maxRootMoves];
static int rootMovesCount;

static int sortScore( struct _rootMove & move) {
    return move.hint * 10000 + move.score;
}

int ValSearchRoot(PCON pcon, PSTE pste, int Alpha, int Beta)
{
    int a = Alpha;
    int b = Beta;
    int bestVal = valMIN;
    int	val;
//    int depth = 1;
    int depth = int(pste - pcon->argste + 1);
    pste->lastMove = 0;
    pste->reduced = false;

    c_sortDebug = 0;

    //	Increment global node counter
    //
    ++pcon->ss.nodes;

    //	Stuff is doone a little differently at the root than at any other
    //	ply.
    //
    //
    //	Mark this move so if I find it in a sub-tree I can detect a rep
    //	draw.  I don't check to see if a rep is already set, because if
    //	we are here, we have to be making a move, not whining about a
    //	blown chance for a draw.  I also don't try to cut off from the
    //	hash table, for the same reason.  I want a move, not a score.
    //
    //	I do probe the table though, in case there is "best move" info
    //	in there.
    //
    PHASH hash_entry = ProbeHashValue(pcon, pste, &val, Alpha, Beta, depth);

    VSetRep(pcon, pste->hashkPc);
    //
    //	If it's the first time we come here (depth == 0)
    //	generate the moves
    //
    if( pste->plyRem == 0 ) {
        // first time with this position
        bGenMoves(pcon, pste);
        pcon->ss.ccmLegalMoves = legalMoveCount(pste);
        rootMovesCount = pcon->ss.ccmLegalMoves;

        PHASHPawn phashp;
        EvalPawns(pcon, pste, phashp);
        PrepareSort( pcon, pste, depth, phashp );

        foreachmove(pste,pcm) {
            CmToSz( pcm, rootMoves[iMove].szmove );
            rootMoves[iMove].cm     = *pcm;
            rootMoves[iMove].hint   = 0;
            rootMoves[iMove].evaluated = false;
#if 0
            rootMoves[iMove].score  = - iMove; // keep the generation order
            if( Capt(pcm->m) )
                rootMoves[iMove].score -= 100;  // captures at the end
#else
            rootMoves[iMove].score = pcm->cmk & 0xffff;
            if( moveisCapture( pcm )) {
                if( moveisGoodCapture( pcm ) )
                    rootMoves[iMove].hint = 1;
                else
                    rootMoves[iMove].score = -1;
            }
#endif
            rootMoves[iMove].nodes  = 0;
        }
    }
    {
        // after the first time (first depth iteration)
        // sort the moves by node count
#ifdef USE_HASH
        PHASH phash = hash_entry;
        U32 hm = phash ? phash->getMove() : 0;
#endif
        for(int i = 0; i < rootMovesCount; i++) {
            PCM pcm = &rootMoves[i].cm;
            if ( rootMoves[i].cm.m == lastPV[1].m )
                rootMoves[i].hint = 10;
#ifdef USE_HASH
		    else if ( pcm->m == hm )
                rootMoves[i].hint = 1;
#endif
        }
    }
	//	Mark the PV and hash moves so they sort first.
	//
    // bubble sort
    for(int i = 0; i < rootMovesCount - 1; i++) {

        for(int j = i+1 ; j < rootMovesCount; j++)
            // most moves have the same score because of null window search
            if( (sortScore(rootMoves[j]) > sortScore(rootMoves[i]) ||
                (sortScore(rootMoves[j]) == sortScore(rootMoves[i])) && (rootMoves[j].nodes > rootMoves[i].nodes) )) {
                _rootMove swap = rootMoves[i];
                rootMoves[i] = rootMoves[j];
                rootMoves[j] = swap;

            }
    }

    CM cmBest(0,0);
	bool fFound = false;
	//
	//	Iterate through the move list, trying everything out.
	//
//	for (PCM pcm = pste->pcmFirst; pcm < (pste + 1)->pcmFirst; pcm++) {
    for(int i = 0; i < rootMovesCount; i++) {
        PCM pcm = &rootMoves[i].cm;
        pcm->cmk = i;
		//
		//
		//	Things are done differently at the root of the search.
		//
        SAN	san;
        //	Record information about which move I am searching ...
        //
        pcon->ss.icmSearching = i+1;
        CmToSan(pcon, pste, pcm, &san);
        SanToSz(&san, pcon->ss.aszSearching);
        CmToSz(pcm, pcon->ss.aszSearchSan);
        rootMoves[i].nodes = pcon->ss.nodes;
        VPrSendUciCurrMove( pcon->ss );

#ifdef _DEBUG
        {
            sortDebug[c_sortDebug].hval = 0;
            sortDebug[c_sortDebug].c_nodes = (int)pcon->ss.nodes;
            strcpy(sortDebug[c_sortDebug].move , pcon->ss.aszSearching);
            strcat(sortDebug[c_sortDebug].move , " ");
            if( pcm->cmk & cmkPV )
                strcat(sortDebug[c_sortDebug].move , "V");
            if( pcm->cmk & cmkHASH )
                strcat(sortDebug[c_sortDebug].move , "H");
            if( pcm->cmk & cmkKILLER )
                strcat(sortDebug[c_sortDebug].move , "K");
            if( pcm->cmk & cmkGOODCAPT )
                strcat(sortDebug[c_sortDebug].move , "G");
            if( pcm->cmk & cmkBADCAPT )
                strcat(sortDebug[c_sortDebug].move , "b");
            if( (pcm->cmk & cmkNONE) == cmkNONE)
                sortDebug[c_sortDebug].hval = pcm->cmk & ~0xfe000000;

        }
#endif
        //
        //	... then go make the move.
        //
        bMove(pcon, pste, pcm->m);
        bool can_reduce = true;
        int extension = 0;
        //
		//	Set up for recursion.
		//
		(pste + 1)->plyRem = pste->plyRem - 1;
#if 0
        if( (pste + 1)->checkf ) {
            can_reduce = false;
            // if I extend before max depth then I could get a doubled depth at the end
            // and it's a waste
            if( pste->plyRem == FRONTIER ) {
                (pste + 1)->plyRem++;
                extension = 1;
            }
        }
#endif
		//
		//	Recurse into normal search or quiescent search, as appropriate.
		//
        nodeType succNType;
        if(fFound)
            succNType = nodeCUT;
        else 
            succNType = nodePV;

        // CAUTION TEST ONLY
        val = -Search(pcon, pste + 1, - b, - a, succNType);
//        val = -Search(pcon, pste + 1, -45000, +45000, succNType, extension);

#ifdef NULL_WINDOW
        if( (fFound) && (val > a) ) {
            val = -Search(pcon, pste + 1, - Beta, - a, nodePV);
            stats.reSearch++;
        }
#endif
        // update the next sort key
        rootMoves[i].hint  = 0;
        if( ! rootMoves[i].evaluated )
            rootMoves[i].score = val;
        rootMoves[i].nodes = pcon->ss.nodes - rootMoves[i].nodes;

        // debugging stuff
        sortDebug[c_sortDebug].c_nodes = int(pcon->ss.nodes) - sortDebug[c_sortDebug].c_nodes;
        c_sortDebug++;

		bUndoMove(pcon, pste, pcm->m);
		if ((pcon->fAbort) || (pcon->fTimeout)) {
			VClearRep(pcon, pste->hashkPc);
			return 0;
		}
		//	We got a value back.  We unmade the move.  We're not dead.  Let's
		//	see how good this move was.  If it was >= "beta", it was so good
		//	that we don't need to search for anything better, so we'll leave.
		//
		//	If it was not >= "beta", but it was > "alpha", this is better than
		//	anything else we've found before, but not so good we have to
		//	leave.  These kinds of moves are actually quite rare.  If I find
		//	one of these, I have to store it in the PV (main-line) that I'm
		//	constructing.  This might end up being the main-line for the whole
		//	search, if it gets backed up all the way to the root.
        //
#if 0
        if( !strcmp(rootMoves[i].szmove, "a8a1" )) {
//                printf("# last bestval in pv is %d\n", pcon->ss.lastVal);
                printf("# Qa1 %d (%d) (%d %d)\n", val, bestVal, a, b);
                fflush(stdout);
//                copyPV( pste, *pcm );
//			    VDumpPv(pcon, pcon->ss.plyDepth, val, prsaNORMAL);
        }
#endif
        if( val > bestVal ) {
            bestVal = val;
		    if (val > a) {
			    if (val >= Beta) {
                    // Note : not reached if not doing aspiration search
				    //
				    //	This move failed high, so we are going to return beta,
				    //	but if we're sitting at the root of the search I will set
				    //	up a one-move PV (if this isn't already the move I'm
				    //	planning to make), so this move will be made if I run out
				    //	of time before resolving the fail-high.
				    //
                    copyPV( pste, *pcm );

				    //	This function needs the root moves generated, which
				    //	they are.
				    //
                    RecordHashValue(pcon, pste, *pcm, bestVal, LOWERBOUND, Beta, depth, false);
                    Beta = val;
				    VDumpPv(pcon, pcon->ss.plyDepth, Beta, prsaFAIL_HIGH);
				    VClearRep(pcon, pste->hashkPc);
				    return bestVal;
			    }
			    //	This move is between alpha and beta, which is actually pretty
			    //	rare.  If this happens I have to add a PV move, and append the
			    //	returned PV to it, and if I'm at the root I'll send the PV to
			    //	the interface so it can display it.
			    //
                rootMoves[i].score = val;
                rootMoves[i].evaluated = true;
			    cmBest = *pcm;
                copyPV( pste, cmBest );
			    a = bestVal;
			    //
			    //	This function needs the root moves generated, and they
			    //	are.
			    //	
			    VDumpPv(pcon, pcon->ss.plyDepth, a, prsaNORMAL);
#if 1 //_DEBUG
                stats.report();
#endif
		    }
        }
#ifdef NULL_WINDOW
        b = a + 1;              // null window
#endif
		fFound = true;
	}
#if 0
    printf("# SORT was ");
    for(int i = 0; i < c_sortDebug && i < 10; i++) {
        if( sortDebug[i].c_nodes < 10000 )
            printf("%s (%d) ", sortDebug[i].move, sortDebug[i].c_nodes );
        else
            printf("%s (%dk) ", sortDebug[i].move, sortDebug[i].c_nodes / 1000 );
        if( sortDebug[i].hval > 0 )
            printf("%d ", sortDebug[i].hval );
    }
    printf("\n");
#endif
	VClearRep(pcon, pste->hashkPc);
	if (!fFound) {
        // We have no move !!
		if (pste->checkf) {
            // Checkmate
			bestVal = mateValue(depth);
		} else
            // Draw
            bestVal = 0;
	}
    RecordHashValue(pcon, pste, cmBest, bestVal, LOWERBOUND, Beta, depth, false);
	return bestVal;
}
